#include <sys/types.h>
#include <stdio.h>
#include <time.h>
#include <sys/times.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "db.h"


const char prefixMsg[] = "suit01";

# define ENV_DIRECTORY "./txnTemp/"
# define TT_DEBUG_FILENAME1 "TTDebugFile1"
# define TT_DEBUG_FILENAME2 "TTDebugFile2"
# define MAX_NUM_RECORDS 1000
# define FALSE 0
# define TRUE 1
# define NUM_WORKERS 2
# define TT_DEBUG 1

typedef struct dbInformation {
    DB *dbp;
    DB_ENV *dbenv;
    int id;
    int numRecords;
} dbInformation;


int displayDB(char * dbName, int openFlag, int numRecords);
int dbStartup(char *dbName, DB **dbpPtr, DB_ENV **dbenvPtr, DB_TXN ** tidPtr, int openFlag);
int dbShutdown(DB *dbp, DB_ENV *dbenv, DB_TXN * tid, int printTT);
int insertData(DB *dbp, DB_TXN *tid, int keyVal, char * strData);

int printTT(DB_ENV *dbenv);

int displayRawDB(DB *dbp, DB_TXN * tid) {

    DBC *dbcp;
    DBT key, data;
    int ret;
    struct timeval start, stop, beginning, forever, prevStop, prevStart;

    printf("Retrieving the content using a cursor (Non Sequenced)\n");

    /* Acquire a cursor for the database. */
    if ((ret = dbp->cursor(dbp, tid, &dbcp, 0)) != 0) {
        dbp->err(dbp, ret, "DB->cursor");
        return(-1);
    }

    /* Initialize the key/data return pair. */
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    /* int DBcursor->c_getnonsequenced(DBC *cursor, DBT *key, DBT *data, struct timeval start,
     * struct timeval stop, u_int32_t flags, struct timeval *outstart, struct timeval *outstop);
     */

    /* Walk through the database and print out the key/data pairs. */
    beginning.tv_sec = BEGINNING;
    beginning.tv_usec = 0;
    forever.tv_sec = FOREVER;
    forever.tv_usec = 0;
    prevStop.tv_sec = 0;
    prevStop.tv_usec = 0;
    prevStart.tv_sec = 0;
    prevStart.tv_usec = 0;

    while ((ret = dbcp->c_get_nonsequenced(dbcp, &key, &data, beginning, forever,
                                          DB_NEXT, &start, &stop)) == 0) {

        /* Todo Comment out the actual time display in Release Test Suite */
        printf("(%d, %.*s) start-time: (((%d))), stop-time: (((%d)))\n",
               *(int *)key.data, (int)data.size, (char *)data.data, start.tv_sec, stop.tv_sec);
    }

    if (ret != DB_NOTFOUND && ret != 0) {
        dbp->err(dbp, ret, "DBcursor->get");
        return(-1);
    }
    if ((ret = dbcp->c_close(dbcp)) != 0) {
        dbp->err(dbp, ret, "db cursor close");
    }

    return (0);
}

int main (int argc, char *argv[]) {
    DB *dbp;
    DBC *dbcp;
    DB_ENV *dbenv;
    DBT key, data;
    DB_TXN * tid;

    int retVal, numRecords, numVersions, numRecordsToRead,
	openFlag = DB_CREATE|DB_TEMPORAL;
	/*openFlag = DB_CREATE; */

    if ((argc == 2) && (strcmp(argv[1], "partition") == 0))
	openFlag |= DB_PARTITION;

    numRecords = 10;
    numVersions = 3;
    numRecordsToRead = 2;

    /* test individual testcase */
	if(argc == 2){
		if(strcmp(argv[1], "1") == 0){
       		printf("[%s]Test 1 --@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n", prefixMsg);
			retVal = advTest1("bdbTest1.db", "./txnTemp/bdbTest1.db",
		     openFlag, 9, numVersions, MAX_NUM_RECORDS,TRUE,TRUE,TRUE);
		     fflush(stdout);
		}
		else if(strcmp(argv[1], "2") == 0){
       		printf("[%s]Test 2 --@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n", prefixMsg);
       		retVal = advTest2("bdbTest2.db", "./txnTemp/bdbTest2.db",
		     openFlag, numRecords, numVersions, MAX_NUM_RECORDS,TRUE,TRUE,TRUE);
		     fflush(stdout);
		}
		else if(strcmp(argv[1], "3") == 0){
       		printf("[%s]Test 3 --@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n", prefixMsg);
       		retVal = advTest3("bdbTest3.db", "./txnTemp/bdbTest3.db",
		     openFlag,numRecords, numVersions, numRecordsToRead, TRUE, TRUE, TRUE);
		     fflush(stdout);
		}
		else if(strcmp(argv[1], "4") == 0){
       		printf("[%s]Test 4 --@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n", prefixMsg);
       		retVal = advTest4("bdbTest3.db", "./txnTemp/bdbTest3.db",
		     openFlag,numRecords, numVersions, numRecordsToRead, TRUE, TRUE, TRUE);
		     fflush(stdout);
		}
 		else if(strcmp(argv[1], "5") == 0){
       		printf("[%s]Test 5 --@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n", prefixMsg);
            retVal = advTest5("bdbTest5.db", "./txnTemp/bdbTest5.db",
		     openFlag,numRecords, numVersions, numRecordsToRead, TRUE, TRUE, TRUE);
       		//retVal = advTest5("bdbTest5.db", "./txnTemp/bdbTest5.db", openFlag, numRecords);
       		fflush(stdout);
		}
		else if(strcmp(argv[1], "6") == 0){
       		printf("[%s]Test 6 --@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n", prefixMsg);
            retVal = advTest6("bdbTest6.db", "./txnTemp/bdbTest6.db",
		     openFlag,numRecords, numVersions, numRecordsToRead, TRUE, TRUE, TRUE);

       		fflush(stdout);
		}
		else if(strcmp(argv[1], "7") == 0){
       		printf("[%s]Test 7 --@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n", prefixMsg);
            retVal = advTest7("bdbTest7.db", "./txnTemp/bdbTest7.db",
		     openFlag,numRecords, numVersions, numRecordsToRead, TRUE, TRUE, TRUE);

       		fflush(stdout);
		}

    }
    else {/* run all the testcase in one iteration */
		numVersions = 1;
		printf("[%s]Test 1 --@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n", prefixMsg);
		retVal = advTest1("bdbTest1.db", "./txnTemp/bdbTest1.db",
		 openFlag, 9, numVersions, MAX_NUM_RECORDS,TRUE,TRUE,TRUE);

		printf("\n[%s]Test 2 --@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n", prefixMsg);
		retVal = advTest2("bdbTest2.db", "./txnTemp/bdbTest2.db",
		 openFlag, numRecords, numVersions, MAX_NUM_RECORDS,TRUE,TRUE,TRUE);

		printf("\n[%s]Test 3 --@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n", prefixMsg);
		retVal = advTest3("bdbTest3.db", "./txnTemp/bdbTest3.db",
		 openFlag,numRecords, numVersions, numRecordsToRead, TRUE, TRUE, TRUE);

		printf("\n[%s]Test 4 --@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n", prefixMsg);
		retVal = advTest4("bdbTest3.db", "./txnTemp/bdbTest3.db",
		 openFlag,numRecords, numVersions, numRecordsToRead, TRUE, TRUE, TRUE);

		printf("\n[%s]Test 5 --@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n", prefixMsg);
		retVal = advTest5("bdbTest5.db", "./txnTemp/bdbTest5.db",
		 openFlag,numRecords, numVersions, numRecordsToRead, TRUE, TRUE, TRUE);

		printf("\n[%s]Test 6 --@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n", prefixMsg);
		retVal = advTest6("bdbTest6.db", "./txnTemp/bdbTest6.db",
		 openFlag,numRecords, numVersions, numRecordsToRead, TRUE, TRUE, TRUE);

		printf("\n[%s]Test 7 --@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n", prefixMsg);
		retVal = advTest7("bdbTest7.db", "./txnTemp/bdbTest7.db",
		 openFlag,numRecords, numVersions, numRecordsToRead, TRUE, TRUE, TRUE);

		fflush(stdout);
 	}
    return(0);
}


/* Test od_pgcnt on tuple insertions --
 *     Open database ->
 *     T1 begins -> insert 9 tuples -> split into two pages [-> print TT, od_pgcnt=0]
 *               -> write pages [-> print TT, od_pgcnt=2] -> T1 commits
 *     T2 begins -> insert 10th tuple [->print TT, T1.od_pgcnt=2, T2.od_pgcnt=0]
 *               -> write page [-> print TT, T1.od_pgcnt=1, T2.od_pgcnt=1] -> T2 commits
 *     T3 begins -> insert the 11st tuple [-> print TT, T1.od_pgcnt=1, T2.od_pgcnt=0(removed),T3.od_pgcnt=0]
 *               -> write page [-> print TT, T1.od_pgcnt=1, T3.od_pgcnt=1] -> T3 commits
 *     Close database
 */
int advTest1(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords, int numVersions, int numRecordsToRead, int printTTAfterTxnCommit, int printTTBforTxnCommit, int printTTBforSync)
{
    DB *dbp;
    DBC *dbcp;
    DB_ENV *dbenv;
    DBT key, data;
    DB_TXN * tid, *tid2;
    int ret, keyVal, version;
    char  strData[200];

    for (version=1; version <= numVersions; version++) {
/* modified by Huilong */
//sleep(2);
/* end Huilong*/
	printf("\n********************  Round %d  ************************\n", version);
	if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0)
	    return ret;

	printf("[%s] Txn1: inserting Data into the database \n", prefixMsg);

	/* make two pages of records */
	for (keyVal=1; keyVal <= 9; keyVal++) {
	    sprintf(strData, " Computer Science For Key %d Version %d. ", keyVal, version);
	    insertData(dbp, tid, keyVal, strData);
	    /* sleep(1); */
	}

	if (printTTBforSync == TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

	printf("[%s] Txn1: calling sync() to write buffer to memory \n", prefixMsg);
	memp_sync(dbenv, NULL);

    /* by now, the od_pgcnt should be 2 */
	if (printTTBforTxnCommit == TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}
	if((ret = txn_commit(tid, 0)) != 0){
		return -1;
	}
	printf("[%s] Txn1: committed \n", prefixMsg);

	printf("[%s] Txn2: inserting Data into the database for Key=10 \n", prefixMsg);
	if((ret = txn_begin(dbenv, NULL, &tid2, 0)) != 0){
		return -1;
	}

    /** insert the 10th Data in the Database **/
    keyVal = 10;
 	sprintf(strData, " Computer Science For Key %d Version %d. ", keyVal, version);
	insertData(dbp, tid2, keyVal, strData);

    /* od_pgcnt for the previous txn should still be 2 */
	if (TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

	printf("[%s] Txn2: calling sync() to write buffer to memory \n", prefixMsg);
    memp_sync(dbenv, NULL);

    /* od_pgcnt for the previous txn should become 1, and od_pgcnt for the current txn is 1 */
	if (TRUE) {
	    printf("[%s] Printing TT\n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

	printf("[%s] Txn2: committed \n", prefixMsg);
 	if((ret = txn_commit(tid2, 0)) != 0){
		return -1;
	}

	if((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0){
		return -1;
	}

    printf("[%s] Txn3: inserting Data in the database For Key = 11\n", prefixMsg);

    /** insert the 11st Data in the Database **/
    keyVal = 11;
 	sprintf(strData, " Computer Science For Key %d Version %d. ", keyVal, version);
	insertData(dbp, tid, keyVal, strData);

	if (TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

   printf("[%s] Txn3: calling sync() to write buffer to memory \n", prefixMsg);

   memp_sync(dbenv, NULL);
	/* by now, the od_pgcnt for the previous txn should become 1 since the first unstamped page being stamped
	   when reading the first record */
	if (TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

 	if((ret = txn_commit(tid, 0)) != 0){
		return -1;
	}
	printf("[%s] Txn3: committed \n", prefixMsg);
    printf("[%s] Shutdown the database\n", prefixMsg);
	tid = NULL;
	if ((ret = dbShutdown(dbp, dbenv, tid, printTTAfterTxnCommit)) != 0){
		printf("[%s] shutdown error\n", prefixMsg);
	    return ret;
	}
    }
    return(0);
}



 /* Test od_pgcnt on tuple insertions and retrieving --
  *     Open database ->
  *     T1 begins -> insert 10 tuples -> split into two pages [-> print TT, od_pgcnt=0]
  *               -> write pages [-> print TT, od_pgcnt=2] -> T1 commits
  *     T2 begins -> get 9th tuple -> stamp page -> write page [-> print TT, T1.od_pgcnt=1, T2.od_pgcnt=0]
  *               -> get 1st tuple -> stamp page -> [print TT, T1.od_pgcnt=0(removed), T2.od_pgcnt=0]
  *               -> write page [-> print TT, T2.od_pgcnt = 0] -> T2 commits [-> print TT]
  *     Close database
  *
  * Repeating this operation will have different TT results
  */

int advTest2(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords, int numVersions, int numRecordsToRead, int printTTAfterTxnCommit, int printTTBforTxnCommit, int printTTBforSync)
{
    DB *dbp;
    DBC *dbcp;
    DB_ENV *dbenv;
    DBT key, data;
    DB_TXN * tid;
    int ret, keyVal, version;
    char  strData[200];

    for (version=1; version <= numVersions; version++) {
//sleep(2);
	printf("\n********************  Round %d  ************************\n", version);
	if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0)
	    return ret;

/* added by Huilong at 11:16am 6/23/2006 */        
printTT(dbenv);
//        displayRawDB(dbp, tid);
//printf("env id:%d, txn id:%d\n", ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->envid, tid->tid);
/* end of add by Huilong*/

	printf("[%s] Txn1: inserting Data into the database \n", prefixMsg);

	/* make two pages of records */
	for (keyVal=1; keyVal <= 10; keyVal++) {
	    sprintf(strData, " Computer Science For Key %d Version %d. ", keyVal, version);
	    insertData(dbp, tid, keyVal, strData);
	    /* sleep(1); */
	}

/* added by Huilong at 11:16am 6/23/2006 */
//        displayRawDB(dbp, tid);
/* end of add by Huilong*/

	if (printTTBforSync == TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

	printf("[%s] Txn1: calling sync() to write buffer to memory \n", prefixMsg);
	memp_sync(dbenv, NULL);

    /* by now, the od_pgcnt should be 2 */
	if (printTTBforTxnCommit == TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}



	if((ret = txn_commit(tid, 0)) != 0){
		return -1;
	}
        
//printTT(dbenv); 
	printf("[%s] Txn1: committed \n", prefixMsg);
	if((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0){
		return -1;
	}
    printf("[%s] Txn2: Getting Data in the database For Key = 9\n", prefixMsg);

    /** Getting  Data in the Database **/
    keyVal = 9;
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    key.data = (int *) &keyVal;
    key.size = sizeof(keyVal);
    data.flags = DB_DBT_MALLOC;

    if ((ret = dbp->get(dbp, tid, &key, &data, 0)) == 0)
        printf("[%s] (key, data) retrieved: (%d, %.*s)\n", prefixMsg, *(int *)key.data, (int)data.size,
	       (char *)data.data);
    else {
        dbp->err(dbp, ret, "DB->get");
        return(-1);
    }

   printf("[%s] Txn2: calling sync() to write buffer to memory \n", prefixMsg);

   memp_sync(dbenv, NULL);

	if (TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

	if((ret = dbp->cursor(dbp, tid, &dbcp, 0)) != 0){
		return -1;
	}

	printf("[%s]Txn2: Getting first record by a cursor\n", prefixMsg);
	if((ret = dbcp->c_get(dbcp, &key, &data, DB_NEXT)) == 0){
		printf("[%s] (key, data) retrieved: (%d, %.*s)\n", prefixMsg,  *(int *)key.data, (int) data.size, (char *)data.data);
	}
    if (ret != DB_NOTFOUND && ret != 0) {
	dbp->err(dbp, ret, "DBcursor->get");
	return(-1);
    }
    if ((ret = dbcp->c_close(dbcp)) != 0) {
	dbp->err(dbp, ret, "db cursor close");
    }
 	if (TRUE) {
 	    printf("[%s] Printing TT \n", prefixMsg);
 	    ret = printTT(dbenv);
 	    if (ret != 0)
 		return (-1);
 	}

	printf("[%s] Txn2: calling sync() to write buffer to memory \n", prefixMsg);
	memp_sync(dbenv, NULL);
 	if (TRUE) {
 	    printf("[%s] Printing TT \n", prefixMsg);
 	    ret = printTT(dbenv);
 	    if (ret != 0)
 		return (-1);
 	}

	if((ret = txn_commit(tid, 0)) != 0){
		return -1;
	}

	printf("[%s] Txn2: committed \n", prefixMsg);
	/* by now, the od_pgcnt should become 1 since the first unstamped page being stamped
	   when reading the first record */
	if (TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

/* added by Huilong at 11:16am 6/23/2006 */
//       dbStartup(dbName, &dbp, &dbenv, &tid, openFlag);
//       displayRawDB(dbp, tid);
/* end of add by Huilong*/

    printf("[%s] Shutdown the database\n", prefixMsg);
	tid = NULL;
	if ((ret = dbShutdown(dbp, dbenv, tid, printTTAfterTxnCommit)) != 0){
		printf("[%s] shutdown error\n", prefixMsg);
	    return ret;
	}
    }

    return(0);
}

 /* Test od_pgcnt on tuple insertions and retrieving --
  *     Open database ->
  *     T1 begins -> insert 10 tuples -> split into two pages [-> print TT, od_pgcnt=0]
  *               -> write pages [-> print TT, od_pgcnt=2] -> T1 commits
  *     Close database ->
  *     Open database ->
  *     T2 begins -> get 9th tuple -> stamp page -> write page [-> print TT, T1.od_pgcnt=1, T2.od_pgcnt=0]
  *               -> get 1st tuple -> stamp page -> T2 commits [-> print TT, ]
  *     Close database
  *
  * Repeating this operation will have different TT results
 */


int advTest3(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords, int numVersions, int numRecordsToRead, int printTTAfterTxnCommit, int printTTBforTxnCommit, int printTTBforSync)
{
    DB *dbp;
    DBC *dbcp;
    DB_ENV *dbenv;
    DBT key, data;
    DB_TXN * tid;
    int ret, keyVal, version,i;
    char  strData[200];

    for (version=1; version <= numVersions; version++) {
/* modified by Huilong */
//sleep(2);
/* end Huilong*/
	printf("\n********************  Round %d  ************************\n", version);
	if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0)
	    return ret;

	printf("[%s] Txn1: inserting Data into the database \n", prefixMsg);

	/* make two pages of records */
	for (keyVal=1; keyVal <= numRecords; keyVal++) {
	    sprintf(strData, " Computer Science For Key %d Version %d. ", keyVal, version);
	    insertData(dbp, tid, keyVal, strData);
	    /* sleep(1); */
	}

	if (printTTBforSync == TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

	printf("[%s] Txn1: calling sync() to write buffer to memory \n", prefixMsg);

	memp_sync(dbenv, NULL);

    /* by now, the od_pgcnt should be 2 */
	if (printTTBforTxnCommit == TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}


        if((ret = txn_commit(tid, 0)) != 0){
		return -1;
	}

	printf("[%s] Txn1: committed \n", prefixMsg);

	tid = NULL;
	if ((ret = dbShutdown(dbp, dbenv, tid, printTTAfterTxnCommit)) != 0)
	    return ret;

/* modified by Huilong */
//sleep(2);
/* end Huilong*/

	if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0)
	    return ret;

    printf("[%s] Txn2: getting Data in the database For Key = 9\n", prefixMsg);

    /** Getting  Data in the Database **/
    keyVal = 9;
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    key.data = (int *) &keyVal;
    key.size = sizeof(keyVal);
    data.flags = DB_DBT_MALLOC;

    if ((ret = dbp->get(dbp, tid, &key, &data, 0)) == 0)
        printf("[%s] (key, data) retrieved: (%d, %.*s)\n", prefixMsg, *(int *)key.data, (int)data.size,
	       (char *)data.data);
    else {
        dbp->err(dbp, ret, "DB->get");
        return(-1);
    }
   printf("[%s] Txn2: calling sync() to write buffer to memory \n", prefixMsg);

   memp_sync(dbenv, NULL);

	if (TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

	if((ret = dbp->cursor(dbp, tid, &dbcp, 0)) != 0){
		return -1;
	}

	printf("[%s] Txn2: getting first record by a cursor\n", prefixMsg);
	if((ret = dbcp->c_get(dbcp, &key, &data, DB_NEXT)) == 0){
		printf("[%s] (key, data) retrieved: (%d, %.*s)\n", prefixMsg, *(int *)key.data, (int) data.size, (char *)data.data);
	}
    if (ret != DB_NOTFOUND && ret != 0) {
	dbp->err(dbp, ret, "DBcursor->get");
	return(-1);
    }
    if ((ret = dbcp->c_close(dbcp)) != 0) {
	dbp->err(dbp, ret, "db cursor close");
    }

    printf("[%s] Txn2: calling sync() to write buffer to memory \n", prefixMsg);
	memp_sync(dbenv, NULL);

	if((ret = txn_commit(tid, 0)) != 0){
		return -1;
	}

	printf("[%s] Txn2: committed \n", prefixMsg);

	/* by now, the od_pgcnt should become 1 since the first unstamped page being stamped
	   when reading the first record */
	if (TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}
	tid == NULL;
	if ((ret = dbShutdown(dbp, dbenv, tid, printTTAfterTxnCommit)) != 0)
	    return ret;
    }

    return(0);
}

/* This test follows the previous test (repeated at least twice) to retrieve all
 * the tuples and then stamp each of them, so the od_pgcnt after this operation
 * should become 0 again --
 *        Open database -> txn begines ->
 *        Retrieve all 10 tuples inserted in the previous operation -> stamp each
 *                                  page -> write page -> print TT -> txn commits
 *        Close database
 *
 */
int advTest4(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords, int numVersions, int numRecordsToRead, int printTTAfterTxnCommit, int printTTBforTxnCommit, int printTTBforSync)
{
    DB *dbp;
    DBC *dbcp;
    DB_ENV *dbenv;
    DBT key, data;
    DB_TXN * tid;
    int ret, keyVal, version,i;
    char  strData[200];

    for (version=1; version <= numVersions; version++) {
/* modified by Huilong */
//sleep(2);
/* end Huilong*/
	printf("\n********************  Round %d  ************************\n", version);
	if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0)
	    return ret;
/* modified by Huilong */
printTT(dbenv);
/* end Huilong*/

	printf("[%s] Txn1: retrieve Data from the database \n", prefixMsg);

	for(i=1; i<=10; i++){

		/* printf("[%s] Txn1: getting Data in the database For Key = %d\n", prefixMsg, i);*/

		/** Getting  Data in the Database **/
		keyVal = i;
		memset(&key, 0, sizeof(key));
		memset(&data, 0, sizeof(data));
		key.data = (int *) &keyVal;
		key.size = sizeof(keyVal);
		data.flags = DB_DBT_MALLOC;

		if ((ret = dbp->get(dbp, tid, &key, &data, 0)) == 0)
			printf("[%s] (key, data) retrieved: (%d, %.*s)\n", prefixMsg, *(int *)key.data, (int)data.size,
			   (char *)data.data);
		else {
			dbp->err(dbp, ret, "DB->get");
			return(-1);
		}

    }
	printf("[%s] Txn1: calling sync() to write buffer to memory \n", prefixMsg);

   	memp_sync(dbenv, NULL);

	if((ret = txn_commit(tid, 0)) != 0){
		return -1;
	}

	printf("[%s] Txn1: committed \n", prefixMsg);

	/* by now, the od_pgcnt should become 1 since the first unstamped page being stamped
	   when reading the first record */
	if (TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}
	tid == NULL;

	if ((ret = dbShutdown(dbp, dbenv, tid, printTTAfterTxnCommit)) != 0)
	    return ret;
    }

    return(0);
}

 /* Test on STP checkpoint recovery --
  *     Open database ->
  *     T1 begins -> insert 9 tuples -> split into two pages ->T1 commits
  *     T2 begins -> insert 10th tuple -> stamp page [-> print TT, T1.od_pgcnt=0, T2.od_pgcnt=0]
  *               -> checkpoint1 [-> print TT] -> T2 commits
  *     T3 begins -> get 10th tuple -> stamp page -> checkpoint2 [->print TT] -> T3 commits
  *               -> print TT
  *     Close database
  *
  * Repeating this operation will have different TT results
 */

int advTest5(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords, int numVersions, int numRecordsToRead, int printTTAfterTxnCommit, int printTTBforTxnCommit, int printTTBforSync)
{
    DB *dbp;
    DBC *dbcp;
    DB_ENV *dbenv;
    DBT key, data;
    DB_TXN * tid, *tid2;
    int ret, keyVal, version;
    char  strData[200];


    for (version=1; version <= numVersions; version++) {
/* modified by Huilong */
//sleep(2);
/* end Huilong*/
	printf("\n********************  Round %d  ************************\n", version);
	if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0)
	    return ret;

	printf("[%s] Txn1: inserting Data into the database \n", prefixMsg);

	/* make two pages of records */
	for (keyVal=1; keyVal <= 9; keyVal++) {
	    sprintf(strData, " Computer Science For Key %d Version %d. ", keyVal, version);
	    insertData(dbp, tid, keyVal, strData);
	    /* sleep(1); */
	}

	if((ret = txn_commit(tid, 0)) != 0){
		return -1;
	}
	printf("[%s] Txn1: committed \n", prefixMsg);
    /* Now TT contains unstamped pages in memory, but od_pgcnt is 0 */


	if((ret = txn_begin(dbenv, NULL, &tid2, 0)) != 0){
		return -1;
	}
    printf("[%s] Txn2: inserting Data in the database For Key = 10\n", prefixMsg);

    /** insert the 10th Data in the Database **/
    keyVal = 10;
 	sprintf(strData, " Computer Science For Key %d Version %d. ", keyVal, version);
	insertData(dbp, tid2, keyVal, strData);

	if (TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

	printf("[%s] Txn2: checkpoint #1 \n", prefixMsg);

	if((ret = txn_checkpoint(dbenv, 0, 0, 0)) != 0){
		return -1;
	}

    /* by now, the od_pgcnt should be 2 */
	if (printTTBforTxnCommit == TRUE) {
	    printf("[%s] Printing TT after checkpoint #1 \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}


 	if((ret = txn_commit(tid2, 0)) != 0){
		return -1;
	}

	printf("[%s] Txn2: committed \n", prefixMsg);

	if((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0){
		return -1;
	}
    printf("[%s] Txn3: getting Data in the database For Key = 10\n", prefixMsg);

    /** Getting  Data in the Database **/
    keyVal = 10;
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    key.data = (int *) &keyVal;
    key.size = sizeof(keyVal);
    data.flags = DB_DBT_MALLOC;

    if ((ret = dbp->get(dbp, tid, &key, &data, 0)) == 0)
        printf("[%s] (key, data) retrieved: (%d, %.*s)\n", prefixMsg, *(int *)key.data, (int)data.size,
	       (char *)data.data);
    else {
        dbp->err(dbp, ret, "DB->get");
        return(-1);
    }

	printf("[%s] Txn3: checkpoint #2 \n", prefixMsg);
	if((ret = txn_checkpoint(dbenv, 0, 0, 0)) != 0){
		return -1;
	}

	if (TRUE) {
	    printf("[%s] Printing TT after checkpoint #2 \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}


 	if((ret = txn_commit(tid, 0)) != 0){
		return -1;
	}

	printf("[%s] Txn3: committed \n", prefixMsg);
	if (TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

    printf("[%s] Shutdown the database\n", prefixMsg);
	tid = NULL;
	if ((ret = dbShutdown(dbp, dbenv, tid, printTTAfterTxnCommit)) != 0){
		printf("[%s] shutdown error\n", prefixMsg);
	    return ret;
	}
    }
    return(0);
}


 /* More test od_pgcnt on tuple insertions and retrieving --
  *     Open database ->
  *     T1 begins -> insert 9 tuples -> split into two pages [-> print TT, od_pgcnt=0]
  *               -> T1 commits
  *     T2 begins -> insert 10th tuple [-> print TT, T1.od_pgcnt=0, T2.od_pgcnt=0]
  *               -> T2 commits
  *     T3 begins -> insert 11th tuple [-> print TT, T1.od_pgcnt=0, T2.od_pgcnt=0, T3.od_pgcnt=0]
  *               -> get 1st tuple -> stamp page [-> print TT, od_pgcnt=0 for all three txns (no page write)]
  *               -> T3 commits
  *     Close database
  *
  * Repeating this operation will have different TT results
 */

int advTest6(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords, int numVersions, int numRecordsToRead, int printTTAfterTxnCommit, int printTTBforTxnCommit, int printTTBforSync)
{
    DB *dbp;
    DBC *dbcp;
    DB_ENV *dbenv;
    DBT key, data;
    DB_TXN * tid, *tid2;
    int ret, keyVal, version;
    char  strData[200];

    for (version=1; version <= numVersions; version++) {
/* modified by Huilong */
//sleep(2);
/* end Huilong*/
	printf("\n********************  Round %d  ************************\n", version);
	if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0)
	    return ret;

	printf("[%s] Txn1: inserting Data into the database \n", prefixMsg);

	/* make two pages of records */
	for (keyVal=1; keyVal <= 9; keyVal++) {
	    sprintf(strData, " Computer Science For Key %d Version %d. ", keyVal, version);
	    insertData(dbp, tid, keyVal, strData);
	    /* sleep(1); */
	}

	/* TT entry for this txn contains un_stamped pages, but od_pgcnt should be 0 */
	if (printTTBforSync == TRUE) {
	    printf("[%s] Printing TT after insertion\n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

 	if((ret = txn_commit(tid, 0)) != 0){
		return -1;
	}

	printf("[%s] Txn1 committed\n", prefixMsg);

	printf("[%s] Txn2: inserting Data in the database For Key = 10\n", prefixMsg);
	if((ret = txn_begin(dbenv, NULL, &tid2, 0)) != 0){
		return -1;
	}

    /** insert the 10th Data into the Database **/
    keyVal = 10;
 	sprintf(strData, " Computer Science For Key %d Version %d. ", keyVal, version);
	insertData(dbp, tid2, keyVal, strData);

    /* od_pgcnt for the txn should still be 0 */
	if (TRUE) {
	    printf("[%s] Printing TT after insertion (before txn2 commit)\n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}


   /* od_pgcnt for the previous txn should stay 0, and od_pgcnt for the current txn is 0
      because no memory write
    */

 	if((ret = txn_commit(tid2, 0)) != 0){
		return -1;
	}
	printf("[%s] Txn2 committed\n", prefixMsg);

	printf("[%s] Txn3: inserting Data in the database For Key = 11\n", prefixMsg);
	if((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0){
		return -1;
	}

    /** insert the 11st Data in the Database **/
    keyVal = 11;
 	sprintf(strData, " Computer Science For Key %d Version %d. ", keyVal, version);
	insertData(dbp, tid, keyVal, strData);

	if (TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

    printf("[%s] Txn3: getting Data in the database For Key = 1\n", prefixMsg);

    /** Getting  Data in the Database -- cause the page containing the data to be stamped **/
    keyVal = 1;
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    key.data = (int *) &keyVal;
    key.size = sizeof(keyVal);
    data.flags = DB_DBT_MALLOC;

    if ((ret = dbp->get(dbp, tid, &key, &data, 0)) == 0)
        printf("[%s] (key, data) retrieved: (%d, %.*s)\n", prefixMsg, *(int *)key.data, (int)data.size,
	       (char *)data.data);
    else {
        dbp->err(dbp, ret, "DB->get");
        return(-1);
    }


 	if (TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

 	if((ret = txn_commit(tid, 0)) != 0){
		return -1;
	}
	printf("[%s] Txn3 committed\n", prefixMsg);

//memp_sync(dbenv, NULL); //added by Huilong

    printf("[%s] Shutdown the database\n", prefixMsg);
	tid = NULL;
	if ((ret = dbShutdown(dbp, dbenv, tid, printTTAfterTxnCommit)) != 0){
		printf("[%s] shutdown error\n", prefixMsg);
	    return ret;
	}
    }
    return(0);
}


 /* More test od_pgcnt on tuple insertions and retrieving --
  *     Open database ->
  *     T1 begins -> insert 10 tuples -> split into two pages [-> print TT, od_pgcnt=0]
  *               -> T1 commits
  *     T2 begins -> insert another 10 tuples [-> print TT, T1.od_pgcnt=0, T2.od_pgcnt=0]
  *               -> T2 commits
  *     T3 begins -> insert 21st tuple [-> print TT, T1.od_pgcnt=0, T2.od_pgcnt=0, T3.od_pgcnt=0]
  *               -> get 11th tuple -> stamp page -> page write [-> print TT, T3.od_pgcnt=1]
  *               -> T3 commits
  *     Close database
  *
  * Repeating this operation will have different TT results
 */

int advTest7(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords, int numVersions, int numRecordsToRead, int printTTAfterTxnCommit, int printTTBforTxnCommit, int printTTBforSync)
{
    DB *dbp;
    DBC *dbcp;
    DB_ENV *dbenv;
    DBT key, data;
    DB_TXN * tid, *tid2;
    int ret, keyVal, version;
    char  strData[200];

    for (version=1; version <= numVersions; version++) {
/* modified by Huilong */
//sleep(2);
/* end Huilong*/
	printf("\n********************  Round %d  ************************\n", version);
	if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0)
	    return ret;

	printf("[%s] Txn1: inserting Data into the database \n", prefixMsg);

	/* make two pages of records */
	for (keyVal=1; keyVal <= 10; keyVal++) {
	    sprintf(strData, " Computer Science For Key %d Version %d. ", keyVal, version);
	    insertData(dbp, tid, keyVal, strData);
	    /* sleep(1); */
	}

	/* TT entry for this txn contains un_stamped pages, but od_pgcnt should be 0 */
	if (printTTBforSync == TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

 	if((ret = txn_commit(tid, 0)) != 0){
		return -1;
	}

	printf("[%s] Txn1 committed\n", prefixMsg);

	printf("[%s] Txn2: inserting Data in the database\n", prefixMsg);
	if((ret = txn_begin(dbenv, NULL, &tid2, 0)) != 0){
		return -1;
	}

    /** insert the another 10 Data into the Database **/
	for (keyVal=11; keyVal <= 20; keyVal++) {
	    sprintf(strData, " Computer Science For Key %d Version %d. ", keyVal, version);
	    insertData(dbp, tid, keyVal, strData);
	    /* sleep(1); */
	}

    /* od_pgcnt for the txn should still be 0 */
	if (TRUE) {
	    printf("[%s] Printing TT \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}


   /* od_pgcnt for the previous txn should stay 0, and od_pgcnt for the current txn is 0
      because no memory write
    */

 	if((ret = txn_commit(tid2, 0)) != 0){
		return -1;
	}
	printf("[%s] Txn2 committed\n", prefixMsg);

	printf("[%s] Txn3: inserting Data in the database For Key = 21\n", prefixMsg);
	if((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0){
		return -1;
	}

    /** insert the 11st Data in the Database **/
    keyVal = 21;
 	sprintf(strData, " Computer Science For Key %d Version %d. ", keyVal, version);
	insertData(dbp, tid, keyVal, strData);

	if (TRUE) {
	    printf("[%s] Printing TT after after insertion\n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

    printf("[%s] Txn3: getting Data in the database For Key = 11\n", prefixMsg);

    /** Getting  Data in the Database -- cause the page containing the data to be stamped **/
    keyVal = 11;
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    key.data = (int *) &keyVal;
    key.size = sizeof(keyVal);
    data.flags = DB_DBT_MALLOC;

    if ((ret = dbp->get(dbp, tid, &key, &data, 0)) == 0)
        printf("[%s] (key, data) retrieved: (%d, %.*s)\n", prefixMsg, *(int *)key.data, (int)data.size,
	       (char *)data.data);
    else {
        dbp->err(dbp, ret, "DB->get");
        return(-1);
    }

	printf("[%s] Calling sync() to write buffer to memory \n", prefixMsg);

    memp_sync(dbenv, NULL);

 	if (TRUE) {
	    printf("[%s] Printing TT after retrieval and memory write \n", prefixMsg);
	    ret = printTT(dbenv);
	    if (ret != 0)
		return (-1);
	}

 	if((ret = txn_commit(tid, 0)) != 0){
		return -1;
	}
	printf("[%s] Txn3 committed\n", prefixMsg);

    printf("[%s] Shutdown the database\n", prefixMsg);
	tid = NULL;
	if ((ret = dbShutdown(dbp, dbenv, tid, printTTAfterTxnCommit)) != 0){
		printf("[%s] shutdown error\n", prefixMsg);
	    return ret;
	}
    }
    return(0);


}


int displayDB(char * dbName, int openFlag, int numRecords) {

    DB *dbp;
    DBC *dbcp;
    DBT key, data;
    int ret;
    DB_ENV *dbenv;
    DB_TXN * tid;

    if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0)
	return ret;

    printf("[%s] Retrieving the content using a cursor (Sequenced)\n", prefixMsg);

    /* Acquire a cursor for the database. */
    if ((ret = dbp->cursor(dbp, tid, &dbcp, 0)) != 0) {
	dbp->err(dbp, ret, "DB->cursor");
	return(-1);
    }

    /* Initialize the key/data return pair. */
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    /* Walk through the database and print out the key/data pairs. */
    while ((ret = dbcp->c_get(dbcp, &key, &data, DB_NEXT)) == 0 && numRecords-- > 0) {
	printf("[%s] (%d, %.*s)\n", prefixMsg, *(int *)key.data, (int)data.size, (char *)data.data);
    }

    if (ret != DB_NOTFOUND && ret != 0) {
	dbp->err(dbp, ret, "DBcursor->get");
	return(-1);
    }
    if ((ret = dbcp->c_close(dbcp)) != 0) {
	dbp->err(dbp, ret, "db cursor close");
    }

        if((ret = txn_commit(tid, 0)) != 0){
                return -1;
        }

    if ((ret = dbShutdown(dbp, dbenv, tid, FALSE)) != 0)
	return ret;
    return (0);
}


int insertData(DB *dbp, DB_TXN *tid, int keyVal, char * strData) {

    DBT key, data;
    int ret;

    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    key.data = (int *) &keyVal;
    key.size = sizeof(keyVal);
    data.data = strData;
    /*data.size = strlen(strData)+1;*/
    data.size = 4000;

    if ((ret = dbp->put(dbp, tid, &key, &data, 0)) == 0)
	printf("[%s] (key, data) stored: (%d, %s) \n", prefixMsg, *(int *)key.data, (char *)data.data);
    else {
	printf("[%s] ERROR in insertData()\n", prefixMsg);
 	exit(-1);
    }

    return(0);
}

int dbStartup(char *dbName, DB **dbpPtr, DB_ENV **dbenvPtr, DB_TXN ** tidPtr, int openFlag) {

    int ret;
    DB_ENV *dbenv;
    DB * dbp;

    /*Create the environment handle. */
    if ((ret = db_env_create(dbenvPtr, 0)) != 0) {
	printf("[%s] dbStartup: db_env_create: %s\n", prefixMsg, db_strerror(ret));
	return (-1);
    }

    dbenv = *dbenvPtr;
    /* Opening the Environment and setting its options */
    if ((ret = dbenv->open(dbenv, ENV_DIRECTORY,
			   DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG |
			   DB_INIT_MPOOL | DB_INIT_TXN | DB_RECOVER | DB_THREAD,
			   0664)) != 0) {
	printf("[%s] dbStartup: DB_ENV->open: %s\n", prefixMsg, db_strerror(ret));
	printf("********1\n");
	(void)dbenv->close(dbenv, 0); printf("********2\n");
	//dbenv->err(dbenv, ret, "dbenv->open: %s", ENV_DIRECTORY);printf("********3\n");
	return(-1);
    }

    /* Creating a db handle under the dbenvironment */
    if ((ret = db_create(dbpPtr, dbenv, 0)) != 0) {
	printf("[%s] dbStartup: db_create: %s\n", prefixMsg, db_strerror(ret));
	return(-1);
    }

    dbp = *dbpPtr;

    if ((ret = dbp->open(dbp, dbName, NULL, DB_BTREE, openFlag, 0664)) != 0) {
	printf("[%s] dbStartup: DB->open: %s\n", prefixMsg, db_strerror(ret));
	dbp->err(dbp, ret, "db->open: %s", dbName);
	return(-1);
    }

    if (tidPtr != NULL) {
	/* Begin the transaction. */
	if ((ret = txn_begin(dbenv, NULL, tidPtr, 0)) != 0) {
	    dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
	    return (-1);
	}
    }
    printf("[%s] DATABASE OPENED\n", prefixMsg);

    return (0);
}


int dbShutdown(DB *dbp, DB_ENV *dbenv, DB_TXN * tid, int printTT) {

    int ret;

/*    if (tid != NULL) {
	if ((ret = txn_commit(tid,0)) != 0) {
	    dbenv->err(dbenv, ret, "DB_TXN->commit");
	    return (-1);
	}

    }*/

    if ((ret = dbp->close(dbp, 0)) != 0) {
	dbp->err(dbp, ret, "db close");
	return(-1);
    }

    if ((ret = dbenv->close(dbenv,0)) != 0) {
	dbenv->err(dbenv, ret, "db env close");
	return(-1);
    }

    /* sleep(2);*/
    printf("[%s] DATABASE CLOSED \n", prefixMsg);
    return (0);
}


