#include <sys/types.h>
#include <stdio.h>
#include <time.h>
#include <sys/times.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "db.h"

/* for debug (Shilong) */
const char prefixMsg[] = "suit01";

# define ENV_DIRECTORY "./txnTemp/"
# define TT_DEBUG_FILENAME1 "TTDebugFile1"
# define TT_DEBUG_FILENAME2 "TTDebugFile2"
# define MAX_NUM_RECORDS 1000
# define FALSE 0
# define TRUE 1
# define NUM_WORKERS 2
# define TT_DEBUG 1

typedef struct dbInformation {
    DB *dbp;
    DB_ENV *dbenv;
    int id;
    int numRecords;
} dbInformation;


int simpleTest(char * dbName, char * dbNameWithEnv, int openFlag);
int advTest(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords,
	    int numVersions, int numRecordsToRead,
	    int printTTAfterTxnCommit, int printTTBforTxnCommit, int printTTBforSync);
int basicCrashTest(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords);
int advCrashTest(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords);
int basicTxnTimeTest(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords);

int displayDB(char * dbName, int openFlag, int numRecords);
int displayNonSeqDB(char * dbName, int openFlag, int numRecords,
		    struct timeval startupTime, struct timeval shutdownTime,
		    int printTTBforSync, int printTTBforTxnCommit);

int dbStartup(char *dbName, DB **dbpPtr, DB_ENV **dbenvPtr, DB_TXN ** tidPtr, int openFlag);
int dbShutdown(DB *dbp, DB_ENV *dbenv, DB_TXN * tid, int printTT);
int insertData(DB *dbp, DB_TXN *tid, int keyVal, char * strData);

void *workers(void *);
int dbReader(dbInformation dbInfo);
int dbWriter(dbInformation dbInfo);

/* TODO : If TT_DEBUG MODE then make calls to printTT */
/* To remove calls to printTT when forming the release test case */
int printTT(DB_ENV *dbenv);

/* For performing diff, time related information must be made as OK or NOT OK */
/* Code for cleanup of the direcoty */

int main (int argc, char *argv[]) {

    int retVal, numRecords, numVersions, numRecordsToRead,
 	openFlag = DB_CREATE|DB_TEMPORAL;

    if ((argc == 2) && (strcmp(argv[1], "partition") == 0))
	openFlag |= DB_PARTITION;

    /* Test 1 */
    printf("[%s]Test 1 --@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n", prefixMsg);
    retVal = simpleTest("bdbTest1.db", "./txnTemp/bdbTest1.db", openFlag);
    
    return 0;
}

/* Working Seq of this test function :
	Open (Env, DB, Txn) -> 	Insert data -> 	Close (Env, DB, Txn) -> Display Data Seq
	Open (Env, DB, Txn) -> 	Update data -> 	Close (Env, DB, Txn) -> Display Data Seq, Non Seq
	Open (Env, DB, Txn) -> 	Delete data -> 	Close (Env, DB, Txn) -> Display Data Seq, Non Seq
*/

int simpleTest(char * dbName, char * dbNameWithEnv, int openFlag)
{
    DB *dbp;
    DBC *dbcp;
    DBT key, data;
    DB_ENV *dbenv;
    DB_TXN * tid;
    int ret, keyVal;
    struct timeval startupTime, shutdownTime;
    char strData[100];


    /**************** DATA INSERTION PHASE *****************/
    printf("[%s] PASS-1: Insert data into the database\n", prefixMsg);

    if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0)
	return ret;

    keyVal = 0;

    keyVal++;
    strcpy(strData, "Computer Science");
    insertData(dbp, tid, keyVal, strData);

    keyVal++;
    strcpy(strData, "Electronics and Communication Engg");
    insertData(dbp, tid, keyVal, strData);

    keyVal++;
    strcpy(strData, "Physics");
    insertData(dbp, tid, keyVal, strData);

    if ((ret = dbShutdown(dbp, dbenv, tid, FALSE)) != 0)
	return ret;
    if ((ret = displayDB(dbName, openFlag, MAX_NUM_RECORDS)) != 0)
	return (-1);

    /**************** DATA UPDATION PHASE *****************/
    printf("[%s] PASS-2: Update Data in the database \n", prefixMsg);

    ret = env_current(&startupTime, 7);
    if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0)
	return ret;

    keyVal = 0;

    keyVal++;
    strcpy(strData, "Computer Science - Changed");
    insertData(dbp, tid, keyVal, strData);

    keyVal++;
    strcpy(strData, "Electronics and Communication Engg - Changed");
    insertData(dbp, tid, keyVal, strData);

    keyVal++;
    strcpy(strData, "Physics - Changed");
    insertData(dbp, tid, keyVal, strData);

    keyVal = 0;

    keyVal++;
    strcpy(strData, "Computer Science - Changed Again");
    insertData(dbp, tid, keyVal, strData);

    if ((ret = dbShutdown(dbp, dbenv, tid, FALSE)) != 0)
	return ret;
    ret = env_current(&shutdownTime, 7);

    if ((ret = displayNonSeqDB(dbName, openFlag, MAX_NUM_RECORDS,
			       startupTime, shutdownTime, FALSE, FALSE)) != 0)
	return (-1);

    printf("[%s] Sleeping... \n", prefixMsg);
    sleep(1);

    if ((ret = displayDB(dbName,openFlag, MAX_NUM_RECORDS)) != 0)
	return (-1);

    /**************** DATA DELETION PHASE *****************/
    printf("[%s] PASS-3: Delete Data in the database \n", prefixMsg);

    ret = env_current(&startupTime, 7);
    if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0)
	return ret;

    keyVal = 2;
    memset(&key, 0, sizeof(key));
    key.data = (int *) &keyVal;
    key.size = sizeof(keyVal);

    if ((ret = dbp->del(dbp, tid, &key, 0)) == 0)
	printf("[%s] %d: key was deleted.\n", prefixMsg,  *(int *)key.data);
    else {
	dbp->err(dbp, ret, "DB->del");
	return(-1);
    }

    keyVal = 3;
    memset(&key, 0, sizeof(key));
    key.data = (int *) &keyVal;
    key.size = sizeof(keyVal);

    if ((ret = dbp->del(dbp, tid, &key, 0)) == 0)
	printf("[%s] %d: key was deleted.\n", prefixMsg, *(int *)key.data);
    else {
	dbp->err(dbp, ret, "DB->del");
	return(-1);
    }

    if ((ret = dbShutdown(dbp, dbenv, tid, FALSE)) != 0) return ret;
    ret = env_current(&shutdownTime, 7);

    if ((ret = displayNonSeqDB(dbName,openFlag, MAX_NUM_RECORDS,
			       startupTime, shutdownTime, FALSE, FALSE)) != 0)
	return (-1);

    printf("[%s] Sleeping... \n", prefixMsg);
    sleep(1);

    if ((ret = displayDB(dbName,openFlag, MAX_NUM_RECORDS)) != 0)
	return (-1);

    /********************* DATA GET PHASE, NON-CURSOR *******************/
    if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0)
	return  ret;
    printf("[%s] Getting Data in the database For Key = 1\n", prefixMsg);

    /** Getting  Data in the Database **/
    keyVal = 1;
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    key.data = (int *) &keyVal;
    key.size = sizeof(keyVal);
    data.flags = DB_DBT_MALLOC;

    if ((ret = dbp->get(dbp, tid, &key, &data, 0)) == 0)
        printf("[%s] %d : %.*s\n", prefixMsg, *(int *)key.data, (int)data.size, 
	       (char *)data.data);
    else {
        dbp->err(dbp, ret, "DB->get");
        return(-1);
    }
    if ((ret = dbShutdown(dbp, dbenv, tid, FALSE)) != 0)
	return ret;                                   
    /****************************************************************************/

    return(0);
}

int displayDB(char * dbName, int openFlag, int numRecords) {

    DB *dbp;
    DBC *dbcp;
    DBT key, data;
    int ret;
    DB_ENV *dbenv;
    DB_TXN * tid;

    if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0)
	return ret;

    printf("[%s] Retrieving the content using a cursor (Sequenced)\n", prefixMsg);

    /* Acquire a cursor for the database. */
    if ((ret = dbp->cursor(dbp, tid, &dbcp, 0)) != 0) {
	dbp->err(dbp, ret, "DB->cursor");
	return(-1);
    }

    /* Initialize the key/data return pair. */
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    /* Walk through the database and print out the key/data pairs. */
    while ((ret = dbcp->c_get(dbcp, &key, &data, DB_NEXT)) == 0 && numRecords-- > 0) {
	printf("[%s] (%d, %.*s)\n", prefixMsg, *(int *)key.data, (int)data.size, (char *)data.data);
    }

    if (ret != DB_NOTFOUND && ret != 0) {
	dbp->err(dbp, ret, "DBcursor->get");
	return(-1);
    }
    if ((ret = dbcp->c_close(dbcp)) != 0) {
	dbp->err(dbp, ret, "db cursor close");
    }

    if ((ret = dbShutdown(dbp, dbenv, tid, FALSE)) != 0)
	return ret;
    return (0);
}

int displayNonSeqDB(char * dbName, int openFlag, int numRecords,
		    struct timeval startupTime, struct timeval shutdownTime,
		    int printTTBforSync, int printTTBforTxnCommit) {

    DB *dbp;
    DBC *dbcp;
    DBT key, data;
    int ret;
    struct timeval start, stop, beginning, forever, prevStop, prevStart;
    DB_ENV *dbenv;
    DB_TXN * tid;
    char startTimeMsg[800], stopTimeMsg[800];

    if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0)
	return ret;

    printf("[%s] Retrieving the content using a cursor (Non Sequenced)\n", prefixMsg);

    /* Acquire a cursor for the database. */
    if ((ret = dbp->cursor(dbp, tid, &dbcp, 0)) != 0) {
	dbp->err(dbp, ret, "DB->cursor");
	return(-1);
    }

    /* Initialize the key/data return pair. */
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    /* int DBcursor->c_get_nonsequenced(DBC *cursor, DBT *key, DBT *data, struct timeval start,
     * struct timeval stop, u_int32_t flags, struct timeval *outstart, struct timeval *outstop);
     */

    /* Walk through the database and print out the key/data pairs. */
    beginning.tv_sec = BEGINNING;
    beginning.tv_usec = 0;
    forever.tv_sec = FOREVER;
    forever.tv_usec = 0;
    prevStop.tv_sec = 0;
    prevStop.tv_usec = 0;
    prevStart.tv_sec = 0;
    prevStart.tv_usec = 0;

    while ((ret = dbcp->c_get_nonsequenced(dbcp, &key, &data, beginning, forever,
					  DB_NEXT, &start, &stop)) == 0 && numRecords-- > 0) {

	if ( (start.tv_sec >= startupTime.tv_sec ||
	      (start.tv_sec == startupTime.tv_sec && start.tv_usec >= startupTime.tv_usec)) &&
	     (start.tv_sec <= shutdownTime.tv_sec ||
	      (start.tv_sec == shutdownTime.tv_sec && start.tv_usec <= shutdownTime.tv_usec)) )
	    strcpy(startTimeMsg, "START Between dbStartup and dbShutdown");
	else
	    strcpy(startTimeMsg, "START NOT Between startup and shutdown");

	if (start.tv_sec == prevStop.tv_sec && start.tv_usec == prevStop.tv_usec)
	    strcat(startTimeMsg, ", START Equal to STOP of Previous Record");
	else
	    strcat(startTimeMsg, ", START NOT Equal to STOP of Previous Record");

	if (start.tv_sec == prevStart.tv_sec && start.tv_usec == prevStart.tv_usec)
	    strcat(startTimeMsg, ", START Equal to START of Previous Record");
	else
	    strcat(startTimeMsg, ", START NOT Equal to START of Previous Record");

	if (stop.tv_sec == FOREVER)
	    strcpy(stopTimeMsg, "STOP Equal to FOREVER");
	else {
	    if ((stop.tv_sec >= startupTime.tv_sec ||
		 (stop.tv_sec == startupTime.tv_sec && stop.tv_usec >= startupTime.tv_usec)) &&
		(stop.tv_sec <= shutdownTime.tv_sec ||
		 (stop.tv_sec == shutdownTime.tv_sec && stop.tv_usec <= shutdownTime.tv_usec)))
		strcpy(stopTimeMsg, "STOP Between startup and shutdown");
	    else
		strcpy(stopTimeMsg, "STOP NOT Between startup and shutdown");
	}

	prevStop = stop;
	prevStart = start;

	/* Todo Comment out the actual time display in Release Test Suite */
	printf("[%s] (%d, %.*s) start-time: (((%d))), stop-time: (((%d)))\n",
	       prefixMsg, *(int *)key.data, (int)data.size, (char *)data.data, start.tv_sec, stop.tv_sec);

	printf("[%s] (%d, %.*s) start: %s, stop: %s\n",
	       prefixMsg, *(int *)key.data, (int)data.size, (char *)data.data, startTimeMsg, stopTimeMsg);
    }

    if (ret != DB_NOTFOUND && ret != 0) {
	dbp->err(dbp, ret, "DBcursor->get");
	return(-1);
    }
    if ((ret = dbcp->c_close(dbcp)) != 0) {
	dbp->err(dbp, ret, "db cursor close");
    }

    /* Doubt : YAO : Why is a sync needed, while read has been done.. */
    if ((ret = dbShutdown(dbp, dbenv, tid, FALSE)) != 0) return ret;
    return (0);
}

int insertData(DB *dbp, DB_TXN *tid, int keyVal, char * strData) {

    DBT key, data;
    int ret;

    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    key.data = (int *) &keyVal;
    key.size = sizeof(keyVal);
    data.data = strData;
    data.size = strlen(strData)+1;

    if ((ret = dbp->put(dbp, tid, &key, &data, 0)) == 0)
	printf("[%s] (key, data) stored: (%d, %s) \n", prefixMsg, *(int *)key.data, (char *)data.data);
    else {
	printf("[%s] ERROR in insertData()\n", prefixMsg);
	dbp->err(dbp, ret, "DB->put");
	/* Modify !!!!!!!!!!!!!!!!!!!!!!!*/
	exit(-1);
    }

    return(0);
}

int dbStartup(char *dbName, DB **dbpPtr, DB_ENV **dbenvPtr, DB_TXN ** tidPtr, int openFlag) {

    int ret;
    DB_ENV *dbenv;
    DB * dbp;

    /*Create the environment handle. */
    if ((ret = db_env_create(dbenvPtr, 0)) != 0) {
	printf("[%s] dbStartup: db_env_create: %s\n", prefixMsg, db_strerror(ret));
	return (-1);
    }

    dbenv = *dbenvPtr;
    /* Opening the Environment and setting its options */
    if ((ret = dbenv->open(dbenv, ENV_DIRECTORY,
			   DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG |
			   DB_INIT_MPOOL | DB_INIT_TXN | DB_RECOVER | DB_THREAD,
			   0664)) != 0) {
	printf("[%s] dbStartup: DB_ENV->open: %s\n", prefixMsg, db_strerror(ret));
	(void)dbenv->close(dbenv, 0);
	dbenv->err(dbenv, ret, "dbenv->open: %s", ENV_DIRECTORY);
	return(-1);
    }

    /* Creating a db handle under the dbenvironment */
    if ((ret = db_create(dbpPtr, dbenv, 0)) != 0) {
	printf("[%s] dbStartup: db_create: %s\n", prefixMsg, db_strerror(ret));
	return(-1);
    }

    dbp = *dbpPtr;

    if ((ret = dbp->open(dbp, dbName, NULL, DB_BTREE, openFlag, 0664)) != 0) {
	printf("[%s] dbStartup: DB->open: %s\n", prefixMsg, db_strerror(ret));
	dbp->err(dbp, ret, "db->open: %s", dbName);
	return(-1);
    }

    if (tidPtr != NULL) {
	/* Begin the transaction. */
	if ((ret = txn_begin(dbenv, NULL, tidPtr, 0)) != 0) {
	    dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
	    return (-1);
	}
    }
    printf("[%s] DATABASE OPENED\n", prefixMsg);

    return (0);
}

int dbShutdown(DB *dbp, DB_ENV *dbenv, DB_TXN * tid, int printTT) {

    int ret;

    if (tid != NULL) {
	if ((ret = txn_commit(tid,0)) != 0) {
	    dbenv->err(dbenv, ret, "DB_TXN->commit");
	    return (-1);
	}
	/* TODO :
	   Check for printTT flag and call Yao's function, also give a msg that this is happ after txn commit
	   [print the TT, the on-disk page count should be 2, the in-memory list should contain 2 pages] */
    }

    if ((ret = dbp->close(dbp, 0)) != 0) {
	dbp->err(dbp, ret, "db close");
	return(-1);
    }

    if ((ret = dbenv->close(dbenv,0)) != 0) {
	dbenv->err(dbenv, ret, "db env close");
	return(-1);
    }

    sleep(2); /* Shilong */
    printf("[%s] DATABASE CLOSED \n", prefixMsg);
    return (0);
}


/*
  Calling sync() says prototype mismatch..
  Check if in normal BDB data can be updated, only if we close the DB and then open it..
  Doesnt look like default DB is opened in Unique Key Mode, find out how to do that..

  Try to solve the makefile (cp to maintain synch issue, in the bdb/build_unix directory)
  Try to solve the libdb.so and libdb.a ????

  BDB:

  Temporal Flag in db env open, check for prerequisit flags (txn, log..)
  db open , add extra check to see if env passed has temporal flag on..
  add extra check to see if opening a temporal db, if yes, then make sure ENV is there.. otherwise err
  Sleep issue, closed and open interval in the txn time..
  Change DOC accordingly
*/

