#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "db.h"

#define	ENV_DIRECTORY	"txnTemp"
#define MAXIMUM_RETRY   10000	

void  env_dir_create();
void  env_open();

void *workers(void *);

DB_ENV *dbenv;
DB *db;
int openFlag = DB_CREATE|DB_THREAD|DB_TEMPORAL;

void usage ()
{
  printf ("cmd **\n");
  exit (1);
}

void
db_open(char *name, int dups)
{
  int ret;

  /* Create the database handle. */
  if ((ret = db_create(&db, dbenv, 0)) != 0) {
    dbenv->err(dbenv, ret, "db_create");
    exit (1);
  }


  /*
   * Open a database in the environment:
   *	create if it doesn't exist
   *	free-threaded handle
   *	read/write owner only
   */
	
  if ((ret = db->open(db, name, NULL,
		      DB_BTREE, openFlag, S_IRUSR | S_IWUSR)) != 0) {
    dbenv->err(dbenv, ret, "db->open: %s", name);
    exit (1);
  }	

}

int numWorkers;


int
main(int argc, char *argv[])
{
  extern char *optarg;
  extern int optind;

  struct timeval mytime;
  pthread_t ptid;
  int ch, i;
	
  /* Thread */
  pthread_t pid[2];
  pthread_attr_t attr;

  if ((argc == 2) && (strcmp(argv[1], "partition") == 0))                                                      
      openFlag |= DB_PARTITION;

  numWorkers=2;
  env_current(&mytime,7);

  pthread_attr_init(&attr);
  pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM); 

	

  /* Thread */


  printf("********The current time is (((%ld,%ld)))*********\n",mytime.tv_sec, mytime.tv_usec);
	
  env_dir_create();
  env_open();

  /* Open database: Key is fruit class; Data is specific type. */
  db_open("fruit", 0);



  /* create workers processes */
  for ( i=0; i<numWorkers; i++) {
    pthread_create(&pid[i], &attr, workers, (void *)i);
  }

  for ( i=0; i<numWorkers; i++) {
    pthread_join(pid[i], NULL);
  }

  return (0);
}

void
env_dir_create()
{
  struct stat sb;

  /*
   * If the directory exists, we're done.  We do not further check
   * the type of the file, DB will fail appropriately if it's the
   * wrong type.
   */
  if (stat(ENV_DIRECTORY, &sb) == 0)
    return;

  /* Create the directory, read/write/access owner only. */
  if (mkdir(ENV_DIRECTORY, S_IRWXU) != 0) {
    fprintf(stderr,
	    "txnapp: mkdir: %s: %s\n", ENV_DIRECTORY, strerror(errno));
    exit (1);
  }
}

void
env_open()
{
  int ret;

  /* Create the environment handle. */
  if ((ret = db_env_create(&dbenv, 0)) != 0) {
    fprintf(stderr,
	    "txnapp: db_env_create: %s\n", db_strerror(ret));
    exit (1);
  }

  /* Set up error handling. */
  dbenv->set_errpfx(dbenv, "txnapp");

  /*
   * Open a transactional environment:
   *	create if it doesn't exist
   *	free-threaded handle
   *	run recovery
   *	read/write owner only
   */
  if ((ret = dbenv->open(dbenv, ENV_DIRECTORY,
			 DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG |
			 DB_INIT_MPOOL | DB_INIT_TXN | DB_RECOVER | DB_THREAD,
			 S_IRUSR | S_IWUSR)) != 0) {
    dbenv->err(dbenv, ret, "dbenv->open: %s", ENV_DIRECTORY);
    exit (1);
  }

}

void *workers(void *arg) {
  int myid;
  
  myid = (int) arg;

  if (myid == 0 ) 
    read_fruit("apple","",0,5,1);
  else 
    add_fruit("apple", "yellow delicious",2,0,2);
}
    
  

int
add_fruit(char *fruit, char *name,int time,int time2,int myid)
{
  DBT key, data;
	
  DB_TXN *tid;
	
  int ret, t_ret;
  struct timeval mytime;

  /* Initialization. */
  memset(&key, 0, sizeof(key));
  memset(&data, 0, sizeof(data));
  key.data = fruit;
  key.size = strlen(fruit);
  data.data = name;
  data.size = strlen(name);
  data.flags = DB_DBT_MALLOC;

  sleep(time);  /* await for the time to start */
  printf("Enter add fruit\n");

  /* Begin the transaction. */
  if ((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0) {
    dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
    exit (1);
  }
  env_current(&mytime,7);
  printf("Current time is (((%ld,%ld))): ------------------------------------->",mytime.tv_sec,mytime.tv_usec);
  printf("Start transaction %d() \n",myid);
	

  sleep(time2);

  env_current(&mytime,7);
  printf("Current time is (((%ld,%ld))): ------------------------------------->",mytime.tv_sec,mytime.tv_usec);
  printf("Put Operation in transaction %d() \n",myid);

  switch (ret = db->put(db, tid, &key, &data, 0)) {

  case 0:
    env_current(&mytime,7); 
    printf("Current time is (((%ld,%ld))): ------------------------------------->",mytime.tv_sec, mytime.tv_usec);
    printf("Commit transaction %d() \n",myid);

    if ((ret = txn_commit(tid, 0)) != 0) {
      dbenv->err(dbenv, ret, "DB_TXN->commit");
      exit (1);
    }
    memp_sync( dbenv, NULL);
    printf("DEBUG: finished memp_sync\n");
    /*ret = db->get(db,NULL, &key, &data,0);
      dbenv->err(dbenv, ret, "DB->get");*/
    break;

  case DB_LOCK_DEADLOCK:
  default:
    env_current(&mytime,7);
    printf("Current time is (((%ld,%ld))): ------------------------------------->",mytime.tv_sec, mytime.tv_usec);

    printf("Abort this transaction %d() \n",myid);		
		
    if ((t_ret = txn_abort(tid)) != 0) {
      dbenv->err(dbenv, t_ret, "DB_TXN->abort");
      exit (1);
    }
		
    break;
  }
}

int
read_fruit(char *fruit, char *name,int time,int time2,int myid)
{
  DBT key, data;
	
  DB_TXN *tid;
	
  int ret, t_ret;
  char *contents;
  struct timeval mytime;
	
  /* Initialization. */	
  memset(&key, 0, sizeof(key));
  memset(&data, 0, sizeof(data));
  key.data = fruit;
  key.size = strlen(fruit);
  data.flags = DB_DBT_MALLOC;

  sleep(time);

  /* Begin the transaction. */
	
  if ((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0) {
    dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
    exit (1);
  }

  env_current(&mytime,7);
  printf("Current time is (((%ld,%ld))): --->",mytime.tv_sec, mytime.tv_usec);
  printf("Start transaction %d() \n",myid);
	
	

  sleep(time2);

  env_current(&mytime,7);
  printf("Current time is (((%ld,%ld))): ---->",mytime.tv_sec, mytime.tv_usec);
  printf("Get Operation in transaction %d() \n",myid);
  ret = db->get(db, tid, &key, &data, 0);
  printf("NNNNNNNNNNNNNNNNNNNNNNNNNNthe ret is %d\n",ret);
  dbenv->err(dbenv, ret, "db->get");
  switch (ret) {
  case 0:
    printf("<DATA: (%.*s)> \n",(int)data.size,(char *)data.data);
    env_current(&mytime,7);
    printf("Current time is (((%ld,%ld))): --->",mytime.tv_sec, mytime.tv_usec);
    printf("Commit transaction %d() \n",myid);
		
    if ((ret = txn_commit(tid, 0)) != 0) {
      dbenv->err(dbenv, ret, "DB_TXN->commit");
      exit (1);
    } else
	printf("[check_msg]Commit successfully\n");
		
		
    break;

  case DB_LOCK_DEADLOCK:
  default:
    env_current(&mytime,7);
    printf("FAILED TO READ THE DATA\n");
    printf("Current time is (((%ld,%ld))): --->",mytime.tv_sec, mytime.tv_usec);
    printf("Abort this transaction %d() \n",myid);
		
    if ((t_ret = txn_abort(tid)) != 0) {
      dbenv->err(dbenv, t_ret, "DB_TXN->abort");
      exit (1);
    }
		
    break;
  }
}
