#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "db.h"

#define	ENV_DIRECTORY	"txnTemp"
#define MAXIMUM_RETRY   10000	

void  env_dir_create();
void  env_open();

DB_ENV *dbenv;
DB *db;
int openFlag = 0;

void usage ()
{
    printf ("cmd **\n");
    exit (1);
}

void
db_open(char *name, int dups, int join)
{
    int ret;
    u_int32_t flags, flags1, flags0;

    flags0 = DB_CREATE|DB_TEMPORAL;
    flags1 = DB_TEMPORAL;

    if (join == 1)
	flags = flags1 | openFlag;
    else
	flags = flags0 | openFlag;;

    if ((ret = db_create(&db, dbenv, 0)) != 0) {
	dbenv->err(dbenv, ret, "db_create");
	exit (1);
    }

    if ((ret = db->open(db, name, NULL,
			DB_BTREE, flags, S_IRUSR | S_IWUSR)) != 0) {
	dbenv->err(dbenv, ret, "db->open: %s", name);
	exit (1);
    }	
}

void
env_dir_create()
{
    struct stat sb;

    /*
     * If the directory exists, we're done.  We do not further check
     * the type of the file, DB will fail appropriately if it's the
     * wrong type.
     */
    if (stat(ENV_DIRECTORY, &sb) == 0)
	return;

    /* Create the directory, read/write/access owner only. */
    if (mkdir(ENV_DIRECTORY, S_IRWXU) != 0) {
	fprintf(stderr,
		"txnapp: mkdir: %s: %s\n", ENV_DIRECTORY, strerror(errno));
	exit (1);
    }
}

void
env_open(int join)
{
    int ret;
    u_int32_t flags, flags1, flags0;

    flags0 = DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG |
	DB_INIT_MPOOL | DB_INIT_TXN | DB_RECOVER | DB_THREAD;
    flags1 = DB_JOINENV | DB_THREAD;

    if (join == 1)
	flags = flags1;
    else
	flags = flags0;

    /* Create the environment handle. */
    if ((ret = db_env_create(&dbenv, 0)) != 0) {
	fprintf(stderr,
		"txnapp: db_env_create: %s\n", db_strerror(ret));
	exit (1);
    }

    /* Set up error handling. */
    dbenv->set_errpfx(dbenv, "txnapp");

    /*
     * Open a transactional environment:
     *	create if it doesn't exist
     *	free-threaded handle
     *	run recovery
     *	read/write owner only
     */
    if ((ret = dbenv->open(dbenv, ENV_DIRECTORY, flags,
			   S_IRUSR | S_IWUSR)) != 0) {
	dbenv->err(dbenv, ret, "dbenv->open: %s", ENV_DIRECTORY);
	exit (1);
    }
}

int
add_fruit(char *fruit, char *name,int time,int time2,int myid)
{
    DBT key, data;
	
    DB_TXN *tid;
	
    int ret, t_ret;
    struct timeval mytime;

    /* Initialization. */
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    key.data = fruit;
    key.size = strlen(fruit);
    data.data = name;
    data.size = strlen(name);
    data.flags = DB_DBT_MALLOC;

    sleep(time);  /* await for the time to start */
    printf("Enter add fruit\n");

    /* Begin the transaction. */
    if ((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0) {
	dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
	exit (1);
    }
    env_current(&mytime,7);
    printf("P - Current time is (((%ld,%ld))): ---------->",
	   mytime.tv_sec, mytime.tv_usec);
    printf("P - Start transaction %d(Transaction ID:) \n",myid);
	

    sleep(time2);

    env_current(&mytime,7);
    printf("P - Current time is (((%ld,%ld))): --------->",
	   mytime.tv_sec,mytime.tv_usec);
    printf("P - Put Operation in transaction %d(Transaction ID:) \n",myid);
    db->put(db, tid, &key, &data, 0); 

    switch (db->put(db, tid, &key, &data, 0)) {

    case 0:
	env_current(&mytime,7);
	        
	printf("P - Current time is (((%ld,%ld))): ------->",
	       mytime.tv_sec, mytime.tv_usec);
	printf("P - Commit transaction %d(Transaction ID: ) \n",myid);

	if ((ret = txn_commit(tid, 0)) != 0) {
	    dbenv->err(dbenv, ret, "DB_TXN->commit");
	    exit (1);
	}

	memp_sync(dbenv, NULL);

	break;
    case DB_LOCK_DEADLOCK:
    default:
	env_current(&mytime,7);
	printf("P - Current time is (((%ld,%ld))): ----->",
	       mytime.tv_sec, mytime.tv_usec);
	printf("P - ABORT this transaction %d() !!!!!!\n",myid);		
		
	if ((t_ret = txn_abort(tid)) != 0) {
	    dbenv->err(dbenv, t_ret, "DB_TXN->abort");
	    exit (1);
	}
	break;
    }
    return;
}

int
read_fruit(char *fruit, char *name,int time,int time2,int myid)
{
    DBT key, data;
	
    DB_TXN *tid;
	
    int ret, t_ret;
    char *contents;
    struct timeval mytime;
	
    /* Initialization. */	
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    key.data = fruit;
    key.size = strlen(fruit);
    data.flags = DB_DBT_MALLOC;

    sleep(time);

    /* Begin the transaction. */
	
    if ((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0) {
	dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
	exit (1);
    }
    printf("C - ** read, txn_begin, tid=(((%d)))\n", tid);

    txn_current(tid, &mytime, 7);

    printf("C - 10 Current txn time is (((%ld,%ld))): --->", mytime.tv_sec, mytime.tv_usec);
    printf("C - Start transaction %d(Transaction ID: ) \n",myid);
	

    /*env_current(&mytime, 7);*/
    printf("C - 20 Current time is (((%ld,%ld))): --->", mytime.tv_sec, mytime.tv_usec);
    printf("C - Call txn_current in transaction %d(Transaction ID:) \n",myid);

    sleep(time2);

    env_current(&mytime,7);
    printf("C - 30 Current time is (((%ld,%ld))): ---->", mytime.tv_sec, mytime.tv_usec);
    printf("C - Get Operation in transaction %d(Transaction ID:) \n",myid);

    switch (ret = db->get(db, tid, &key, &data, 0)) {
    case 0:
	printf("<DATA: (%.*s)> \n",(int)data.size,(char *)data.data);
	env_current(&mytime,7);
	printf("C - 40 Current time is (((%ld,%ld))): --->", mytime.tv_sec, mytime.tv_usec);
	printf("C - Commit transaction %d(Transaction ID:) \n",myid);
		
	printf(" **** read, txn_commit,tid=(((%d)))\n", tid);
	if ((ret = txn_commit(tid, 0)) != 0) {
	    dbenv->err(dbenv, ret, "DB_TXN->commit");
	    exit (1);
	}
	break;
    case DB_LOCK_DEADLOCK:
	printf("C - DB_LOCK_DEADLOCK\n");
    case TEMPORAL_CLK_CHECK:
	printf("C - CLK_ERR\n");
    default:
	env_current(&mytime,7);
	printf("C - FAILED TO GET THE DATA\n and THAT IS WHAT WE EXPECT\n");
	printf("C - Current time is (((%ld,%ld))): --->", mytime.tv_sec, mytime.tv_usec);
	printf("[check_msg]C - Abort this transaction %d(Transaction ID: ) \n",myid);
		
	if ((t_ret = txn_abort(tid)) != 0) {
	    dbenv->err(dbenv, t_ret, "DB_TXN->abort");
	    exit (1);
	}
	break;
    }
    return;
}

int
main(int argc, char *argv[])
{
    extern char *optarg;
    extern int optind;

    struct timeval mytime;
    int ch, i;
    int pid;

    if ((argc == 2) && (strcmp(argv[1], "partition") == 0))
	openFlag =  DB_PARTITION;

    env_dir_create();
	
    printf("I am Parent===========\n");

    env_current(&mytime,7);
    printf("P - ********The current time is (((%ld)))*********\n",mytime);

    env_open(0);
    printf("P - env created\n");

    if ((pid = fork()) == 0) {
	sleep(3);
	printf("I am Child============\n");

	env_current(&mytime,7);
	printf("C - ********The current time is (((%ld)))*********\n",mytime);

	env_open(1);
	printf("C - env joined\n");
	
	/* Open database: Key is fruit class; Data is specific type. */
	db_open("fruit", 0, 1);
	printf("C - db opened\n");

	read_fruit("apple", "", 0, 10, 111);
	printf("C - data read\n");

	printf("STOP Child============\n");
	return(0);
    } else {

	/* Open database: Key is fruit class; Data is specific type. */
	db_open("fruit", 0, 0);
	printf("P - db opened\n");

	sleep(6);

	add_fruit("apple", "red_RED_red", 0, 0, 999);
	printf("P - data added\n");

	printf("STOP Parent===========\n");
    }
    sleep(10);
    return (0);
}










