#include <sys/types.h>
#include <stdio.h>
#include <time.h>
#include <sys/times.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "db.h"

/* for debug (Shilong) */
const char prefixMsg[] = "suit01";

# define ENV_DIRECTORY "./txnTemp/"
# define TT_DEBUG_FILENAME1 "TTDebugFile1"
# define TT_DEBUG_FILENAME2 "TTDebugFile2"
# define MAX_NUM_RECORDS 1000
# define FALSE 0
# define TRUE 1
# define NUM_WORKERS 2
# define TT_DEBUG 1

typedef struct dbInformation {
    DB *dbp;
    DB_ENV *dbenv;
    int id;
    int numRecords;
} dbInformation;

int simpleTest(char * dbName, char * dbNameWithEnv, int openFlag);
int advTest(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords,
	    int numVersions, int numRecordsToRead,
	    int printTTAfterTxnCommit, int printTTBforTxnCommit, int printTTBforSync);
int basicCrashTest(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords);
int advCrashTest(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords);
int basicTxnTimeTest(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords);

int displayDB(char * dbName, int openFlag, int numRecords);
int displayNonSeqDB(char * dbName, int openFlag, int numRecords,
		    struct timeval startupTime, struct timeval shutdownTime,
		    int printTTBforSync, int printTTBforTxnCommit);

int dbStartup(char *dbName, DB **dbpPtr, DB_ENV **dbenvPtr, DB_TXN ** tidPtr, int openFlag);
int dbShutdown(DB *dbp, DB_ENV *dbenv, DB_TXN * tid, int printTT);
int insertData(DB *dbp, DB_TXN *tid, int keyVal, char * strData);

void *workers(void *);
int dbReader(dbInformation dbInfo);
int dbWriter(dbInformation dbInfo);

/* TODO : If TT_DEBUG MODE then make calls to printTT */
/* To remove calls to printTT when forming the release test case */
int printTT(DB_ENV *dbenv);

/* For performing diff, time related information must be made as OK or NOT OK */
/* Code for cleanup of the direcoty */

int main (int argc, char *argv[]) {

    int retVal, numRecords, numVersions, numRecordsToRead,
 	openFlag = DB_CREATE|DB_TEMPORAL;

    if ((argc == 2) && (strcmp(argv[1], "partition") == 0))
	openFlag |= DB_PARTITION;

    /* Test 1 */
    printf("[%s]Test 1 --@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n",
	   prefixMsg);
    retVal = simpleTest("bdbTest1.db", "./txnTemp/bdbTest1.db", openFlag);

    return(0);
}

int simpleTest(char * dbName, char * dbNameWithEnv, int openFlag)
{
    DB *dbp;
    DBC *dbcp1, *dbcp2;
    DBT key, data;
    DB_ENV *dbenv;
    DB_TXN *tid1, *tid2;
    int ret, keyVal;
    struct timeval beginning, forever, start, stop; 
    char strData[100];

    /*Create the environment handle. */
    if ((ret = db_env_create(&dbenv, 0)) != 0) {
	printf("[%s] dbStartup: db_env_create: %s\n", prefixMsg, db_strerror(ret));
	return (-1);
    }

    /* Opening the Environment and setting its options */
    if ((ret = dbenv->open(dbenv, ENV_DIRECTORY,
			   DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG |
			   DB_INIT_MPOOL | DB_INIT_TXN | DB_RECOVER | DB_THREAD,
			   0664)) != 0) {
	printf("[%s] dbStartup: DB_ENV->open: %s\n", prefixMsg, db_strerror(ret));
	(void)dbenv->close(dbenv, 0);
	dbenv->err(dbenv, ret, "dbenv->open: %s", ENV_DIRECTORY);
	return(-1);
    }

    /* Creating a db handle under the dbenvironment */
    if ((ret = db_create(&dbp, dbenv, 0)) != 0) {
	printf("[%s] dbStartup: db_create: %s\n", prefixMsg, db_strerror(ret));
	return(-1);
    }

    if ((ret = dbp->open(dbp, dbName, NULL, DB_BTREE, openFlag, 0664)) != 0) {
	printf("[%s] dbStartup: DB->open: %s\n", prefixMsg, db_strerror(ret));
	dbp->err(dbp, ret, "db->open: %s", dbName);
	return(-1);
    }
    printf("[%s] DATABASE OPENED\n", prefixMsg);

    /* Begin the transaction. */
    if ((ret = txn_begin(dbenv, NULL, &tid1, 0)) != 0) {
	dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
	return (-1);
    }

    /* Begin the transaction. */
    if ((ret = txn_begin(dbenv, NULL, &tid2, 0)) != 0) {
	dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
	return (-1);
    }

    printf("[%s] txn1 insert data item\n", prefixMsg);

    keyVal = 0;

    keyVal++;
    strcpy(strData, "Computer Science");
    insertData(dbp, tid1, keyVal, strData);

    printf("[%s] txn1 non-sequenced get the data item\n", prefixMsg);
    
    /* Acquire a cursor for the database. */
    if ((ret = dbp->cursor(dbp, tid1, &dbcp1, 0)) != 0) {
	dbp->err(dbp, ret, "DB->cursor");
	return(-1);
    }

    /* Initialize the key/data return pair. */
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    /* Walk through the database and print out the key/data pairs. */
    beginning.tv_sec = BEGINNING;
    beginning.tv_usec = 0;
    forever.tv_sec = FOREVER;
    forever.tv_usec = 0;

    while ((ret = dbcp1->c_get_nonsequenced(dbcp1, &key, &data, beginning, forever,
					  DB_NEXT, &start, &stop)) == 0) {
	if (is_us(start))
	    printf("txn 1 (is_us()): OK\n");
	else
	    printf("txn 1 (is_us()): ERROR\n");

	if (is_other(start))
	    printf("txn 1 (is_other()): ERROR\n");
	else
	    printf("txn 1 (is_other()): OK\n");

	if (is_until_changed(stop))
	    printf("txn 1 (is_until_changed()): OK\n");
	else
	    printf("txn 1 (is_until_changed()): ERROR\n");
    }

    if (ret != DB_NOTFOUND && ret != 0) {
	dbp->err(dbp, ret, "DBcursor->get");
	return(-1);
    }
    if ((ret = dbcp1->c_close(dbcp1)) != 0) {
	dbp->err(dbp, ret, "db cursor close");
    }

    if (tid1 != NULL) {
        if ((ret = txn_commit(tid1,0)) != 0) {
            dbenv->err(dbenv, ret, "DB_TXN->commit");
            return (-1);
        }
    }

    printf("[%s] txn2 non-sequenced get the data item\n", prefixMsg);
    
    /* Acquire a cursor for the database. */
    if ((ret = dbp->cursor(dbp, tid2, &dbcp2, 0)) != 0) {
	dbp->err(dbp, ret, "DB->cursor");
	return(-1);
    }

    /* Initialize the key/data return pair. */
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    /* Walk through the database and print out the key/data pairs. */
    beginning.tv_sec = BEGINNING;
    beginning.tv_usec = 0;
    forever.tv_sec = FOREVER;
    forever.tv_usec = 0;

    while ((ret = dbcp2->c_get_nonsequenced(dbcp2, &key, &data, beginning, forever,
					  DB_NEXT, &start, &stop)) == 0) {
	if (is_us(start))
	    printf("txn 2 (is_us()): ERROR\n");
	else
	    printf("txn 2 (is_us()): OK\n");

	/* Because txn1 has committed */
	if (is_other(start))
	    printf("txn 2 (is_other()): ERROR\n");
	else
	    printf("txn 2 (is_other()): OK\n");

	if (is_until_changed(stop))
	    printf("txn 2 (is_until_changed()): OK\n");
	else
	    printf("txn 2 (is_until_changed()): ERROR\n");
    }

    if (ret != DB_NOTFOUND && ret != 0) {
	dbp->err(dbp, ret, "DBcursor->get");
	return(-1);
    }
    if ((ret = dbcp2->c_close(dbcp2)) != 0) {
	dbp->err(dbp, ret, "db cursor close");
    }
    
    if (tid2 != NULL) {
        if ((ret = txn_commit(tid2,0)) != 0) {
            dbenv->err(dbenv, ret, "DB_TXN->commit");
            return (-1);
        }
    }

    if ((ret = dbp->close(dbp, 0)) != 0) {
        dbp->err(dbp, ret, "db close");
        return(-1);
    }

    if ((ret = dbenv->close(dbenv,0)) != 0) {
        dbenv->err(dbenv, ret, "db env close");
        return(-1);
    }

    return(0);
}

int insertData(DB *dbp, DB_TXN *tid, int keyVal, char * strData) {

    DBT key, data;
    int ret;

    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    key.data = (int *) &keyVal;
    key.size = sizeof(keyVal);
    data.data = strData;
    data.size = strlen(strData)+1;

    if ((ret = dbp->put(dbp, tid, &key, &data, 0)) == 0)
	printf("[%s] (key, data) stored: (%d, %s) \n",
	       prefixMsg, *(int *)key.data, (char *)data.data);
    else {
	printf("[%s] ERROR in insertData()\n", prefixMsg);
	dbp->err(dbp, ret, "DB->put");
	/* Modify !!!!!!!!!!!!!!!!!!!!!!!*/
	exit(-1);
    }

    return(0);
}
