#!/bin/bash

if [ $# -ne 3 ]; then
    echo "USAGE: ${0} bdbsrc beeheader_path beelib_path"
    exit 1
fi

bdbsrc=${1}
beeheader=${2}
beelib=${3}

echo "******************** Begin the Regression Test ********************"

let "num_failed = 0"
let "num_total = 0"

# Loop on each testcase subdirectory
for t in *; do
    # Non directories are not testcases
    if [ ! -d ${t} ]; then
	continue
    fi
    if [ "${t}" == "CVS" ]; then
	continue
    fi

    # Do the testcase
    echo "==========> TESTCASE: ${t} <=========="
    let "num_total += 1"
    cd ${t}
    ./testme.sh ${bdbsrc} ${beeheader} ${beelib}
    if [ $? -ne 0 ]; then
	let "num_failed += 1"
	echo "    =====> Failed"
    else
	echo "    =====> OK"
    fi
    cd ..
    echo
done

echo "******************** Regression Test Summary ********************"
echo " Total number of testcases: ${num_total}"
echo " Number of failed testcases: ${num_failed}"
if [ $num_failed -eq 0 ]; then
    echo " Final result: SUCCESSFUL :-)"
#else
#    echo " Final result: FAILED :-("
fi
echo "*****************************************************************"
