#include "exp.h"

int dbStartup(char *dbName, char * dbEnvName, DB **dbpPtr, DB_ENV **dbenvPtr, DB_TXN ** tidPtr, int openFlag);
int dbShutdown(DB *dbp, DB_ENV *dbenv, DB_TXN * tid);
int displayDB(char *dbName, char * dbEnvName, int openFlag, int keyVal, int numRecords);

int numPageReads, numPageWrites;

int dbStartup(char *dbName, char * dbEnvName, DB **dbpPtr, DB_ENV **dbenvPtr, DB_TXN ** tidPtr, int openFlag) {

    int ret;
    DB_ENV *dbenv;
    DB * dbp;

	/* printf("\n creating env "); */
    /*Create the environment handle. */
    if ((ret = db_env_create(dbenvPtr, 0)) != 0) {
		printf(" dbStartup: db_env_create: %s\n", db_strerror(ret));
		return (-1);
    }

	/* printf("\n creating env %s \n ", dbEnvName); */
    dbenv = *dbenvPtr;
    /* Opening the Environment and setting its options */
    if ((ret = dbenv->open(dbenv, dbEnvName,
			   DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG |
			   DB_INIT_MPOOL | DB_INIT_TXN | DB_RECOVER | DB_THREAD,
			   0664)) != 0) {
		printf(" dbStartup: DB_ENV->open: %s\n", db_strerror(ret));
		(void)dbenv->close(dbenv, 0);
		dbenv->err(dbenv, ret, "dbenv->open: %s", ENV_DIRECTORY);
		return(-1);
    }

	/* printf("\n creating db "); */

    /* Creating a db handle under the dbenvironment */
    if ((ret = db_create(dbpPtr, dbenv, 0)) != 0) {
		printf(" dbStartup: db_create: %s\n", db_strerror(ret));
		return(-1);
    }

    dbp = *dbpPtr;

    if ((ret = dbp->open(dbp, dbName, NULL, DB_BTREE, openFlag, 0664)) != 0) {
		printf(" dbStartup: DB->open: %s\n", db_strerror(ret));
		dbp->err(dbp, ret, "db->open: %s", dbName);
		return(-1);
    }

    if (tidPtr != NULL) {
		/* Begin the transaction. */
		if ((ret = txn_begin(dbenv, NULL, tidPtr, 0)) != 0) {
			dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
			return (-1);
		}
    }

    return (0);
}

int dbShutdown(DB *dbp, DB_ENV *dbenv, DB_TXN * tid ) {

    int ret;

    if (tid != NULL) {
		if ((ret = txn_commit(tid,0)) != 0) {
			dbenv->err(dbenv, ret, "DB_TXN->commit");
			return (-1);
		}
    }

    if ((ret = dbp->close(dbp, 0)) != 0) {
		dbp->err(dbp, ret, "db close");
		return(-1);
    }

    if ((ret = dbenv->close(dbenv,0)) != 0) {
		dbenv->err(dbenv, ret, "db env close");
		return(-1);
    }

    return (0);
}

int displayDB(char *dbName, char * dbEnvName, int openFlag, int keyVal, int numRecords) {

    DB *dbp;
    DB_ENV *dbenv;
    DB_TXN * tid;
    DBC *dbcp;
    DBT key, data;
    int ret;

	/* printf("\n calling DB startup "); */
    if ((ret = dbStartup(dbName, dbEnvName, &dbp, &dbenv, &tid, openFlag)) != 0)
	return ret;

	numPageReads = numPageWrites = 0;

    /* Acquire a cursor for the database. */
    if ((ret = dbp->cursor(dbp, tid, &dbcp, 0)) != 0) {
		dbp->err(dbp, ret, "DB->cursor");
		return(-1);
    }

    /* Initialize the key/data return pair. */
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    key.data = (int *) &keyVal;
    key.size = sizeof(keyVal);

	if ((ret = dbcp->c_get(dbcp, &key, &data, DB_SET)) != 0) {

		if (ret != DB_NOTFOUND && ret != 0) {
			dbp->err(dbp, ret, "DBcursor->get");
			return(-1);
		}
		if ((ret = dbcp->c_close(dbcp)) != 0) {
			dbp->err(dbp, ret, "db cursor close");
			return(-1);
		}
	}

	numRecords--;
printf("=====\n");
    /* Walk through the database and print out the key/data pairs. */
    while ((ret = dbcp->c_get(dbcp, &key, &data, DB_NEXT)) == 0 && numRecords-- > 0) {
		
		printf(" (%d, %.*s)\n", *(int *)key.data, (int)data.size, (char *)data.data);
/*
		if (( (*(int *)key.data) % numRecords) == 0) {
		}
		*/
    }

    if (ret != DB_NOTFOUND && ret != 0) {
		dbp->err(dbp, ret, "DBcursor->get");
		return(-1);
    }
    if ((ret = dbcp->c_close(dbcp)) != 0) {
		dbp->err(dbp, ret, "db cursor close");
		return(-1);
    }

    printf(" numPageReads : %d , numPageWrites : %d \n", numPageReads, numPageWrites);

    if ((ret = dbShutdown(dbp, dbenv, tid)) != 0) return ret;

    return (0);
}

int main (int argc, char** argv) {

	int numVersions, dbType, openFlag;
	char dbName[30], dbEnvName[60];
    int ret, keyVal, startTime, stopTime;
    struct tms buffer;

	if (argc != 4) {
		printf(" Usage: exp1 dbType numVersions keyVal \n");
		return (0);
	}

	dbType = atoi(argv[1]);
	numVersions = atoi(argv[2]);
	keyVal = atoi(argv[3]);

	if (dbType == 1) {
		openFlag = DB_CREATE;
		sprintf(dbName, "bdbc_%d.db", numVersions);
		sprintf(dbEnvName, "%sbdbc_%d/", ENV_DIRECTORY, numVersions);
	}
	if (dbType == 2) {
		openFlag = DB_CREATE|DB_TEMPORAL;
		sprintf(dbName, "tbdb_%d.db", numVersions);
		sprintf(dbEnvName, "%stbdb_%d/", ENV_DIRECTORY, numVersions);
	}
	if (dbType == 3) {
		openFlag = DB_CREATE|DB_TEMPORAL|DB_PARTITION;
		sprintf(dbName, "pbdb_%d.db", numVersions);
		sprintf(dbEnvName, "%spbdb_%d/", ENV_DIRECTORY, numVersions);
	}

    startTime = times(&buffer);

	/* printf("\n calling display DB "); */
	ret = displayDB(dbName, dbEnvName, openFlag, keyVal, NUMRECORDS);

	stopTime = times(&buffer);
	printf(" Time taken by bdbc : %d \n ", stopTime - startTime);

    return (0);

}
