#include "exp.h"

int populateDB(char * dbName, char * dbEnvName, int openFlag, int numVersions);
int dbStartup(char *dbName, char * dbEnvName, DB **dbpPtr, DB_ENV **dbenvPtr, DB_TXN ** tidPtr, int openFlag);
int dbShutdown(DB *dbp, DB_ENV *dbenv, DB_TXN * tid);
int insertData(DB *dbp, DB_TXN *tid, int keyVal, char * strData, int dataLen);

int main (int argc, char** argv) {

	int numVersions, dbType, openFlag;
	char dbName[30], dbEnvName[60], strCommand[100];

	if (argc != 3) {
		printf(" Usage: expSetup dbType numVersions \n");
		return (0);
	}

	numVersions = atoi(argv[2]);
	dbType = atoi(argv[1]);

	if (dbType == 1) {
		openFlag = DB_CREATE;
		sprintf(dbName, "bdbc_%d.db", numVersions);
		sprintf(dbEnvName, "%sbdbc_%d/", ENV_DIRECTORY, numVersions);
	}
	if (dbType == 2) {
		openFlag = DB_CREATE|DB_TEMPORAL;
		sprintf(dbName, "tbdb_%d.db", numVersions);
		sprintf(dbEnvName, "%stbdb_%d/", ENV_DIRECTORY, numVersions);
	}
	if (dbType == 3) {
		openFlag = DB_CREATE|DB_TEMPORAL|DB_PARTITION;
		sprintf(dbName, "pbdb_%d.db", numVersions);
		sprintf(dbEnvName, "%spbdb_%d/", ENV_DIRECTORY, numVersions);
	}

	sprintf(strCommand, "mkdir %s", dbEnvName);
	/* printf(" %s \n ", strCommand); */
	system(strCommand);

	populateDB(dbName, dbEnvName, openFlag, numVersions);

	return(0);

}

int populateDB(char * dbName, char * dbEnvName, int openFlag, int numVersions) {

    DB *dbp;
    DBC *dbcp;
    DBT key, data;
    DB_ENV *dbenv;
    DB_TXN * tid;
    int ret, keyVal;
    struct timeval startupTime, shutdownTime;
    char strData[RECSIZE];
    int numTotalRecs, numCurrentRecs, numRecsPerPage, version;

	/* Number of versions: 1, 4, 8, 12... */

	if ((ret = dbStartup(dbName, dbEnvName, &dbp, &dbenv, &tid, openFlag)) != 0)
	    return ret;

	numRecsPerPage = (PAGESIZE - PAGEHEADERSIZE) / RECSIZE;
	numTotalRecs = numRecsPerPage * DBSIZE;
	numCurrentRecs = numTotalRecs / numVersions;

	/* If the db is non temporal then its BDBC */
	if (!(openFlag & DB_TEMPORAL)) numVersions = 1;

	memset(strData, 80, RECSIZE);

	/* Inserting without huge page splits, but with a txn per insert */
	for (keyVal=1; keyVal <= numCurrentRecs; keyVal++) {

	   	for (version=1; version <= numVersions; version++) {
		        //printf("(%d,%d)", keyVal, version);fflush(stdout);
			sprintf(strData, " Key %d Version %d ", keyVal, version);
			insertData(dbp, tid, keyVal, strData, RECSIZE);

			if ((ret = txn_commit(tid,0)) != 0) {
				dbenv->err(dbenv, ret, "DB_TXN->commit");
				return (-1);
			}

			if ((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0) {
				dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
				return (-1);
			}
		}
		/*
		if ((keyVal % 100) == 0) {

			if ((ret = dbShutdown(dbp, dbenv, tid)) != 0)
				return ret;

			if ((ret = dbStartup(dbName, dbEnvName, &dbp, &dbenv, &tid, openFlag)) != 0)
				return ret;
		}
		*/
	}


	if ((ret = dbShutdown(dbp, dbenv, tid)) != 0)
	    return ret;

}

int insertData(DB *dbp, DB_TXN *tid, int keyVal, char * strData, int dataLen) {

    DBT key, data;
    int ret;

    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    key.data = (int *) &keyVal;
    key.size = sizeof(keyVal);
    data.data = strData;
    data.size = dataLen;

    if ((ret = dbp->put(dbp, tid, &key, &data, 0)) != 0)
	{
		printf(" ERROR in insertData()\n");
		dbp->err(dbp, ret, "DB->put");
		return (ret);
    }

    return(0);
}

int dbStartup(char *dbName, char * dbEnvName, DB **dbpPtr, DB_ENV **dbenvPtr, DB_TXN ** tidPtr, int openFlag) {

    int ret;
    DB_ENV *dbenv;
    DB * dbp;


    /*Create the environment handle. */
    if ((ret = db_env_create(dbenvPtr, 0)) != 0) {
		printf(" dbStartup: db_env_create: %s\n", db_strerror(ret));
		return (-1);
    }

    dbenv = *dbenvPtr;
    /* Opening the Environment and setting its options */
    if ((ret = dbenv->open(dbenv, dbEnvName,
			   DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG |
			   DB_INIT_MPOOL | DB_INIT_TXN | DB_RECOVER | DB_THREAD,
			   0664)) != 0) {
		printf(" dbStartup: DB_ENV->open: %s\n", db_strerror(ret));
		(void)dbenv->close(dbenv, 0);
		dbenv->err(dbenv, ret, "dbenv->open: %s", ENV_DIRECTORY);
		return(-1);
    }

    /* Creating a db handle under the dbenvironment */
    if ((ret = db_create(dbpPtr, dbenv, 0)) != 0) {
		printf(" dbStartup: db_create: %s\n", db_strerror(ret));
		return(-1);
    }

    dbp = *dbpPtr;

    if ((ret = dbp->open(dbp, dbName, NULL, DB_BTREE, openFlag, 0664)) != 0) {
		printf(" dbStartup: DB->open: %s\n", db_strerror(ret));
		dbp->err(dbp, ret, "db->open: %s", dbName);
		return(-1);
    }

    if (tidPtr != NULL) {
		/* Begin the transaction. */
		if ((ret = txn_begin(dbenv, NULL, tidPtr, 0)) != 0) {
			dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
			return (-1);
		}
    }

    return (0);
}

int dbShutdown(DB *dbp, DB_ENV *dbenv, DB_TXN * tid ) {

    int ret;

    if (tid != NULL) {
		if ((ret = txn_commit(tid,0)) != 0) {
			dbenv->err(dbenv, ret, "DB_TXN->commit");
			return (-1);
		}
    }

    if ((ret = dbp->close(dbp, 0)) != 0) {
		dbp->err(dbp, ret, "db close");
		return(-1);
    }

    if ((ret = dbenv->close(dbenv,0)) != 0) {
		dbenv->err(dbenv, ret, "db env close");
		return(-1);
    }

    return (0);
}


