#include <sys/types.h>
#include <stdio.h>
#include <time.h>
#include <sys/times.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "db.h"

# define ENV_DIRECTORY "/tmp/txnTemp/"
# define MAX_NUM_RECORDS 1000
# define FALSE 0
# define TRUE 1
# define NUM_WORKERS 2

typedef struct dbInformation {
	DB *dbp;
	DB_ENV *dbenv;
	int id;
	int numRecords;
} dbInformation;


int simpleTest(char * dbName, char * dbNameWithEnv, int openFlag);
int advTest(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords, int numVersions, int numRecordsToRead, int printTTAfterTxnCommit, int printTTBforTxnCommit, int printTTBforSync);
int basicCrashTest(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords);
int advCrashTest(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords);
int basicTxnTimeTest(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords);

int displayDB(char * dbName, int openFlag, int numRecords);
int dbStartup(char *dbName, DB **dbpPtr, DB_ENV **dbenvPtr, DB_TXN ** tidPtr, int openFlag);
int dbShutdown(DB *dbp, DB_ENV *dbenv, DB_TXN * tid, int printTT);
int insertData(DB *dbp, DB_TXN *tid, int keyVal, char * strData);

void *workers(void *);
int dbReader(dbInformation dbInfo);
int dbWriter(dbInformation dbInfo);


/* 			For performing diff, time related information must be made as OK or NOT OK */
/*			Code for cleanup of the direcoty */

int main () {

	int retVal, numRecords, numVersions, numRecordsToRead, openFlag = DB_CREATE;

	/* Test 1 */
	printf("\nTEST 1 @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
	retVal = simpleTest("bdbTest1.db", "/tmp/txnTemp/bdbTest1.db", openFlag);
	
	return (0);
}

/* Working Seq of this test function :
	Open (Env, DB, Txn) -> 	Insert data -> 	Close (Env, DB, Txn) -> Display Data Seq
	Open (Env, DB, Txn) -> 	Update data -> 	Close (Env, DB, Txn) -> Display Data Seq, Non Seq
	Open (Env, DB, Txn) -> 	Delete data -> 	Close (Env, DB, Txn) -> Display Data Seq, Non Seq
*/

int simpleTest(char * dbName, char * dbNameWithEnv, int openFlag)
{
	DB *dbp;
	DBC *dbcp;
	DBT key, data;
	DB_ENV *dbenv;
	DB_TXN * tid;
	int ret, keyVal = 0;
	char strData[100];

	/**************** DATA INSERTION PHASE *****************/

	if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0) return ret;

	printf(" Inserting Data into the database \n");

	keyVal++;
	strcpy(strData, "Computer Science");
	insertData(dbp, tid, keyVal, strData);

	keyVal++;
	strcpy(strData, "Electronics and Communication Engg");
	insertData(dbp, tid, keyVal, strData);

	keyVal++;
	strcpy(strData, "Physics");
	insertData(dbp, tid, keyVal, strData);

	if ((ret = dbShutdown(dbp, dbenv, tid, FALSE)) != 0) return ret;

	if ((ret = displayDB(dbNameWithEnv, openFlag, MAX_NUM_RECORDS)) != 0) return (-1);


	/**************** DATA UPDATION PHASE *****************/

	if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0) return ret;

	printf(" Updating Data in the database \n");

	/** Updating Data in the Database **/

	keyVal = 0;

	keyVal++;
	strcpy(strData, "Computer Science - Changed");
	insertData(dbp, tid, keyVal, strData);

	keyVal++;
	strcpy(strData, "Electronics and Communication Engg - Changed");
	insertData(dbp, tid, keyVal, strData);

	keyVal++;
	strcpy(strData, "Physics - Changed");
	insertData(dbp, tid, keyVal, strData);

	keyVal = 0;

	keyVal++;
	strcpy(strData, "Computer Science - Changed Again");
	insertData(dbp, tid, keyVal, strData);

	if ((ret = dbShutdown(dbp, dbenv, tid, FALSE)) != 0) return ret;

	printf(" Sleeping... \n");
	sleep(1);
	if ((ret = displayDB(dbNameWithEnv,openFlag, MAX_NUM_RECORDS)) != 0) return (-1);

	/**************** DATA DELETION PHASE *****************/

	if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0) return ret;

	printf(" Deleting Data in the database \n");

	/** Deleting Data in the Database **/

	keyVal = 2;
	memset(&key, 0, sizeof(key));
	key.data = (int *) &keyVal;
	key.size = sizeof(keyVal);

	if ((ret = dbp->del(dbp, tid, &key, 0)) == 0)
		printf("db: %d: key was deleted.\n", *(int *)key.data);
	else {
		dbp->err(dbp, ret, "DB->del");
		return(-1);
	}

	keyVal = 3;
	memset(&key, 0, sizeof(key));
	key.data = (int *) &keyVal;
	key.size = sizeof(keyVal);

	if ((ret = dbp->del(dbp, tid, &key, 0)) == 0)
		printf("db: %d: key was deleted.\n", *(int *)key.data);
	else {
		dbp->err(dbp, ret, "DB->del");
		return(-1);
	}

	if ((ret = dbShutdown(dbp, dbenv, tid, FALSE)) != 0) return ret;

	if ((ret = displayDB(dbNameWithEnv,openFlag, MAX_NUM_RECORDS)) != 0) return (-1);

	/********************* DATA GET PHASE, NON-CURSOR *******************/
	if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0)
	    return  ret;
	printf("Getting Data in the database For Key = 1\n");

	/** Getting  Data in the Database **/
	keyVal = 1;
	memset(&key, 0, sizeof(key));
	memset(&data, 0, sizeof(data));
	key.data = (int *) &keyVal;
	key.size = sizeof(keyVal);
	data.flags = DB_DBT_MALLOC;

	if ((ret = dbp->get(dbp, tid, &key, &data, 0)) == 0)
	    printf("%d : %.*s\n", *(int *)key.data, (int)data.size, 
		   (char *)data.data);
	else {
	    dbp->err(dbp, ret, "DB->get");
	    return(-1);
	}
	if ((ret = dbShutdown(dbp, dbenv, tid, FALSE)) != 0)
	    return ret;                                   
	/****************************************************************************/

	return(0);

}


/* Working Seq of this test function :
	For the number of versions of updates required :
		-> Open (Env, DB, Txn)
		-> Insert data (effectively Insert first time, later update) - as many records as numRecords
		-> Close (Env, DB, Txn)
	End For
	-> Display Data Seq, Non Seq (Display as many records as specified by numRecordsToRead)
*/

int advTest(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords, int numVersions, int numRecordsToRead, int printTTAfterTxnCommit, int printTTBforTxnCommit, int printTTBforSync)
{
	DB *dbp;
	DBC *dbcp;
	DB_ENV *dbenv;
	DBT key, data;
	DB_TXN * tid;
	int ret, keyVal, version;
	char  strData[200];

	for (version=1; version <= numVersions; version++) {

		if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0) return ret;

		printf(" Inserting Data into the database \n");

		for (keyVal=1; keyVal <= numRecords; keyVal++) {

			sprintf(strData, " Computer Science For Key %d Version %d. ", keyVal, version);
			insertData(dbp, tid, keyVal, strData);
		}

		if ((ret = dbShutdown(dbp, dbenv, tid, printTTAfterTxnCommit)) != 0) return ret;

	}

	if ((ret = displayDB(dbNameWithEnv,openFlag, numRecordsToRead)) != 0) return (-1);

	return(0);

}

int basicCrashTest(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords) {

	DB *dbp;
	DB_ENV *dbenv;
	DBT key, data;
	DB_TXN * tid;
	int ret, keyVal;
	char  strData[200];
	int childProcessId, childProcessStatus;

	childProcessId = fork();
	if (childProcessId == 0) {

		/* I am child */

		if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0) return ret;

		printf(" Inserting Data into the database under txn 1 \n");

		for (keyVal=1; keyVal <= numRecords; keyVal++) {

			sprintf(strData, " Computer Science For Key %d.  ", keyVal);
			insertData(dbp, tid, keyVal, strData);
		}

		/* TODO : sync call */

		if ((ret = txn_commit(tid,0)) != 0) {
			dbenv->err(dbenv, ret, "DB_TXN->commit");
			return (-1);
		}

		if ((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0) {
			dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
			return (-1);
		}

		printf(" Inserting Data into the database under txn 2 \n");

		for (keyVal=numRecords+1; keyVal <= 2*numRecords; keyVal++) {

			sprintf(strData, " Computer Science For Key %d.  ", keyVal);
			insertData(dbp, tid, keyVal, strData);
			/* sleep(1); */
		}

		/* TODO : sync call */

		if ((ret = txn_commit(tid,0)) != 0) {
			dbenv->err(dbenv, ret, "DB_TXN->commit");
			return (-1);
		}

		/* TODO : print the TT to File1 */
		if ((ret = txn_checkpoint(dbenv, 0, 0, DB_FORCE)) != 0) {
			dbenv->err(dbenv, ret, "DB_ENV->checkpoint");
			return (-1);
		}

		printf(" child exiting \n");
		exit(1);
	}

	/* I am parent process */
	ret = wait(&childProcessStatus);

	if ((ret = displayDB(dbNameWithEnv,openFlag, MAX_NUM_RECORDS)) != 0) return (-1);

	return(0);
}

int advCrashTest(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords) {

	DB *dbp;
	DB_ENV *dbenv;
	DBT key, data;
	DB_TXN * tid;
	int ret, keyVal;
	char  strData[200];
	int childProcessId, childProcessStatus;

	childProcessId = fork();
	if (childProcessId == 0) {

		/* I am child */

		if ((ret = dbStartup(dbName, &dbp, &dbenv, &tid, openFlag)) != 0) return ret;

		printf(" Inserting Data into the database under txn 1 \n");

		for (keyVal=1; keyVal <= numRecords; keyVal++) {

			/* TODO : replace calls to timestamp by calls to env_current */
			sprintf(strData, " Computer Science For Key %d.  ", keyVal);
			insertData(dbp, tid, keyVal, strData);
			/* sleep(1); */
		}

		/* TODO : sync call */

		if ((ret = txn_commit(tid,0)) != 0) {
			dbenv->err(dbenv, ret, "DB_TXN->commit");
			return (-1);
		}

		if ((ret = txn_checkpoint(dbenv, 0, 0, DB_FORCE)) != 0) {
			dbenv->err(dbenv, ret, "DB_ENV->checkpoint");
			return (-1);
		}

		if ((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0) {
			dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
			return (-1);
		}

		printf(" Inserting Data into the database under txn 2 \n");

		for (keyVal=numRecords+1; keyVal <= 2*numRecords; keyVal++) {

			/* TODO : replace calls to timestamp by calls to env_current */
			sprintf(strData, " Computer Science For Key %d.  ", keyVal);
			insertData(dbp, tid, keyVal, strData);
		}

		if ((ret = txn_commit(tid,0)) != 0) {
			dbenv->err(dbenv, ret, "DB_TXN->commit");
			return (-1);
		}

		/* TODO : sync call */

		printf(" child exitting \n");
		exit(1);
	}

	/* I am parent process */
	ret = wait(&childProcessStatus);

	if ((ret = displayDB(dbNameWithEnv, openFlag, MAX_NUM_RECORDS)) != 0) return (-1);

	return(0);
}

int basicTxnTimeTest(char * dbName, char * dbNameWithEnv, int openFlag, int numRecords) {

	DB *dbp;
	DB_ENV *dbenv;
	dbInformation dbInfo[NUM_WORKERS];

	time_t mytime;
	pthread_t ptid;
	int i, ret;

	/* Thread */
	pthread_t pid[NUM_WORKERS];
	pthread_attr_t attr;


	pthread_attr_init(&attr);
	pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);

	/* Thread */

	if ((ret = dbStartup(dbName, &dbp, &dbenv, NULL, openFlag)) != 0) return ret;

	/* create workers processes */
	for (i=0; i<NUM_WORKERS; i++) {
		dbInfo[i].dbp = dbp;
		dbInfo[i].dbenv = dbenv;
		dbInfo[i].id = i;
		dbInfo[i].numRecords = numRecords;
		ret = pthread_create(&pid[i], &attr, workers, (void *)&dbInfo[i]);
	}

	for ( i=0; i<NUM_WORKERS; i++) {
		ret = pthread_join(pid[i], NULL);
	}

	if ((ret = dbShutdown(dbp, dbenv, NULL, FALSE)) != 0) return ret;

	if ((ret = displayDB(dbNameWithEnv,openFlag, MAX_NUM_RECORDS)) != 0) return (-1);


	return(0);
}

void *workers(void *arg) {
	dbInformation dbInfo;
	int ret;

	dbInfo = *(dbInformation *) arg;

	if ((dbInfo.id % NUM_WORKERS) == 0 ) {

		ret = dbReader(dbInfo);
	}
	else {

		ret = dbWriter(dbInfo);
	}
}

int dbReader(dbInformation dbInfo) {

	DB *dbp;
	DB_ENV *dbenv;
	DBC *dbcp;
	DBT key, data;
	int ret;
	DB_TXN * tid;

	printf("MyId: %d : Reader : Starting Txn \n", dbInfo.id);

	dbp = dbInfo.dbp;
	dbenv = dbInfo.dbenv;

	/* Begin the transaction. */
	if ((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0) {
		dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
		return (-1);
	}

	printf("MyId: %d : Reader : Transaction Started \n", dbInfo.id);

	printf("MyId: %d : Retrieving the content using a cursor (Non Sequenced) \n", dbInfo.id);

	/* Acquire a cursor for the database. */
	if ((ret = dbp->cursor(dbp, tid, &dbcp, 0)) != 0) {
		dbp->err(dbp, ret, "DB->cursor");
		return(-1);
	}

	/* Initialize the key/data return pair. */
	memset(&key, 0, sizeof(key));
	memset(&data, 0, sizeof(data));

	/* Walk through the database and print out the key/data pairs. */
	while ((ret = dbcp->c_get(dbcp, &key, &data, DB_NEXT )) == 0) {
		printf("MyId: %d : %d : %.*s \n", dbInfo.id, *(int *)key.data, (int)data.size, (char *)data.data);
	}

	if (ret != DB_NOTFOUND && ret != 0) {
		dbp->err(dbp, ret, "DBcursor->get");
		return(-1);
	}
	if ((ret = dbcp->c_close(dbcp)) != 0) {
		dbp->err(dbp, ret, "db cursor close");
	}

	if ((ret = txn_commit(tid,0)) != 0) {
		dbenv->err(dbenv, ret, "DB_TXN->commit");
		return (-1);
	}

	printf("MyId: %d : Reader : Sleeping before commit \n", dbInfo.id);
	sleep(dbInfo.numRecords+2);

	/* Begin the transaction. */
	if ((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0) {
		dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
		return (-1);
	}

	printf("MyId: %d : Reader : Transaction Commit \n", dbInfo.id);

	printf("MyId: %d : Reader : Sleeping bfor starting new txn \n", dbInfo.id);
	sleep(5);

	/* Acquire a cursor for the database. */
	if ((ret = dbInfo.dbp->cursor(dbInfo.dbp, tid, &dbcp, 0)) != 0) {
		dbInfo.dbp->err(dbInfo.dbp, ret, "DB->cursor");
		return(-1);
	}

	printf("MyId: %d : Reader : New Transaction Started \n", dbInfo.id);

	printf("MyId: %d : Retrieving the content using a cursor (Non Sequenced) \n", dbInfo.id);

	/* Initialize the key/data return pair. */
	memset(&key, 0, sizeof(key));
	memset(&data, 0, sizeof(data));

	/* Walk through the database and print out the key/data pairs. */
	while ((ret = dbcp->c_get(dbcp, &key, &data, DB_NEXT)) == 0) {
		printf("MyId: %d : %d : %.*s  \n", dbInfo.id, *(int *)key.data, (int)data.size, (char *)data.data);
	}

	if (ret != DB_NOTFOUND && ret != 0) {
		dbp->err(dbp, ret, "DBcursor->get");
		return(-1);
	}
	if ((ret = dbcp->c_close(dbcp)) != 0) {
		dbp->err(dbp, ret, "db cursor close");
	}

	if ((ret = txn_commit(tid,0)) != 0) {
		dbenv->err(dbenv, ret, "DB_TXN->commit");
		return (-1);
	}
	printf("MyId: %d : Reader : Transaction Committed \n", dbInfo.id);
	return(0);

}

int dbWriter(dbInformation dbInfo) {

	DB *dbp;
	DB_ENV *dbenv;
	DBT key, data;
	int ret, keyVal;
	DB_TXN * tid;
	char  strData[200];

	printf("MyId: %d : Writer : Called at time ...\n", dbInfo.id);

	dbp = dbInfo.dbp;
	dbenv = dbInfo.dbenv;

	/* TODO :
	  mytime=env_current(6);
	  printf("Current time is %s(%ld): ------------------------------------->",asctime(localtime(&mytime)), mytime);
	  printf("Start transaction %d(Transaction ID: %ld) \n",myid, tid->txnid);
	*/

	/* Begin the transaction. */
	if ((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0) {
		dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
		return (-1);
	}

	printf("MyId: %d : Writer : Transaction Started at time... \n", dbInfo.id);


	for (keyVal=1; keyVal <= dbInfo.numRecords; keyVal++) {

		/* TODO : replace calls to timestamp by calls to env_current */
		sprintf(strData, " Txn Time Data For Key %d Put at time ...  ", keyVal);
		insertData(dbp, tid, keyVal, strData);
		sleep(1);
	}

	printf("MyId: %d : Writer : Sleeping at time... \n", dbInfo.id);
	sleep(5);

	printf("MyId: %d : Writer : Transaction Commit at time... \n", dbInfo.id);

	if ((ret = txn_commit(tid,0)) != 0) {
		dbenv->err(dbenv, ret, "DB_TXN->commit");
		return (-1);
	}

	return (0);

}


int displayDB(char * dbName, int openFlag, int numRecords) {

	DB *dbp;
	DBC *dbcp;
	DBT key, data;
	int ret;

	if ((ret = db_create(&dbp, NULL, 0)) != 0) {
		printf("db_create: %s\n", db_strerror(ret));
		return(-1);
	}

	if ((ret = dbp->open(dbp, dbName, NULL, DB_BTREE, openFlag, 0664)) != 0) {
		dbp->err(dbp, ret, "%s", dbName);
		return(-1);
	}

	printf(" Retrieving the content using a cursor (Sequenced) \n");

	/* Acquire a cursor for the database. */
	if ((ret = dbp->cursor(dbp, NULL, &dbcp, 0)) != 0) {
		dbp->err(dbp, ret, "DB->cursor");
		return(-1);
	}

	/* Initialize the key/data return pair. */
	memset(&key, 0, sizeof(key));
	memset(&data, 0, sizeof(data));

	/* Walk through the database and print out the key/data pairs. */
	while ((ret = dbcp->c_get(dbcp, &key, &data, DB_NEXT)) == 0 && numRecords-- > 0) {
		printf("%d : %.*s\n", *(int *)key.data, (int)data.size, (char *)data.data);
	}

	if (ret != DB_NOTFOUND && ret != 0) {
		dbp->err(dbp, ret, "DBcursor->get");
		return(-1);
	}
	if ((ret = dbcp->c_close(dbcp)) != 0) {
		dbp->err(dbp, ret, "db cursor close");
	}
	if ((ret = dbp->close(dbp, 0)) != 0) {
		dbp->err(dbp, ret, "db close");
		return(-1);
	}

	return (0);
}

int insertData(DB *dbp, DB_TXN *tid, int keyVal, char * strData){

	DBT key, data;
	int ret;

	memset(&key, 0, sizeof(key));
	memset(&data, 0, sizeof(data));
	key.data = (int *) &keyVal;
	key.size = sizeof(keyVal);
	data.data = strData;
	data.size = strlen(strData)+1;

	if ((ret = dbp->put(dbp, tid, &key, &data, 0)) == 0)
		printf("db: key,data stored. : %d, %s \n", *(int *)key.data, (char *)data.data);
	else {
		dbp->err(dbp, ret, "DB->put");
		return(-1);
	}

	return(0);
}

int dbStartup(char *dbName, DB **dbpPtr, DB_ENV **dbenvPtr, DB_TXN ** tidPtr, int openFlag) {

	int ret;
	DB_ENV *dbenv;
	DB * dbp;

	/*Create the environment handle. */
	if ((ret = db_env_create(dbenvPtr, 0)) != 0) {
		printf("txnapp: db_env_create: %s\n", db_strerror(ret));
		return (-1);
	}

	dbenv = *dbenvPtr;
	/* Opening the Environment and setting its options */
	if ((ret = dbenv->open(dbenv, ENV_DIRECTORY,
		DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG |
	    DB_INIT_MPOOL | DB_INIT_TXN | DB_RECOVER | DB_THREAD,
	    0664)) != 0) {
			(void)dbenv->close(dbenv, 0);
			dbenv->err(dbenv, ret, "dbenv->open: %s", ENV_DIRECTORY);
			return(-1);
	}

	/* Creating a db handle under the dbenvironment */
	if ((ret = db_create(dbpPtr, dbenv, 0)) != 0) {
		printf("db_create: %s\n", db_strerror(ret));
		return(-1);
	}
	printf(" dbName : %s \n", dbName); fflush(stdout);
	dbp = *dbpPtr;
	if ((ret = dbp->open(dbp, dbName, NULL, DB_BTREE, openFlag, 0664)) != 0) {
		dbp->err(dbp, ret, "%s", dbName);
		return(-1);
	}

	if (tidPtr != NULL) {
		/* Begin the transaction. */
		if ((ret = txn_begin(dbenv, NULL, tidPtr, 0)) != 0) {
			dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
			return (-1);
		}
	}

	printf(" Database Opened \n");

	return (0);

}

int dbShutdown(DB *dbp, DB_ENV *dbenv, DB_TXN * tid, int printTT) {

	int ret;

	if (tid != NULL) {
		if ((ret = txn_commit(tid,0)) != 0) {
			dbenv->err(dbenv, ret, "DB_TXN->commit");
			return (-1);
		}

		/* TODO :
		Check for printTT flag and call Yao's function, also give a msg that this is happ after txn commit
		[print the TT, the on-disk page count should be 2, the in-memory list should contain 2 pages] */

	}

	if ((ret = dbp->close(dbp, 0)) != 0) {
		dbp->err(dbp, ret, "db close");
		return(-1);
	}

	if ((ret = dbenv->close(dbenv,0)) != 0) {
		dbenv->err(dbenv, ret, "db env close");
		return(-1);
	}

	printf(" Database Closed \n");

	return (0);

}


/*
	Sleep issue, closed and open interval in the txn time..
	Calling sync() says prototype mismatch..
	Check if in normal BDB data can be updated, only if we close the DB and then open it..
	Doesnt look like default DB is opened in Unique Key Mode, find out how to do that..

	Try to solve the makefile (cp to maintain synch issue, in the bdb/build_unix directory)
	Try to solve the libdb.so and libdb.a ????

	BDB:

	Temporal Flag in db env open, check for prerequisit flags (txn, log..)
	db open , add extra check to see if env passed has temporal flag on..
	Change DOC accordingly

   */

