#ifndef _TTDB_H_
#define _TTDB_H_

/*-
 * Developed (2002-2003) by:
 *     TAU Project
 *     Computer Science Dept., University of Arizona
 *     Tucson, AZ 85721, USA
 *     URL  : http://www.cs.arizona.edu/tau
 *     Email: tau@cs.arizona.edu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <sys/types.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <limits.h>
#include "db.h"

#ifndef =====BDBVER=====
#define =====BDBVER=====
#endif

#define DB struct __ttdb         /*the user always uses our TTDB*/
#define DBC struct __ttdbc       /*the user always uses our TTDBC*/
#define db_create __ttdb_create  /*our TTDB gets created by a new function*/

/* New user specified flags */
/* Shilong Yao, 06/01/03
 * In BDB 4.1.25, up to 0x10000000 has been used in DB->flags.
 * So 0x02000000 is changed to 0x020000000, and also TT_DUP and TT_DB_EXISTS
 */
#define DB_TEMPORAL  0x020000000 /* we can not specify flag values that clash
                                  * with already defined flags. (we have looked
                                  * at the flags passed to db_open and the
                                  * flags that are stored in struct __db). */
#define TT_DUP       0x040000000 /* the flags that are already kept in the
                                  * struct __db should not conflict with this
                                  * flag */
#define TT_DB_EXISTS 0x080000000 /* this means that the database is already existing.*/
/* Modified by Huilong */
//#ifdef TBDB_AUDIT
#define DB_AUDIT     0x02000000 
#define DB_INIT_AUDIT  0x100000  /* Concurrent Access Methods. */
//#endif
/* end of modification */
/* PARTITION Mani, Added flag for partitioned support */
#define DB_PARTITION 0x10000000
/* Using the next flag available, after the ones used for temporal support
 * This flag lets the user specify if he wants the db to be partitioned or not
 * Once a db is partitioned, it can be opened in that mode only
 * just like the idea of temporal, where in even if the user doesn't specifty
 * the temporal flag, the db will be opened with temporal flag on.
 */
/* Mani End */

#define CURRENT_YEAR 1
#define CURRENT_MONTH 2
#define CURRENT_DAY 3
#define CURRENT_HOUR 4
#define CURRENT_MINUTE 5
#define CURRENT_SECOND 6
#define CURRENT_MIC 7

/* New error code */
#define TEMPORAL_TUC_ERR (-33000)
#define TEMPORAL_CLK_CHECK (-33001)
#define TEMPORAL_CLK_PREC (-33002)
#define TEMPORAL_CLK_NL (-33003)

#define UNTILCHANGED 10         /* this value is never used as a flag,
				   therefore, we don't have to worry about
				   free values */

#define ID_ID   1       /* these constants define the state of the*/
#define ID_UC   2       /* particular record. Whether the start time*/
#define ID_TIME 3       /* is stamped, or a particular time is*/
#define TIME_TIME 4     /* UNTILCHANGED, or if both the start andstop*/
#define TIME_ID 5       /* times are stamped(should not occur) ... etc*/
#define TIME_UC 6
#define INVALID_STATE 7

#define FOREVER LONG_MAX       /* these values will be used by c_get_nonsequenced.*/
#define BEGINNING (LONG_MAX - 1)
#define US (LONG_MAX - 2)      /* US == this transaction, OTHER == some other */
#define OTHER (LONG_MAX - 3)   /* transaction, FOREVER == valid forever */
#define UC (LONG_MAX - 4)      /* Until changed. -Stanley yao (08/01/03) */


struct __ttdbc {
    /* new methods */
    int (*c_get_nonsequenced) (struct __ttdbc *ttdbc, DBT *key, DBT *data,
        struct timeval start, struct timeval stop, u_int32_t flags,
        struct timeval *outstart, struct timeval *outstop);

    /* member elements */
    struct __dbc *dbc;
    struct __ttdb *ttdb;

    /* PARTITION Mani, 13 April 2003, Added new member, a cursor pointer for the cursor which will run thru the archive db */
    struct __dbc *adbc;
    /* PARTITION Mani 25 May 2003, Added new member variables to help in cursor movement */
    char moveArchiveFlag, moveCurrentFlag;
    char firstArchiveRec, firstCurrentRec;
    /* Mani End */

    /* redefined methods */
    int (*c_del) (struct __ttdbc *dbc_arg, u_int32_t flags);
    int (*c_get) (struct __ttdbc *dbc_arg, DBT *key, DBT *data, u_int32_t flags);
    int (*c_put) (struct __ttdbc *dbc_arg, DBT *key, DBT *data, u_int32_t flags);
    int (*c_close) (struct __ttdbc *cursor);
};

struct __ttdb{
    struct  __db    *db;
    /* PARTITION Mani, 16th Mar 2003, Adding a db pointer, which will hold the pointer to the archive store */
    struct __db *adb;
    /* Mani End */

    /* old elements that are accessed from outside */
    void *app_private;      /* Application private handle */

    /* extra elements */
    struct timeval  current_time;   /* data extracted at this time*/

    /* new methods */
    int (*is_tt) (struct __ttdb *ttdb);
    int (*set_current) (struct __ttdb *db, struct timeval curr);
    struct timeval (*get_current) (struct __ttdb *ttdb);

    /* redefined methods*/

    int  (*key_range) (struct __ttdb *,DB_TXN *, DBT *, DB_KEY_RANGE *,
        u_int32_t);

    int  (*stat) (DB *, void *, void *(*)(size_t), u_int32_t);

    int  (*remove) (DB *, const char *, const char *, u_int32_t);

    int (*set_bt_compare) (struct __ttdb*, int (*)(struct __ttdb*, const DBT *,
        const DBT*));

#ifdef BDB_4
    int (*open) (struct __ttdb *ttdb, DB_TXN *tid, const char *file, const char *database,
            DBTYPE type, u_int32_t flags, int mode);
#endif
#ifdef BDB_3
    int (*open) (struct __ttdb *ttdb, const char *file, const char *database,
            DBTYPE type, u_int32_t flags, int mode);
#endif

    int (*put) (struct __ttdb *db, DB_TXN *txnid, DBT *key, DBT *data,
            u_int32_t flags);

    int (*get) (struct __ttdb *db, DB_TXN *txnid, DBT *key, DBT *data,
            u_int32_t flags);

    int (*cursor) (struct __ttdb *ttdb, DB_TXN *txnid, struct __ttdbc **cursorp,
            u_int32_t flags);

    int (*pget) (struct __ttdb *db, DB_TXN *txnid, DBT *key, DBT *pkey,
            DBT *data, u_int32_t flags);

    int (*del) (struct __ttdb *db, DB_TXN *txnid, DBT *key, u_int32_t flags);
    void (*err) (struct __ttdb *db, int error, const char *fmt, ...);
    int (*close) (struct __ttdb *ttdb, u_int32_t flags);
    int (*set_flags) (struct __ttdb *db, u_int32_t flags);
    void (*set_errfile) (struct __ttdb *db, FILE *errfile);
    void (*set_errpfx) (struct __ttdb *db, const char *errpfx);
    int (*set_pagesize) (struct __ttdb *db, u_int32_t pagesize);
    int (*set_cachesize) (struct __ttdb *db, u_int32_t gbytes, u_int32_t bytes, int
        ncache);
};
typedef struct __ttdb TTDB;
typedef struct __ttdbc TTDBC;

/* the data structure has to be changed in order to make space for the start
 * and stop times. note the union that we have specified in case of the start
 * and stop times.
 typedef union{
 struct timeval   curr;
 u_int32_t        stamperId;
 }IdOrTime;
*/

/* Modification of Huilong at 15:06 7/11/2006 */
struct __env_id_type {
    u_int32_t   tv_sec;
    u_int16_t   tv_msec;
};
typedef struct __env_id_type  ENV_ID_TYPE;
struct __stp_id {
    ENV_ID_TYPE envid;
    u_int32_t   txnid;
};
/*struct __stp_id {
    time_t		envid;
    u_int32_t	txnid;
};*/
/* end of modification */

typedef struct __stp_id STP_ID;

typedef union{
    struct timeval   curr;
    STP_ID  stamperId;
} IdOrTime;

struct __tt_wrapper{
    int         state;
    IdOrTime    start;
/* Added by Huilong on 7/13/2006 */
#ifdef TBDB_AUDIT
    int         start_seq; /* for audit */     
#endif
/* End of modification */
    IdOrTime    stop;
/* Added by Huilong on 7/13/2006 */
#ifdef TBDB_AUDIT
    int         stop_seq; /* for audit */
#endif
/* End of modification */
};
typedef struct __tt_wrapper TTDATA;

/* Added by Huilong (from Shilong's code) on 7/13/2006*/
/* Check the state of a tuple
 * Tuple state (32 bits):
 * 0-3 bit: state (ID_UC, TIME_UC, TIME_ID, TIME_TIME)
 * 4 bit: start_time need hashing in AUD or not?
 * 5 bit: stop_time need hashing in AUD or not?
 * 6 bit: tuple is TEMPORAL_AUD or not?
 */
#ifdef TBDB_AUDIT
#define TTDB_DBT_STATE(state) ((state) & 0x0F)
#define TTDB_DBT_SETSTATE(state, value) state = (((state) & 0xFFF0) | (value))

#define TTDB_DBT_HASH1(state) (((state) >> 4) & 0x01)
#define TTDB_DBT_SETHASH1(state) state |= 0x010;
#define TTDB_DBT_CLRHASH1(state) state &= (~0x010)

#define TTDB_DBT_HASH2(state) (((state) >> 5) & 0x01)
#define TTDB_DBT_SETHASH2(state) state |= 0x020;
#define TTDB_DBT_CLRHASH2(state) state &= (~0x020)

#define TTDB_DBT_AUD(state) ((state) & 0x040)
#define TTDB_DBT_SETAUD(state) state |= 0x040
#define TTDB_DBT_CLRAUD(state) state &= (~0x040)
#else
#define TTDB_DBT_STATE(state) (state)
#define TTDB_DBT_SETSTATE(state, value) state = (value)
#endif
/* end of modification */

/* All API call signitures are listed below.
 * All internal function call signitures are moved to
 *   internal header files: ttdb_ext.h, etc.
 * Stanley Yao (06/24/03)
 */
#ifdef __cplusplus
extern "C" {
#endif
int __ttdb_c_get_nonsequenced(struct __ttdbc *ttdbc, DBT *key, DBT *data,
			      struct timeval start, struct timeval stop, u_int32_t flags,
			      struct timeval *outstart, struct timeval *outstop);

/* TTDB interface */
int is_tt(struct __ttdb *ttdb);
/* PARTITION Mani - 9th Aug 2003 - Added function decl. to interface */
int is_partitioned (struct __ttdb *ttdb);
/* Mani End */
int is_until_changed(struct timeval time_value);
int is_us(struct timeval time_value);
int is_other(struct timeval time_value);
int __ttdb_create(struct __ttdb **ttdb, DB_ENV *env, u_int32_t flags);

/* CLK interface */
int set_nesting_level(DB_ENV *dbenv, int lvflag);
int env_current(struct timeval *time, int precision);
int txn_current(DB_TXN *tid, struct timeval *time, int precision);

/* Modified by Huilong to eliminate the unnecessary headers in forensic analysis programs */
int __aud_get_time_hash(char *path, char *dbname, char *hash_value, struct timeval *ttime, struct timeval *cur_time);
#define AUD_HASH_LEN 20
struct __aud_txn_hash_data {
    char prev_hash_value[AUD_HASH_LEN];
    struct timeval time;
    char hash_value[AUD_HASH_LEN];
};
typedef struct __aud_txn_hash_data AUD_TXN_HASH_DATA;
struct __nid {
   struct timeval ntime;
   char nid[AUD_HASH_LEN];
};
typedef struct __nid NID;

#ifndef HEAP_H_INCLUDED
#define HEAP_H_INCLUDED

extern char const Heap_rcsid[];

#define T Heap_T
typedef struct T *T;

typedef struct HeapItem_T *HI_T;

typedef enum {
  HEAP_OKAY = 0,        /*  all is well                     */
  HEAP_ERR_OVFL,        /*  insert failed                   */
  HEAP_ERR_BOUNDARY,    /*  internal inconsistency          */
  HEAP_ERR_NOTHEAP,     /*  heap property violated          */
  HEAP_ERR_REF          /*  referential integrity violated  */
} HeapErr_T;

extern T
Heap_new(const int init_size, int cmp(const void *a, const void *b),
         HI_T *(*self)(void *p));

extern HeapErr_T
Heap_insert(T heap, void *p);

void *
Heap_top(const T heap);

void *
Heap_extract_min(T heap);

void
Heap_check(T heap, void func(HeapErr_T err));

int
Heap_size(T heap);

void
Heap_free(T *pheap);

int
Heap_member(const T heap, void *p);

void *
Heap_remove(T heap, void *p);

void
Heap_update(T heap, void *p);

#undef T
#endif

/* end of modification */

#ifdef __cplusplus
}
#endif

#endif /* !_TTDB_H_ */
