/*-
 * Developed (2002-2003) by:
 *     TAU Project
 *     Computer Science Dept., University of Arizona
 *     Tucson, AZ 85721, USA
 *     URL  : http://www.cs.arizona.edu/tau
 *     Email: tau@cs.arizona.edu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef DB_TEMPORAL
#define DB_TEMPORAL  0x020000000
#endif

#ifndef CLK_TXN_INC
#define CLK_TXN_INC

#include <sys/types.h>
#include <sys/stat.h>

#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "db.h"
#include "db_int.h"

#undef DB
#undef DBC
#undef db_create

#define BDB_3_2_9

/*#define LONG_MAX 2147483647L*/
#define ENTRIES_OF_TABLE 1024
/*#define CLK_ERR -1   ---> define this in db.h*/
#define FXNOK 0
/*#define PRE_ERR -2   ---> define this in db.h*/
#define CLK_REG_SIZE                  6553500


/* This is created to hold the proper info for the rtt and wtt table which 
is treated
* as a table in database, and these info is per-ENV
* Thus, we need to put them into the shared region to support
* multi-process
*/

/* struct timeval rtt_table[ENTRIES_OF_TABLE];
* struct timeval wtt_table[ENTRIES_OF_TABLE];
* int NESTED_LEVEL;
*/

struct __clk_share {
	/*
	 * These fields need to be protected for multi-threaded support.
	 *
	 * !!!
	 * As this structure is allocated in per-process memory, the mutex may need
	 * to be stored elsewhere on architectures unable to support mutexes in 
heap
	 * memory, e.g., HP/UX 9.
	 */
#ifdef BDB_4
	DB_MUTEX
#endif
#ifdef BDB_3
	MUTEX
#endif
	*mutexp;		/* Mutex for thread protection. */

  struct timeval *rtt_table_ptr;
  struct timeval *wtt_table_ptr;


	/* These fields are not thread-protected. */
	DB_ENV			*dbenv;			/* Reference to error information. */
	REGINFO			reginfo;		/* Underlying cache regions. */
	u_int32_t		flags;
};
typedef struct __clk_share CLK_SHARE;

/*
* STP --
*	Shared stamper region.  One of these is allocated in shared memory,
*	and describes the stamper.
*/
struct __clk_share_node {

	roff_t		wtt_off;			/* Offset of TT */
	roff_t		rtt_off;			/* Offset of TL */

#ifdef BDB_4
	DB_MUTEX
#endif
#ifdef BDB_3
	MUTEX
#endif
	*mutexp;                /* Mutex for thread protection. */

#ifdef MUTEX_SYSTEM_RESOURCES
#define	CLK_MAINT_SIZE	(sizeof(roff_t) * DB_MAX_HANDLES)

	roff_t		maint_off;		/* offset of region maintenance info */
#endif
  	u_int32_t	flags; 		/* Environment-wide flag values. */
};
typedef struct __clk_share_node SHARE_NODE;


/* For each transaction, we will need to allocate a piece of memory to store 
   the following data
   In this way, the callers who got the pointers to this piece of data can 
   easily get the info
   in the structure, or pass it as the parameter for the callee to modify.*/

struct clk_txndata {
  struct timeval t_l;/*This is the earliest possible txntime*/
  struct timeval t_h;/*This is the latest possible txntime*/
  struct timeval t_d; /* timestamp value for data item d */
  struct timeval t_x; /* timestamp that is finally assigned to this transaction */

  struct timeval t_s; /* at transaction befin, we set start time to be NOW */
  int precision;

  char *V_R;
  char *V_W;
  DB_TXN *tid;
  DB_ENV *env;
  struct clk_txndata *next;
};

typedef struct clk_txndata CLK_TXNDATA;

/* commented by: Shilong Yao
 * This will cause multi definition of clk_header, when this
 *	header file is included in more than one module.
 * It is moved to clk_txn.c
 * - 10/17/02
 
 CLK_TXNDATA *clk_header = NULL;
 */

/*this is to find out the time of the rtt entry & wtt entry*/
struct timeval clk_rtt_lookup (DB_ENV *,int);
struct timeval clk_wtt_lookup (DB_ENV *,int);

/*This will be called by the txn_begin to init the data*/
CLK_TXNDATA* clk_init ();

/*This will be called at the commit time to update the rtt and wtt*/
int clk_rtt_update (CLK_TXNDATA *txndata);
int clk_wtt_update (CLK_TXNDATA *txndata);
int clk_hash (DBT *data);
int clk_bitmanage (char* array_rec, int bitpos);

/*The following functions are for the operations of a global linklist*/
int clk_insert_data (CLK_TXNDATA *insert_data);
int clk_delete_data (DB_TXN *tid);
int clk_search_data (CLK_TXNDATA **curr_data, DB_TXN *tid);


/*This part is the functions that is currently only called by the clock module and
 *    it might be called by others some time later*/
int env_current ( struct timeval *time, int precision );

/*This part of functions will be called by the other modules
 *   It will set a specific time to be the transaction time.
 *   Right now, we choose the lower bound(the earliest time) to be
 *   the timestamp when it is called
 *   The timestamp will be available in the txn_data,
 *   and 1 is returned when it is successful
 */

int clk_txn_begin (DB_ENV *env,DB_TXN *parent, DB_TXN **tid, u_int32_t flags);
int txn_current (DB_TXN *tid,  struct timeval *time, int precision);

  /* When we access the data during the transaction, it will be either read, 
   * delete, insert
   * In each of the following function,it will take care of one of the case.
   * It will check up the transaction timestamp and the RTT table, 
   * whenever there is a confliction, it will return -1
   * Otherwise, it will return 1 to show that it is a success
   * Then the clock_bounds will also be changed by the 
   * pointer passed as a parameter
   */
int clk_check (DB_TXN *dbc,DB* dbpp, db_pgno_t pgno, int mode,DBT* data, 
	       u_int32_t flag);

  /*The following function will be called at commit time.
   *   It will do the rtt update, and the loging process */

int clk_txn_commit (DB_ENV *dbenv, DB_TXN *tid);

int clk_txn_abort(DB_TXN *tid);

int propagate(CLK_TXNDATA *txn_data);

int __share_init(DB_ENV *, CLK_SHARE *);
int __clk_open(DB_ENV *);
int setnstlv(DB_ENV *, int );

#endif

#ifndef timerisset
#define timerisset(tvp)         ((tvp)->tv_sec || (tvp)->tv_usec)
#define timercmp(tvp, uvp, cmp) \
        (((tvp)->tv_sec == (uvp)->tv_sec) ? \
            /* CSTYLED */ \
            ((tvp)->tv_usec cmp (uvp)->tv_usec) : \
            /* CSTYLED */ \
            ((tvp)->tv_sec cmp (uvp)->tv_sec))

#define timerclear(tvp)         (tvp)->tv_sec = (tvp)->tv_usec = 0
#endif


