/*-
 * Developed (2002-2003) by:
 *     TAU Project
 *     Computer Science Dept., University of Arizona
 *     Tucson, AZ 85721, USA
 *     URL  : http://www.cs.arizona.edu/tau
 *     Email: tau@cs.arizona.edu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "db_config.h"
#include "db_int.h"
#include "db_shash.h"
#include "txn.h"
#include "lock.h"
#include "log.h"
#include "db_dispatch.h"
#include "db_page.h"
#ifdef BDB_3
#include "db_ext.h"
#endif
#include "TTDB.h"
#include "db.h"
#include "stp_int.h"
#include "clk_txn.h"
/* Huilong 7/16/06 */
#include "mp.h"
#ifdef TBDB_AUDIT
#include "../stp/aud_int.h"
#endif
/* end */

#include "temporal.h"

CLK_TXNDATA *clk_header = NULL;
unsigned char   usedmask[8] = { 1, 2, 4, 8, 16, 32, 64, 128 };

#define MYSQL_LEVEL 1
#define NESTED_LEVEL 2
#define PLAIN_LEVEL 3
#define NOVALUE_LEVEL 10

#ifndef MAX_LEVEL_VAR
#define MAX_LEVEL_VAR 7
#endif

void
__set_nstlv(DB_ENV *dbenv)
{
    /* We now setup the NESTLEVEL code. As at this step, we have lost the chance to change
     * this unique value in the environment. I am going to set this value
     * to be a non-changable value when transaction begins
     */
    if(((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->NESTLEVEL < MAX_LEVEL_VAR)
	((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->NESTLEVEL = MAX_LEVEL_VAR;
}

/* After env->open, this value should be 0.
 * set_nesting_level() should be called between env->open and db->open. Otherwise
 *   the nest level remains 0 until db->open().
 * db->open will set this value to default value 1, if set_nesting_level()
 *   hasn't been called by that time.
 * After db->open, the nesting level must NOT be 0. If the set_nesting_level() is
 *   called after db->open, we should return error code.
 */
int set_nesting_level(DB_ENV *dbenv, int lvflag){
    if(((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->NESTLEVEL == 0)
	((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->NESTLEVEL = lvflag;
    else
	return TEMPORAL_CLK_NL;
}

int clk_insert_data(CLK_TXNDATA *insert_data){

    /*check the clk_header is NULL or not*/
    insert_data->next = clk_header;
    clk_header = insert_data;
    return 0;
}

int clk_delete_data (DB_TXN *tid) {

    CLK_TXNDATA* next_data, *temp;	/* is this correct */

    /* check the clk_header is NULL or not */
    if (clk_header==NULL) return -1;

    next_data = clk_header;
    if (next_data->tid == tid) {
	clk_header = next_data->next;
	free(next_data->V_R);
	free(next_data->V_W);
	free(next_data);
	return 0;
    }
    else {
	while ( (next_data->next)!=NULL ) {
	    temp = (next_data->next);
	    if ( (temp->tid) == tid) {
		next_data->next = temp->next;
		free(temp->V_R);
		free(temp->V_W);
		free(temp);
		return 0;
	    }
	    next_data = temp; /* Shilong */
	}
    }
    /* This point should never be reached */
    return -1;
}

int clk_search_data (CLK_TXNDATA **curr_data, DB_TXN *tid) {

    CLK_TXNDATA* next_data;

    /*check the clk_header is NULL or not*/
    if (clk_header==NULL) {
	return -1;
    }
    /*printf("DEBUG: in search, the header is not null\n");*/
    next_data = clk_header;
    if ( next_data->tid == tid) {
	*curr_data = next_data;
	return 0;
    }
    else {
	while( next_data->next!=NULL) {
	    next_data = next_data->next;
	    if ( next_data->tid == tid) {
		*curr_data = next_data;
		return 0;
	    }
	}
	/*printf("DEBUG: I should not reach here\n");*/
	return -1;		/* there is no matching list */
    }
}

int env_current ( struct timeval *time, int precision ) {
    struct timeval mytime;
    struct tm *tm_ptr;
    struct tm tm_store;
    int errno;

    /* Modify according to the document, Stanley Yao (08/04/03) */
    if ((precision <= 0) || (precision > 7))
	precision = CURRENT_SECOND;

    gettimeofday(&mytime, NULL);
    /* Stanley yao, 08/25/03
    if (precision == 0) {
	time->tv_sec = 0;
	time->tv_usec = 0;
	return FXNOK;
    }
    */
    tm_ptr = localtime_r(&(mytime.tv_sec), &tm_store);
    tm_ptr = &tm_store;
    switch(precision) {
    /* Stanley Yao, 08/25/03
    case 0: *This is for all_time*
	tm_ptr->tm_year = 0;
    */
    case CURRENT_YEAR: /*This is for year*/
	tm_ptr->tm_mon = 0;
    case CURRENT_MONTH: /*This is for month*/
	tm_ptr->tm_mday = 1;
    case CURRENT_DAY: /*This is for day*/
	tm_ptr->tm_hour = 0;
    case CURRENT_HOUR: /*This is for hour*/
	tm_ptr->tm_min = 0;
    case CURRENT_MINUTE: /*This is for minute*/
	tm_ptr->tm_sec = 0;
    case CURRENT_SECOND: /*This is for the second*/
	mytime.tv_usec = 0;
    case CURRENT_MIC: /*This is for the microsecond*/
	/*I just simply do nothing*/
	break;
	/*That is the end of all cases*/
    default:
	printf("This is not an accepted granularity!\n");
	printf("Now the precision is set to microsecond(default)\n");
	break;
	/*Defualt: I do nothing here*/
    }
    mytime.tv_sec = mktime(tm_ptr);
    /* Stanley Yao, 08/25/03
    if(precision == 0) {
	mytime.tv_sec = 0;
	mytime.tv_usec = 0;
    }
    */
    time->tv_sec = mytime.tv_sec;
    time->tv_usec = mytime.tv_usec;
    /* memcpy(time, &mytime, sizeof(struct timeval));*/
    return FXNOK; /*time got from the system and round it up to the first instance of that range*/
}

/* Here are the AUX functions for the calculation of "current function"
 * No return value is needed as we use the pointers to pass back the value
 * we have calculated
 */

void clk_floor( struct timeval *timeout, struct timeval *timein, int precision) {
    struct tm *tm_ptr;
    struct tm tm_store;

    /* Modify according to the document, Stanley Yao (08/04/03) */
    if ((precision <= 0) || (precision > 7))
	precision = CURRENT_SECOND;

    if(precision >= 7) {
	/* The value will not be changed at all */
	timeout->tv_sec = timein->tv_sec;
	timeout->tv_usec = timein->tv_usec;
	return;
    }
    tm_ptr = localtime_r(&(timein->tv_sec), &tm_store);
    tm_ptr = &tm_store;
    switch(precision) {
    /* Stanley Yao, 08/25/03
    case 0: *This is for all_time*
	tm_ptr->tm_year = 0;
    */
	/* Notice here that we set the last possible time to be year 2011 */
    case CURRENT_YEAR: /*This is for year*/
	tm_ptr->tm_mon = 0;
    case CURRENT_MONTH: /*This is for month*/
        /* printf("DEBUG: THE MON IS %d\n", tm_ptr->tm_mon); */
	tm_ptr->tm_mday = 1;
    case CURRENT_DAY: /*This is for day*/
	tm_ptr->tm_hour = 0;
    case CURRENT_HOUR: /*This is for hour*/
	tm_ptr->tm_min = 0;
    case CURRENT_MINUTE: /*This is for minute*/
	tm_ptr->tm_sec = 0;
    case CURRENT_SECOND: /*This is for the second*/
	timeout->tv_usec = 0;
	break;
    }
    timeout->tv_sec = mktime(tm_ptr);
    return;
}

void granule( struct timeval *timeout, struct timeval *timein, int precision, int flag) {
    struct timeval temptime;
    struct tm *tm_ptr;
    struct tm tm_store;
    int sign;

    temptime.tv_sec = timein->tv_sec;
    temptime.tv_usec = timein->tv_usec;
    if(flag == 0) {
	sign = 1;
    }
    else if(flag == 1) {
	sign = -1;
    }
    else {
        /* printf("Wrong flag for the granule function\n"); */
	timeout->tv_sec = timeout->tv_usec = 0;
    }

    /* Modify according to the document, Stanley Yao (08/04/03) */
    if ((precision <= 0) || (precision > 7))
	precision = CURRENT_SECOND;

    tm_ptr = localtime_r(&(temptime.tv_sec), &tm_store);
    tm_ptr = &tm_store;
  ONCEMORE:
    switch(precision) {
    /* Stanley Yao, 08/25/03
    case 0: {
	* This is already forever,there is no need to add or minus *
	tm_ptr->tm_year = 1;
	tm_ptr->tm_mon = 0;
	tm_ptr->tm_mday = 1;
	tm_ptr->tm_min = 0;
	tm_ptr->tm_sec = 0;
	temptime.tv_usec = 0;
	break;
    }
    */
    case CURRENT_YEAR: {
	/* This is for the granule of YEAR */
	tm_ptr->tm_year += sign;
	break;
    }

    case CURRENT_MONTH: {
	/* This is for the granule of Mon */
	if(flag == 0) {
	    if(tm_ptr->tm_mon <11) {
		tm_ptr->tm_mon ++;
		break;
	    }
	    else {
		tm_ptr->tm_year ++;
		tm_ptr->tm_mon = 0;
		break;
	    }
	}
	else {
	    /* Now the flag must be 1 */
	    if(tm_ptr->tm_mon > 1) {
		tm_ptr->tm_mon --;
		break;
	    }
	    else {
		tm_ptr->tm_year--;
		tm_ptr->tm_mon = 0;
		if(tm_ptr->tm_year <=0) {
		    tm_ptr->tm_year = 1;
		    tm_ptr->tm_mon = 0;
		    tm_ptr->tm_mday = 1;
		    tm_ptr->tm_min = 0;
		    tm_ptr->tm_sec = 0;
		    temptime.tv_usec = 0;
		}
		break;
	    }
	}
    }

    case CURRENT_DAY: {
	/* This is for the granule of Day */
	if(flag == 0) {
	    switch(tm_ptr->tm_mon){
	    case 0:
	    case 2:
	    case 4:
	    case 6:
	    case 7:
	    case 9:
	    case 11:
		/*These are the mon of 31 days*/
	    {
		if(tm_ptr->tm_mday <31) {
		    tm_ptr->tm_mday ++;
		}
		else {
		    tm_ptr->tm_mday = 1;
		    precision--;
		    goto ONCEMORE;
		}
		break;
	    }

	    case 3:
	    case 5:
	    case 8:
	    case 10: {
		/* These are the mon of 30 days */
		if(tm_ptr->tm_mday < 30) {
		    tm_ptr->tm_mday ++;
		    break;
		}
		else {
		    tm_ptr->tm_mday = 1;
		    precision --;
		    goto ONCEMORE;
		}
	    }
	    case 1: {
		if(tm_ptr->tm_mday < 28) {
		    tm_ptr->tm_mday ++;
		    break;
		}
		else {
		    tm_ptr->tm_mday = 1;
		    precision--;
		    goto ONCEMORE;
		}
	    }
	    }
	}
	else {
	    /* Flag must be 1 now */
	    switch(tm_ptr->tm_mon){
	    case 0:
	    case 1:
	    case 3:
	    case 5:
	    case 7:
	    case 8:
	    case 10:
		/*These are the mon that pre mon is of 31 days*/
	    {
		if(tm_ptr->tm_mday >1) {
		    tm_ptr->tm_mday --;
		}
		else {
		    tm_ptr->tm_mday = 31;
		    precision--;
		    goto ONCEMORE;
		}
		break;
	    }

	    case 4:
	    case 6:
	    case 9:
	    case 11: {
		/* These are the mon that pre mon is of 30 days */
		if(tm_ptr->tm_mday > 1) {
		    tm_ptr->tm_mday --;
		    break;
		}
		else {
		    tm_ptr->tm_mday = 30;
		    precision --;
		    goto ONCEMORE;
		}
	    }
	    case 2: {
		if(tm_ptr->tm_mday > 1) {
		    tm_ptr->tm_mday --;
		    break;
		}
		else {
		    tm_ptr->tm_mday = 28;
		    precision--;
		    goto ONCEMORE;
		}
	    }
	    }
	}
	break;
    }

    case CURRENT_HOUR: {
	/*This is for the granule of hour */
	if(flag == 0) {
	    /* We add one hour */
	    if(tm_ptr->tm_hour <23) {
		tm_ptr->tm_hour ++;
		break;
	    }
	    else {
		/* It is now 23:00 */
		tm_ptr->tm_hour = 0;
		precision --;
		goto ONCEMORE;
	    }
	}
	else {
	    /* We minus one hour */
	    if(tm_ptr->tm_hour > 0) {
		tm_ptr->tm_hour --;
		break;
	    }
	    else {
		/* It is now 0:00 */
		tm_ptr->tm_hour = 23;
		precision --;
		goto ONCEMORE;
	    }
	}
    }

    case CURRENT_MINUTE: {
	/* This is for the granule of minute */
	if(flag == 0) {
	    /* We add one minute */
	    if(tm_ptr->tm_min < 59) {
		tm_ptr->tm_min++;
		break;
	    }
	    else {
		tm_ptr->tm_min = 0;
		precision--;
		goto ONCEMORE;
	    }
	}
	else {
	    if(tm_ptr->tm_min > 0) {
		tm_ptr->tm_min--;
		break;
	    }
	    else {
		tm_ptr->tm_min = 59;
		precision --;
		goto ONCEMORE;
	    }
	}
    }

    case CURRENT_SECOND: {
	/* This is for the granule of second */
	if(flag == 0) {
	    if(tm_ptr->tm_sec < 59) {
		tm_ptr->tm_sec ++;
		break;
	    }
	    else {
		tm_ptr->tm_sec = 0;
		precision--;
		goto ONCEMORE;
	    }
	}
	else {
	    if(tm_ptr->tm_sec > 0) {
		tm_ptr->tm_sec-- ;
		break;
	    }
	    else {
		tm_ptr->tm_sec = 59;
		precision--;
		goto ONCEMORE;
	    }
	}
    }

    case CURRENT_MIC: {
	/* This is for the granule of microsecond */
	if (flag == 0) {
	    if(temptime.tv_usec < 999999) {
		temptime.tv_usec ++;;
		break;
	    }
	    else {
		temptime.tv_usec = 0;
		precision --;
		goto ONCEMORE;
	    }
	}
	else {
	    if(temptime.tv_usec > 0) {
		temptime.tv_usec--;
		break;
	    }
	    else {
		temptime.tv_usec = 999999;
		precision --;
		goto ONCEMORE;
	    }
	}
    }
    }
    timeout->tv_sec = mktime(tm_ptr);
    timeout->tv_usec = temptime.tv_usec;
    return;
}


int txn_current (DB_TXN *tid, struct timeval *time, int precision) {
    CLK_TXNDATA *txn_data, *parent_data;
    struct timeval mytime;
    struct timeval temp_time1, temp_time2;
    struct tm *tm_ptr;
    DB_TXN *ulti_parent;
    int nothing;
    DB_TXN *temptid;
    /* Now I need to find out the next to highest level parent,
     * All propagation stops at the second level of the nested txn
     */

    txn_data = NULL;
    ulti_parent = tid;

    /* Modify according to the document, Stanley Yao (08/04/03) */
    if ((precision <= 0) || (precision > 7))
	precision = CURRENT_SECOND;

    /*  printf("DEBUG: I am starting the txn_current, precision is %d\n", precision);
     */
    nothing = clk_search_data(&txn_data, tid);
    /*  printf("DEBUG: I am after search list, %d\n", nothing);
     */
    if(((TEMPORAL_HANDLE *)(txn_data->env->temporal_handle))->NESTLEVEL == MYSQL_LEVEL ||
       ((TEMPORAL_HANDLE *)(txn_data->env->temporal_handle))->NESTLEVEL == PLAIN_LEVEL) {
	/* In this case, I do not need to change the information */
    }
    else if(((TEMPORAL_HANDLE *)(txn_data->env->temporal_handle))->NESTLEVEL == NESTED_LEVEL) {
	nothing = clk_search_data(&parent_data, tid->parent);
    }
    if( ((txn_data->t_x).tv_sec != 0) && (txn_data->precision == precision)  ) {
	/* We have already set up the time stamp for the current parent transaction*/
	/*    printf("DEBUG: What's wrong here\n");
	 */
	time->tv_sec = (txn_data->t_x).tv_sec;
	time->tv_usec = (txn_data->t_x).tv_usec;
	/*printf("DEBUG:Have tmp, no need to change in txn_cur\n");
	 */
	return FXNOK;
    }

    /* So now, either the time is not setted, or the precision need to be changed */
    /*  printf("DEBUG: I am here somewhere\n");
     */
    /* If we just want to change the precision, only the more precise one be accepted */
    /* Stanley Yao, 08/28/03, Comment the code below ---
    if (precision < txn_data->precision) {
	* The precision order is wrong *
	*    printf("DEBUG:The precision misorder\n");
	*
	return TEMPORAL_CLK_PREC;
    }
    */

    /* Stanley Yao, 08/25/03
    if(precision == 0) {
	* This transaction time do not need to be set as it is valid for all time *
	(txn_data->t_l).tv_sec = (txn_data->t_x).tv_sec = time->tv_sec = 0;
	(txn_data->t_l).tv_usec = (txn_data->t_x).tv_usec = time->tv_usec = 0;
	return FXNOK;
    }
    */
    gettimeofday(&temp_time2, NULL);
    /*  printf("The cur sec %ld, sec time1%ld, %d\n",
	temp_time2.tv_sec, temp_time1.tv_sec, precision);
     */
    clk_floor( &temp_time1, &temp_time2, precision);
    /*  printf("DEBUG: After the clk_floor, sec is %ld\n", temp_time1.tv_sec);
     */
    granule( &temp_time2,&temp_time1, precision, 0);
    /*  printf("DEBUG: After 1st granule in txn_cur, sec is %ld\n", temp_time2.tv_sec );
     */

    mytime.tv_usec = temp_time2.tv_usec;
    mytime.tv_sec = temp_time2.tv_sec;
    if(timercmp((&mytime),(&(txn_data->t_h)),<)) {
	/*    printf("DEBUG: mytime < t_h\n");
	 */
	(txn_data->t_h).tv_sec = mytime.tv_sec;
	(txn_data->t_h).tv_usec = mytime.tv_usec;
    }
    granule(&temp_time1, &(txn_data->t_h), precision, 1);
    /*  printf("DEBUG: The tv_sec is now %ld\n\n\n\n", temp_time1.tv_sec);
     */
    mytime.tv_sec = temp_time1.tv_sec;
    mytime.tv_usec = temp_time1.tv_usec;

    if (timercmp (&mytime, &(txn_data->t_l), >) ) {
	/*    printf("DEBUG: mytime>t_l\n");
	 */
	(txn_data->t_l).tv_sec = mytime.tv_sec;
	(txn_data->t_l).tv_usec = mytime.tv_usec;
    }
    /*  printf("DEBUG: The sec of t_l is %ld, %ld\n",
	(txn_data->t_l).tv_sec, (txn_data->t_h).tv_sec);
     */
    if( timercmp((&(txn_data->t_l)), (&(txn_data->t_h)), > )) {
	/*    printf("DEBUG: In txn_cur, need to abort txn\n");
	 */
	/*((DB_TXN*)tid)->abort((DB_TXN*)tid);*/
	/*MINGDE*/
	/* Not sure if it is correct */
	/*txn_abort(txn_data->tid);*/
	/* Shilong 05/27/03
	 * (1) When conflict found, we just return error msg and
	 *     let the user to choose whether abort or not.
	 * (2) In BDB 4.1.xxx, __txn_end() is static!!!
	 *
	 __txn_end(txn_data->tid, 0);
	 */
	return TEMPORAL_CLK_CHECK;
    }


    txn_data->precision = precision;
    /*  printf("DEBUG: sec of mytime is %ld \n", mytime.tv_sec);
     */
    clk_floor(time,&mytime, precision);
    (txn_data->t_x).tv_sec = time->tv_sec;
    (txn_data->t_x).tv_usec = time->tv_usec;
    /*  printf("DEBUG: End of the txn_current, sec is %ld\n", (txn_data->t_x).tv_sec);
     */

    return FXNOK;
}


/* this function returns error no if there is an error.
 *
 */
int clk_txn_begin(DB_ENV *env,DB_TXN *parent, DB_TXN **tid, u_int32_t flags) {
    int errorno;
    int val;
    int ret;
    CLK_TXNDATA *dataptr, *tempp;

    /* printf("DEBUG: I am entering the clk_txn_begin\n");*/
    dataptr = clk_init();
    dataptr->env = env;
    dataptr->tid = *tid;
    gettimeofday(&(dataptr->t_s), NULL);
    ret =clk_search_data(&tempp,dataptr->tid);
    if(ret == 0) {
	clk_delete_data(dataptr->tid);
	clk_insert_data(dataptr);
    }
    else
	clk_insert_data(dataptr);

     /*printf("DEBUG: I am at the end of clk_txn_begin\n"); */
    __stp_txn_begin(env,(*tid)->txnid);
     /*printf("(DEBUG: I am leaving the clk_txn_begin \n");*/
    return 0;
}

CLK_TXNDATA* clk_init () {
    int num_of_char;
    CLK_TXNDATA *data;
    data = (CLK_TXNDATA*)malloc(sizeof(CLK_TXNDATA));
    num_of_char = ENTRIES_OF_TABLE/sizeof(char)+1;
    (data->t_l).tv_sec  = (data->t_l).tv_usec = 0;
    env_current(&(data->t_s), 7);
    /*NO: IS is a LONG_MAX or -1*/
    (data->t_h).tv_sec = LONG_MAX;
    (data->t_h).tv_usec = 999999;
    (data->t_x).tv_sec = (data->t_x).tv_usec = 0;
    data->V_R = (char*)malloc(num_of_char);
    data->V_W = (char*)malloc(num_of_char);
    memset(data->V_R, 0, num_of_char);
    memset(data->V_W, 0, num_of_char);
    data->tid = NULL;
    data->env = NULL;
    data->next = NULL;

    /* printf("DEBUG: THis is the end of init\n"); */
    return data;
}

struct timeval clk_rtt_lookup(DB_ENV *env,int position) {

    CLK_SHARE *cshare;
    cshare = (CLK_SHARE*)(((TEMPORAL_HANDLE *)(env->temporal_handle))->clk_handler);
    return(cshare->rtt_table_ptr[position]);
    /*return(rtt_table[position]);*/

}

struct timeval clk_wtt_lookup(DB_ENV *env,int position) {
    CLK_SHARE *cshare;
    cshare = (CLK_SHARE*)(((TEMPORAL_HANDLE *)(env->temporal_handle))->clk_handler);
    /* printf("DEBUG: WTT: position is %d\n",position);
       printf("DEBUG: WTT: the content is %ld(%ld)\n",
       ((cshare->wtt_table_ptr)[position]).tv_sec,((cshare->wtt_table_ptr)[position]).tv_usec);
    */
    return(cshare->wtt_table_ptr[position]);
    /*return(wtt_table[position]);*/
}

int clk_check (DB_TXN *tidin, DB *dbpp,db_pgno_t pgno, int mode, DBT* data,u_int32_t flag){

    DB_ENV *dbenv;
    CLK_SHARE *cshare;
    DB_TXN *ulti_parent, *tid;
    int ret;
    u_int32_t idp;
    CLK_TXNDATA* txn_data = NULL;
    struct timeval read_time;

    struct timeval start_time;
    struct timeval temp;
    TTDATA ttdata;
    int state;
    int position;
    struct timeval *comp1, *comp2;
    /*  printf("DEBUG: Enter CLK_CHECK\n");
     */
    dbenv = dbpp->dbenv;
    tid = tidin;
    cshare = (CLK_SHARE*)(((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->clk_handler);
    position = pgno%ENTRIES_OF_TABLE;
    read_time.tv_sec = start_time.tv_sec = 0;
    read_time.tv_usec = start_time.tv_usec = 0;
    /*printf("DEBUG:38 data=%d, data->data\n", data, data->data);
     */
    ret = clk_search_data(&txn_data, tid);

    /*printf("DEBUG:40 The returned value in clk_check_search is %d\n",ret);
     */
    if(ret != 0) {
	/* If I can not find the txndata I need, then it is for sure a lock required outside
	 * transaction, can be a lock requested by db_open to open a db for example,
	 * and I should not fill any of the information which is only usable for the
	 * transaction control
	 printf("DEBUG: not really a data lock\n");
	*/
	return 0;
    }

    /*For the modification of WTT: Mingde :CLK*/
    if( (dbpp->flags)& DB_TEMPORAL){
	/*This is a temporal database*/
	if(mode == 0){
	    /*	    printf("DEBUG: check tep read\n");
	     */
	    if(data->data == NULL) printf("NOTHING IS IN DATA\n");

	    memcpy((char*)(&ttdata),(char*)(data->data),sizeof(TTDATA));
	    /*      printf("DEBUG: 60 The ttdata.state is %d\n",ttdata.state);
	     */
	    switch(TTDB_DBT_STATE(ttdata.state)) {
	    case 1:
	    case 2:
	    case 3:
		/*	      printf("DEBUG:35 before R_LOCK\n");
		 */
		R_LOCK(dbenv,&(cshare->reginfo));
		/*      printf("DEBUG: 36 get the lock\n");
		 */
		start_time = clk_wtt_lookup(dbenv,position);
		/*      printf("DEBUG: 38 try to unlease the lock\n");
		 */
		R_UNLOCK(dbenv,&(cshare->reginfo));
		/*      printf("DEBUG: 37 after R_LOCK\n");
		 */
		break;
	    case 4:
	    case 5:
	    case 6:
		/*	printf("DEBUG: 65 I am here to process temporal data\n");
		 */
		start_time.tv_sec = ttdata.start.curr.tv_sec;
		start_time.tv_usec = ttdata.start.curr.tv_usec;
		break;
	    default:
		/*	printf("suit01: not expected mode, return success\n");
		 */
		return 0;
	    }
	    /* Mingde: Well, this part is acutally of no use. But for the
	     * non-temporal database, this comparison is a must
	     */
	    /*      printf("DEBUG: start the time check\n");
	     */
	    comp1 = &read_time;
	    comp2 = &start_time;
	    /*      printf("DEBUG: 68 the start_time is %ld(%ld)\n",
		    start_time.tv_sec, start_time.tv_usec);
	     */
	    if( timercmp ( comp1, comp2, > )) {
		/*	printf("readtime > starttime\n");
		 */
		start_time.tv_sec = read_time.tv_sec;
		start_time.tv_usec = read_time.tv_usec;
	    }
	    /*      printf("DEBUG:70 The data comes to the comparision\n");
	     */
	    comp1 =  &(txn_data->t_l);
	    comp2 = &start_time;
	    if( timercmp( comp1, comp2, <)) {
		/*
		  printf("The lower bound has been moved up \n");
		  printf("The start time is %ld(%ld)\n",
		  start_time.tv_sec, start_time.tv_usec);
		  printf("The lower bound is %ld(%ld)\n",
		  txn_data->t_l.tv_sec, txn_data->t_l.tv_usec);
		*/
		(txn_data->t_l).tv_sec = start_time.tv_sec;
		(txn_data->t_l).tv_usec = start_time.tv_usec;
	    }
	    /* That means we move up the lbound to avoid the reading of "future" data */
	    comp1 =  &(txn_data->t_l);
	    comp2 =  &(txn_data->t_h);
	    if( timercmp( comp1, comp2, >)) {
		/*	printf("suit01: The lower bound is too large now in read\n");*/
		/* This means the lower bound is now greater than the higer bound */
		TTDB_DBT_SETSTATE(((TTDATA*)(data->data))->state, INVALID_STATE);
                /*txn_abort(txn_data->tid);*/
		return TEMPORAL_CLK_CHECK;
	    }
	    /*printf("DEBUG: 75 before clk_check :read return success\n");
	     */
	    return 0;
	}
	if(mode == 1 || mode == 2) {
	    /*printf("DEBUG:76 before the R_LOCK in the clk_check:write\n");
	     */
	    R_LOCK(dbenv,&(cshare->reginfo));
	    read_time = clk_rtt_lookup(dbenv,position);
	    R_UNLOCK(dbenv, &(cshare->reginfo));
	    /*printf("DEBUG:80 finish the lock\n");
	     */
	    memcpy(&ttdata,data->data,sizeof(TTDATA));
	    switch(TTDB_DBT_STATE(ttdata.state)) {
	    case 1:
	    case 2:
	    case 3:
		R_LOCK(dbenv, &(cshare->reginfo));
		start_time = clk_wtt_lookup(dbenv,position);
		R_UNLOCK(dbenv, &(cshare->reginfo));
		break;
	    case 4:
	    case 5:
	    case 6:start_time.tv_sec = ttdata.start.curr.tv_sec;
		start_time.tv_usec = ttdata.start.curr.tv_usec;
		break;
		/*      default:printf("suit01: not expected mode, return success\n");
		 */
	    default:
		return 0;
		break;
	    }
	    comp1 = &read_time;
	    comp2 = &start_time;
	    if( timercmp(comp1, comp2, >)) {
		start_time.tv_sec = read_time.tv_sec;
		start_time.tv_usec = read_time.tv_usec;
	    }
	    if( timercmp(&start_time, &(txn_data->t_l), > )) {
		(txn_data->t_l).tv_sec = start_time.tv_sec;
		(txn_data->t_l).tv_sec = start_time.tv_sec;
	    }
	    if(timercmp( &(txn_data->t_l), &(txn_data->t_h), > )) {
		/*	printf("suit01: The lower bound is too large now in write\n");*/
		/*txn_abort(txn_data->tid);*/
		TTDB_DBT_SETSTATE(((TTDATA*)(data->data))->state, INVALID_STATE);
		return TEMPORAL_CLK_CHECK;
	    }
	    return FXNOK;
	}
    }

    else /*It is not a temporal*/ {
	if(mode == 0) {
	    /*It is read*/
	    /*	    printf("IN NONTEMPORAL\n");*/
	    R_LOCK(dbenv, &(cshare->reginfo));
	    start_time = clk_wtt_lookup(dbenv,position);
	    R_UNLOCK(dbenv, &(cshare->reginfo));
	    if( timercmp( &start_time,&(txn_data->t_l), > )) {
		(txn_data->t_l).tv_sec = start_time.tv_sec;
		(txn_data->t_l).tv_usec = start_time.tv_usec;
	    }
	    /*That means we move up the lbound to avoid the reading of "future" data*/
	    if( timercmp( &(txn_data->t_l), &(txn_data->t_h), >)) {
		 TTDB_DBT_SETSTATE(((TTDATA*)(data->data))->state, INVALID_STATE);
		/*txn_abort(txn_data->tid);*/
		return TEMPORAL_CLK_CHECK;
	    }

	    /* update the array rtt rec bit*/
	    clk_bitmanage(txn_data->V_R, position);
	    return FXNOK;
	}


	else if(mode == 1 || mode == 2) {
	    /*insert or delete*/
	    /*	   printf("IN NONTEMPORAL\n");*/
	    R_LOCK(dbenv, &(cshare->reginfo));
	    read_time = clk_rtt_lookup(dbenv,position);
	    start_time = clk_wtt_lookup(dbenv,position);
	    R_UNLOCK(dbenv, &(cshare->reginfo));
	    if( timercmp( &read_time, &start_time, > )) {
		start_time.tv_sec = read_time.tv_sec;
		start_time.tv_usec = read_time.tv_usec;
	    }
	    if( timercmp( &start_time, &(txn_data->t_l), > )) {
		(txn_data->t_l).tv_sec = start_time.tv_sec;
		(txn_data->t_l).tv_usec = start_time.tv_usec;
	    }

	    if( timercmp( &(txn_data->t_l), &(txn_data->t_h), > )) {
		/*txn_abort(txn_data->tid);*/
		return TEMPORAL_CLK_CHECK;
	    }
	    if(mode == 1) {
		clk_bitmanage(txn_data->V_W, position);
	    }
	    return FXNOK;
	}
    }
}


int clk_bitmanage(char *table, int position) {

    table[position/8] = (unsigned char) table[position/8] | (unsigned char)usedmask[position%8];
    return 0;
}



/*update the RTT table with transacstion time.*/
int clk_rtt_update(CLK_TXNDATA* txn_data) {

    DB_ENV *env;
    CLK_SHARE *cshare;
    int i;
    int ret;
    env = txn_data->env;
    cshare = (CLK_SHARE*)(((TEMPORAL_HANDLE *)(env->temporal_handle))->clk_handler);
    /*update the read time of all the data*/
    R_LOCK(env, &(cshare->reginfo));
    for ( i=0; i<ENTRIES_OF_TABLE; i++) {
	if  ( (unsigned char)((txn_data->V_R)[i/8] & usedmask[i%8]) ==
	      (unsigned char) (usedmask[i%8]) ) {
	    cshare->rtt_table_ptr[i].tv_sec = (txn_data->t_l).tv_sec;
	    cshare->rtt_table_ptr[i].tv_usec = (txn_data->t_l).tv_usec;
	}
    }
    R_UNLOCK(env,&(cshare->reginfo));
    return 0;
}


int clk_wtt_update(CLK_TXNDATA* txn_data) {

    DB_ENV *env;
    CLK_SHARE *cshare;
    int i;
    int ret;
    env = txn_data->env;
    cshare = (CLK_SHARE*)(((TEMPORAL_HANDLE *)(env->temporal_handle))->clk_handler);

    /*update the insert time of all the data*/
    R_LOCK(env, &(cshare->reginfo));
    for ( i=0; i<ENTRIES_OF_TABLE; i++) {
	if (  (unsigned char)((txn_data->V_W)[i/8] & usedmask[i%8] ) ==
	      (unsigned char)(usedmask[i%8]) ) {
	    cshare->wtt_table_ptr[i].tv_sec = (txn_data->t_l).tv_sec;
	    cshare->wtt_table_ptr[i].tv_usec = (txn_data->t_l).tv_usec;
	}
    }
    R_UNLOCK(env, &(cshare->reginfo));
    return 0;
}


int clk_txn_commit (DB_ENV *dbenv, DB_TXN* txnid) {

    int ret;
    struct timeval cur_time;
    CLK_TXNDATA *txn_data=NULL;

/* Huilong: 2/1/2007
 */
    DB_STP              *dbstp;
    DB_MPOOL            *dbmp;
/* Huilong */

    /*search for the list*/
    /*  printf("DEBUG: This is the beginning of txn_commit\n");
     */
    ret = clk_search_data(&txn_data, txnid);
    /*printf("20 shilong: txnid=%d, dbenv=%d\n", txnid, txnid->mgrp->dbenv);
     */
    if(ret != 0) {
	/*gdbenv->err(gdbenv, ret, "search_data: %s", "Can not find the data");*/
	return -1;
    }
    
/* Huilong: 2/1/2007
 * Compete for the exclusive lock with the notarizor/validator
 * (Stop and wait protocol)
 */
    dbmp = (DB_MPOOL *)dbenv->mp_handle;
    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;
    R_LOCK(dbenv, dbmp->reginfo);
    R_LOCK(dbenv, &dbstp->reginfo);
/* Huilong */

    gettimeofday(&(txn_data->t_h), NULL);
    /*printf("The timestamp before final setting is %ldsec, %ldusec\n",
      (txn_data->t_x).tv_sec, (txn_data->t_x).tv_usec);
    */
    if( (txn_data->t_x).tv_sec==0 && (txn_data->t_x).tv_usec == 0) {
	(txn_data->t_l).tv_usec ++;
	if( (txn_data->t_l).tv_usec >= 1000000 ) {
	    (txn_data->t_l).tv_usec = 0;
	    (txn_data->t_l).tv_sec ++;
	}
	if( timercmp( &(txn_data->t_l), &(txn_data->t_s), > )) {
	    /*  printf("t_l > t_s\n");*/
	    (txn_data->t_x).tv_sec = (txn_data->t_l).tv_sec;
	    (txn_data->t_x).tv_usec = (txn_data->t_l).tv_usec;
	}
	else {
	    /*printf("t_s > t_l\n");*/
	    (txn_data->t_x).tv_sec = (txn_data->t_s).tv_sec;
	    (txn_data->t_x).tv_usec = (txn_data->t_s).tv_usec;
	}
    }
    /*printf("40The t_l is %ld(%ld), the t_h is %ld(%ld)\n",txn_data->t_l.tv_sec,
      txn_data->t_l.tv_usec,txn_data->t_h.tv_sec,
      txn_data->t_h.tv_usec);
    */
    if( timercmp( &(txn_data->t_l), &(txn_data->t_h) ,> )) {
	/*    printf("t_l > t_h\n");
	 */
	/*MINGDE*//*txn_abort(txn_data->tid);*/

/* Huilong: 5/29/2007
 */
    R_UNLOCK(dbenv, dbmp->reginfo);
    R_UNLOCK(dbenv, &dbstp->reginfo);
/* Huilong */

	return TEMPORAL_CLK_CHECK;
    }
    /*printf("50The t_x of transaction is %ld(%ld)\n",txn_data->t_x.tv_sec,
      txn_data->t_x.tv_usec);
    */
    if( timercmp( &(txn_data->t_x), &(txn_data->t_h), > )) {
	/*printf("t_x > t_h\n");*/
	/*MINGDE*//*txn_abort(txn_data->tid);*/

/* Huilong: 5/29/2007
 */
    R_UNLOCK(dbenv, dbmp->reginfo);
    R_UNLOCK(dbenv, &dbstp->reginfo);
/* Huilong */

	return TEMPORAL_CLK_CHECK;
    }
    /*printf("60 shilong: txnid=%d, dbenv=%d\n", txnid, txnid->mgrp->dbenv);*/

    /* Here we need to add the STAMP module call to send out the t_x */
    clk_rtt_update(txn_data);
    /*printf("70 shilong: txnid=%d, dbenv=%d\n", txnid, txnid->mgrp->dbenv);*/
    clk_wtt_update(txn_data);
    /*printf("75 shilong: txnid=%d, dbenv=%d\n", txnid, txnid->mgrp->dbenv);*/
    /*printf("%d\n\n\n\n\n",sizeof(__db_txn));*/
    /* Here is the main difference in difference way of nesting */
    /* Change of NESTLEVEL code */
    if(((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->NESTLEVEL == PLAIN_LEVEL) {
	/* This is a plain level nesting, no parent and no child, thus we need
	 * to do nothing
	 */
    }
    else if( ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->NESTLEVEL == MYSQL_LEVEL) {
	/*This is a special way of nesting, as the meat is at the second level
	 * while the first level txn merely opens and close the second level
	 * transactions. Besides this, there will be no deeper nesting
	 * In view of this, we can consider it the same as the single level
	 * nesting, and we do not need to propagate the information upward
	 */
	/*
	  if((txn_data->tid)->parent != NULL) {
	  if(propagate(txn_data)!=0){
	  return CLK_ERR;
	  }
	  }
	*/
	;;
    }
    else if(((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->NESTLEVEL == NESTED_LEVEL) {
	/* This is the arbitrary nesting mode. The
	 * nesting can be any levels as long as the levels are more than two
	 * However, due to the nature of nesting, it must be true that :
	 * when child txn committed, its parent txn must be uncommitted, so
	 * we only need to propagate one step up
	 */

	if((txn_data->tid)->parent != NULL) {
	    if(propagate(txn_data)!=0){

/* Huilong: 5/29/2007
 */
    R_UNLOCK(dbenv, dbmp->reginfo);
    R_UNLOCK(dbenv, &dbstp->reginfo);
/* Huilong */

		return TEMPORAL_CLK_CHECK;
	    }
	}
    }
    /*printf("The timestamp is %ldsec, %ldusec\n",
      (txn_data->t_x).tv_sec, (txn_data->t_x).tv_usec);
    */

    /*printf("80 shilong: txnid=%d, dbenv=%d\n", txnid, txnid->mgrp->dbenv);*/
    __stp_txn_commit(dbenv, txnid->txnid,&(txn_data->t_x));
    /*printf("90 shilong: txnid=%d, dbenv=%d\n", txnid, txnid->mgrp->dbenv);*/
    clk_delete_data(txnid);
    /*printf("100 shilong: txnid=%d, dbenv=%d\n", txnid, txnid->mgrp->dbenv);*/
     


/* Huilong: 2/1/2007
 * Return the exclusive lock
 * (Stop and wait protocol)
 */
    R_UNLOCK(dbenv, &dbstp->reginfo);
    R_UNLOCK(dbenv, dbmp->reginfo);
/* Huilong */

    return FXNOK;
}

int pushdown(CLK_TXNDATA *txn_data) {

}

int propagate(CLK_TXNDATA *txn_data) {
    CLK_TXNDATA *parent_data;
    int i;
    int ret = 0;

    ret = clk_search_data(&parent_data, txn_data->tid->parent);
    if(ret != 0) {
	/*txn_abort(txn_data->tid);*/
	return TEMPORAL_CLK_CHECK;
    }
    for( i= 0; i< ENTRIES_OF_TABLE/sizeof(char)+ 1; i++) {
	((parent_data->V_R)[i]) = (char)((parent_data->V_R)[i]) | (char)((txn_data->V_R)[i]);
	((parent_data->V_W)[i]) = (char)((parent_data->V_W)[i]) | (char)((txn_data->V_W)[i]);
    }
    (parent_data->t_l).tv_sec = (txn_data->t_x).tv_sec;
    (parent_data->t_l).tv_usec = (txn_data->t_x).tv_usec;
    return 0;
}

int clk_txn_abort(DB_TXN *tid) {
    CLK_TXNDATA *txn_data;
    int ret;
    ret = clk_search_data(&txn_data,tid);
    if(ret != 0) {
	return -1;
    }
    __stp_txn_abort(txn_data->env, tid->txnid);
    ret = clk_delete_data(tid);
    /* Shilong
     * this cause env panic */
    /* ret = txn_abort(tid);*/
    return ret;
}


/*
 * __clk_open --
 *	Open the shared region of CLK module.
 *	Called in __dbenv_open (env/env_open.c).
 */

int  __clk_open(DB_ENV *dbenv)
{
    CLK_SHARE		*clk_share;
    SHARE_NODE		*cnode;
    int		ret;

#ifdef CLK_DEBUG
    printf("\tEnter __CLK_open()\n");
#endif

    /*gdbenv = dbenv;*/
/* Modified by Huilong at 15:11 7/11/2006 */
//    ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->envid = time(NULL);
/* end of modification */

    /* if nesting level has not been set till now, we set it
       to default value to prevent it being set later.
    */
    if(((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->NESTLEVEL == 0)
	((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->NESTLEVEL = 1;

    /* Create/initialize the CLK_SHARE structure. */
    if ((ret = __os_calloc(dbenv, 1, sizeof(CLK_SHARE), &clk_share)) != 0)
	return (ret);
    clk_share->dbenv = dbenv;

    /* Join/create the log region. */
    clk_share->reginfo.type = REGION_TYPE_CLK;
    clk_share->reginfo.id = INVALID_REGION_ID;
    clk_share->reginfo.mode = dbenv->db_mode;
    clk_share->reginfo.flags = REGION_JOIN_OK;
    if (F_ISSET(dbenv, DB_ENV_CREATE))
	F_SET(&clk_share->reginfo, REGION_CREATE_OK);
    if ((ret = __db_r_attach(
	dbenv, &clk_share->reginfo, CLK_REG_SIZE)) != 0)
	goto err;

    /* If we created the region, initialize it. */
    if (F_ISSET(&clk_share->reginfo, REGION_CREATE))
	if ((ret = __share_init(dbenv, clk_share)) != 0)
	    goto err;

    /* Set the local addresses. */
    cnode = clk_share->reginfo.primary =
	R_ADDR(&clk_share->reginfo, clk_share->reginfo.rp->primary);

    clk_share->rtt_table_ptr = R_ADDR(&clk_share->reginfo, cnode->rtt_off);
    clk_share->wtt_table_ptr = R_ADDR(&clk_share->reginfo, cnode->wtt_off);

    /*
     * If the region is threaded, then we have to lock both the handles
     * and the region, and we need to allocate a mutex for that purpose.
     */
    if (F_ISSET(dbenv, DB_ENV_THREAD)) {
#ifdef BDB_4
	if ((ret = __db_mutex_setup(dbenv, &clk_share->reginfo, &clk_share->mutexp,
				    MUTEX_ALLOC | MUTEX_NO_RLOCK | MUTEX_THREAD)) != 0)
	    goto err;
#endif
#ifdef BDB_3
	if ((ret = __db_mutex_alloc(dbenv, &clk_share->reginfo,
				    &clk_share->mutexp)) != 0)
	    goto err;
	if ((ret = __db_mutex_init(dbenv, clk_share->mutexp,
				   0, MUTEX_THREAD)) != 0)
	    goto err;
#endif
	/* Shilong 05/27/03
	if ((ret = __db_shmutex_init(dbenv, clk_share->mutexp, 0,
				     MUTEX_THREAD, &clk_share->reginfo,
				     (REGMAINT *)R_ADDR(&clk_share->reginfo,
							cnode->maint_off))) != 0)
	    goto err;
	*/
    }

    /* Initialize the rest of the structure. */

    ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->clk_handler = clk_share;

    R_UNLOCK(dbenv, &(clk_share->reginfo));
    return (0);

  err:
    if (clk_share->reginfo.addr != NULL) {
	if (F_ISSET(&clk_share->reginfo, REGION_CREATE))
	    ret = __db_panic(dbenv, ret);
	R_UNLOCK(dbenv, &clk_share->reginfo);
	(void)__db_r_detach(dbenv, &clk_share->reginfo, 0);
    }

    if (clk_share->mutexp != NULL)
	__db_mutex_free(dbenv, &clk_share->reginfo, clk_share->mutexp);

    __os_free(
#ifndef BDB_3_2_9
	dbenv,
#endif
	clk_share, sizeof(*clk_share));

    return (ret);
}

int __share_init(DB_ENV *dbenv, CLK_SHARE *clk_share)
{
    int ret;
    SHARE_NODE			*node;

    if ((ret = __db_shalloc(clk_share->reginfo.addr,
			    sizeof(*node), 0, &clk_share->reginfo.primary)) != 0)
	goto mem_err;
    clk_share->reginfo.rp->primary =
	R_OFFSET(&clk_share->reginfo, clk_share->reginfo.primary);
    node = clk_share->reginfo.primary;
    memset(node, 0, sizeof(*node));


    if ((ret =
	 __db_shalloc(clk_share->reginfo.addr, ENTRIES_OF_TABLE*sizeof(struct timeval), 0, &(clk_share->wtt_table_ptr))) != 0) {
	__db_err(dbenv, "Unable to allocate enough memory for the CLK write table");
	return (ret);
    }
    node->wtt_off = R_OFFSET(&clk_share->reginfo, clk_share->wtt_table_ptr);

    if ((ret =
	 __db_shalloc(clk_share->reginfo.addr, ENTRIES_OF_TABLE*sizeof(struct timeval), 0, &(clk_share->rtt_table_ptr))) != 0) {
      mem_err:
	__db_err(dbenv, "Unable to allocate enough memory for the CLK read table");
	return (ret);
    }
    node->rtt_off = R_OFFSET(&clk_share->reginfo, clk_share->rtt_table_ptr);
    memset(clk_share->rtt_table_ptr, 0, ENTRIES_OF_TABLE*sizeof(struct timeval));
    memset(clk_share->wtt_table_ptr, 0, ENTRIES_OF_TABLE*sizeof(struct timeval));
    return (0);
}

