/*-
 * Developed (2002-2003) by:
 *     TAU Project
 *     Computer Science Dept., University of Arizona
 *     Tucson, AZ 85721, USA
 *     URL  : http://www.cs.arizona.edu/tau
 *     Email: tau@cs.arizona.edu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
/*-
 * Transaction Support to BDB - Time Stamper
 *
 * By: Shilong Yao (yao@CS.Arizona.EDU)
 *     Yuhong Liu            (yuhongl@CS.Ariozna.EDU)
 * Advisor: Dr. Richard Snodgrass (rts@CS.Arizona.EDU )
 *
 * The University of Arizona
 *
 * 04/08/02: Create
 *		Only BTree is implemented. Hooks for other page layout are provided.
 * 06/20/02: Version 2 Begin Coding
 * 04/09/05: version 3 Finished
 *
 */
/***************************************************
 * Configurations
 ***************************************************/


/* #define _STAMPER_DEBUG_ */




/* Audit requires this! */
#define _STAMP_AT_COMMIT_

/*
#define _EXPERIMENT_
*/

/* Sizes */
/* Huilong 7/31/06 */
#define STP_REG_SIZE			35535000
//#define STP_REG_SIZE			(35535000+12288) // add ov_table size
/* Huilong */
#define STP_TT_HTAB_SIZE                1024
#define STP_BT_HTAB_SIZE                2048
#define STP_MAX_TT_ENT                  102400
//#define STP_MAX_BT_ENT                  4096
#define STP_MAX_BT_ENT                    4096 // modified by Huilong
#define STP_MAX_BT_NODE                 (STP_MAX_BT_ENT*10)
#define IDX_INVALID                     (-1)

/* Time consts */
#define STP_TIME_INVALID		0

/* Error Code */
#define STP_ERROR_TXN_BEGIN		1
#define STP_ERROR_TXN_COMMIT	2
#define STP_ERROR_TXN_ABORT		3
#define STP_ERROR_RECORD		5
#define STP_ERROR_PAGE			6
#define STP_ERROR_BHFREE		7

/* Debug processing */
#ifdef _STAMPER_DEBUG_
#define STP_DEBUG(s) s
#else
#define STP_DEBUG(s)
#endif

/* Let all STP functions return without doing anything */
/*#define _STP_STOP_WORKING_*/
#ifdef  _STP_STOP_WORKING_
#define STP_STOP_WORKING return 0
#else
#define STP_STOP_WORKING()
#endif

/* Experiment processing */
#ifdef _EXPERIMENT_
#define STP_EXP(s) s
#else
#define STP_EXP(s)
#endif

/* Error processing */
#define STP_ERROR() do { \
    fprintf(stderr, " ***** STP module error at (file, line) = (%s, %d) *****\n", __FILE__, __LINE__); \
    goto err; \
} while (0)


/* Whether a record passed to __stp_record() gets stamped or not */
#define UNCHANGED			0
#define CHANGED				1

/* Action code in function __stp_page() */
#define PAGE_READ			1
#define PAGE_WRITE			2
#define MPF_PUT				3

/* Typedef of existing types */
typedef struct timeval TV;

/* Record types in TL (u_int_32) */
/* modified by yuhong -- 2/22/05
   Only two bits: STP_READ_NONE and STP_WRITE_STAMPED needed
   for read state and write state (for each TL(tnx, pgno) node) --
     read_state = NONE if set STP_READ_NONE
     read_state = UNSTAMPED if not set STP_READ_NONE
     write_state = UNSTAMPED if not set STP_WRITE_STAMPED
     write_state = STAMPED if set STP_WRITE_STAMPED
*/
#define STP_READ_NONE          1 /* Bit 0 */
#define STP_WRITE_STAMPED      2 /* Bit 1 */

/* Data structure definition */
struct __stp_tt_entry {
    STP_ID	txnid;
    TV		time;
    u_int32_t	od_pgcnt;
    SH_TAILQ_HEAD(__btn_intt) btn;
    SH_TAILQ_ENTRY links;
};
typedef struct __stp_tt_entry STP_TT_ENTRY;

struct __stp_bt_entry {
    db_pgno_t	pgno;
    u_int8_t 	fileid[DB_FILE_ID_LEN];
    SH_TAILQ_HEAD(__btn_inbt) btn;
    SH_TAILQ_ENTRY links;
};
typedef struct __stp_bt_entry STP_BT_ENTRY;

struct __stp_bt_node {
    STP_TT_ENTRY    *tt;             /* index to the TT entry */
    SH_TAILQ_ENTRY  vlinks, hlinks; /* vertical/horizontal links */
    u_int32_t       flags;
 #ifdef STAMPER_DEBUG
    db_pgno_t       pgno;           /* just for the convenient of print_tt() */
 #endif
};
typedef struct __stp_bt_node STP_BTN;

/* Huilong 7/27/06 */
/*#define MAX_NUM_OVN    1024
struct __stp_ov_item {
    DB_MPOOLFILE   *dbmfp;
    db_pgno_t      rec_pgno;
    db_pgno_t      ov_pgno;
};
typedef struct __stp_ov_item STP_OVN;
/* Huilong */

/* Huilong 10/11/2006 */
#define INCR_CHUNK_SIZE  128
struct __stp_incr_item {
    STP_ID      txnid;
    int32_t     fileid;
    db_pgno_t   pgno;
};
typedef struct __stp_incr_item STP_INCR;
struct __stp_incr_chunk {
    STP_INCR    incrs[INCR_CHUNK_SIZE];
    int         count;
    struct __stp_incr_chunk *next;
};
typedef struct __stp_incr_chunk STP_INCR_CHUNK;
/* Huilong */

struct __db_stp {
    /*
     * These fields need to be protected for multi-threaded support.
     *
     * !!!
     * As this structure is allocated in per-process memory, the mutex may need
     * to be stored elsewhere on architectures unable to support mutexes in heap
     * memory, e.g., HP/UX 9.
     */
#ifdef BDB_4
    DB_MUTEX
#endif
#ifdef BDB_3
    MUTEX
#endif
    *mutexp;		/* Mutex for thread protection. */

    DB_HASHTAB          *tt_htab;               /* TT hash table */
    u_int32_t 		tt_htab_size;		/* Number of entries in the TT hash table */

    DB_HASHTAB          *bt_htab;               /* BT hash table */
    u_int32_t 		bt_htab_size;		/* Number of entries in the TT hash table */

/* Huilong 7/27/06 */
/*    STP_OVN             *ov_table;
    u_int32_t           ov_count;
/* Huilong */

    u_int32_t max_tt_ent;   /* max number of total TT entries */
    u_int32_t max_bt_node;  /* max number of total BTN nodes */
    u_int32_t max_bt_ent;   /* max number of total BT entries */

    /* These fields are not thread-protected. */
    DB_ENV			*dbenv;			/* Reference to error information. */
    REGINFO			reginfo;		/* Underlying cache regions. */
    u_int32_t		flags;
};
typedef struct __db_stp DB_STP;

/*
 * STP --
 *	Shared stamper region.  One of these is allocated in shared memory,
 *	and describes the stamper.
 */
struct __stp {

    u_int32_t	tt_htab_size;		/* Number of the entries in TT htab */
    roff_t		tt_htab_off;		/* Offset of TT htab */
    u_int32_t	bt_htab_size;		/* Number of the entries in BT htab */
    roff_t		bt_htab_off;			/* Offset of BT htab */

/* Huilong 7/27/06*/
/*    u_int32_t   ov_count;
    roff_t      ov_table_off;
/* Huilong */

    u_int32_t max_tt_ent;   /* max number of total TT entries */
    u_int32_t max_bt_node;  /* max number of total BTN nodes */
    u_int32_t max_bt_ent;   /* max number of total BT entries */

    SH_TAILQ_HEAD(__tt_indbstp) free_tt_ent;
    SH_TAILQ_HEAD(__btn_indbstp) free_bt_node;
    SH_TAILQ_HEAD(__bt_indbstp) free_bt_ent;

    int			in_flush;		/* Log flush in progress. */
#ifdef BDB_4
    DB_MUTEX
#endif
#ifdef BDB_3
    MUTEX
#endif
    flush;			/* Mutex for flushing. */

#ifdef MUTEX_SYSTEM_RESOURCES
#define	STP_MAINT_SIZE	(sizeof(roff_t) * DB_MAX_HANDLES)
    roff_t		maint_off;	/* offset of region maintenance info */
#endif

    u_int32_t	flags;			/* Environment-wide flag values. */
};
typedef struct __stp STP;

/* Macros */
/* Global temporary variable used to help the linking list operations */
u_int32_t list_index_tmp;

/* In BDB 3.2.9, there is no IS_WRITELOCK(mode) macro */
#ifdef BDB_3
#define IS_WRITELOCK(mode) ( \
	( (mode == DB_LOCK_WRITE) || \
	  (mode == DB_LOCK_IWRITE) || \
	  (mode == DB_LOCK_IWR) )? 1:0 \
)
#endif

/* Time related operations to deal with "struct
 * timeval
 */
#define TIME(time, sec, usec) do { \
	time.tv_sec = sec; \
	time.tv_usec = usec; \
} while(0)
#define INVALID_TIME(time) TIME(time, 0, 0)
#define IS_INVALID_TIME(time) ((time.tv_sec == 0)&&(time.tv_usec == 0))

/* STP_ID related operations */
/* Modified by Huilong at 15:34 7/11/2006 */
#define STPID_EQUAL(i1, i2) ((i1.envid.tv_sec == i2.envid.tv_sec) \
                             && (i1.envid.tv_msec == i2.envid.tv_msec) \
                             && (i1.txnid == i2.txnid))
#define STPID_INVALID(i) ((i.envid.tv_sec == 0) \
                          && (i.envid.tv_msec == 0) \
                          && (i.txnid == 0))
#define STPID_INIT(i) do { i.envid.tv_sec = 0; \
                           i.envid.tv_msec= 0; \
                           i.txnid = 0; } while(0)
//#define STPID_EQUAL(i1, i2) ((i1.envid == i2.envid) && (i1.txnid == i2.txnid))
//#define STPID_INVALID(i) ((i.envid == 0) && (i.txnid == 0))
//#define STPID_INIT(i) do { i.envid = i.txnid = 0; } while(0)
/* end of modification */
