#include "db_config.h"

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/* Original BDB system */
#include "db_int.h"
#include "db_shash.h"
#include "lock.h"
#include "db_page.h"
#include "mp.h"
#include "qam.h"
#include "log.h"

/* TT support */
#include "db.h"
#include "TTDB.h"
#include "clk_txn.h"
#include "stp_int.h"
#include "aud_int.h"
#include "temporal.h"

/* Huilong 7/24/06 */
#include "stp_tt.h"
#define __AUD
/* Huilong */
/* FOR DEBUG !!!!! */
int
print_hash(char *h)
{
    int i;
    for (i = 0; i < AUD_HASH_LEN; i ++)
	printf("%02X", 0xff & h[i]);
    printf("\n");
    return 0;
}
int
print_tol(dbenv)
    DB_ENV *dbenv;
{
    DB_AUD *dbaud;
    AUD *aud;
    AUD_TOL_ENTRY *tolep, *new_tolep, *prev_tolep;
return 0;

    dbaud = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->aud_handler;
    if (dbaud == NULL)
	return 0;
    aud = dbaud->reginfo.primary;

    printf("+++++++++++ TOL ++ tail_time=(%d, %d) tail_hash=%d, %d++++++++++\n",
	   aud->hash_tail_time.tv_sec, aud->hash_tail_time.tv_usec,
	   aud->hash_value[0], aud->hash_value[1]);
    /* Search on the current TOL for the beginning txn */
    for (tolep = SH_TAILQ_FIRST(&aud->tol, __aud_tol_entry);
	 tolep != NULL;
	 tolep = SH_TAILQ_NEXT(tolep, links, __aud_tol_entry)) {

	printf("\ttxnid=%d, ready=%d, time=(%d, %d), hash=",
	       tolep->txnid, tolep->ready, 
	       tolep->time.tv_sec, tolep->time.tv_usec);
	print_hash(tolep->hash_value);
    }
    printf("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");

    return 0;
}
/* END OF DEBUG CODE !!!! */

static int
__tol_cmp(AUD_TOL_ENTRY *a, AUD_TOL_ENTRY *b)
{
    TV ta = a->time,
	tb = b->time;
    if (ta.tv_sec < tb.tv_sec)
	return -1;
    else if (ta.tv_sec > tb.tv_sec)
	return 1;
    else if (ta.tv_usec < tb.tv_usec)
	return -1;
    else if (ta.tv_usec > tb.tv_usec)
	return 1;
    else
	return 0;
}

static int
__aud_init(dbenv, dbaud)
    DB_ENV	*dbenv;
    DB_AUD 	*dbaud;
{
    AUD *region;
    AUD_TOL_ENTRY *entryp;
    int ret, i;

#ifdef BDB_4
#ifdef HAVE_MUTEX_SYSTEM_RESOURCES
    u_int8_t *addr;
#endif
#endif

#ifdef BDB_3
    u_int8_t *addr;
#ifdef MUTEX_SYSTEM_RESOURCES
    size_t maint_size;
#endif
#endif

#ifndef __AUD
    return 0;
#endif

    /*printf("ppp 10\n"); fflush(stdout);*/
    if ((ret = __db_shalloc(dbaud->reginfo.addr,
			    sizeof(*region), 0, &dbaud->reginfo.primary)) != 0)
	goto mem_err;

    dbaud->reginfo.rp->primary =
	R_OFFSET(&dbaud->reginfo, dbaud->reginfo.primary);
    region = dbaud->reginfo.primary;
    memset(region, 0, sizeof(*region));

    /* Initialize the table size */
    region->tol_size = AUD_TOL_SIZE;
    region->ready_to_hash = 1;

#ifdef BDB_4
#ifdef HAVE_MUTEX_SYSTEM_RESOURCES
    /* Allocate room for the txn maintenance info and initialize it. */
    if ((ret = __db_shalloc(dbaud->reginfo.addr,
			    sizeof(REGMAINT) + TXN_MAINT_SIZE, 0, &addr)) != 0) {
	__db_err(dbenv,
		 "Unable to allocate memory for mutex maintenance");
	return (ret);
    }
    __db_maintinit(&dbaud->reginfo, addr, TXN_MAINT_SIZE);
    region->maint_off = R_OFFSET(&dbaud->reginfo, addr);
#endif
#endif

#ifdef BDB_3
#ifdef MUTEX_SYSTEM_RESOURCES
    /* Allocate room for the txn maintenance info and initialize it. */
    if ((ret = __db_shalloc(dbaud->reginfo.addr,
			    sizeof(REGMAINT) + AUD_MAINT_SIZE, 0, &addr)) != 0)
	goto mem_err;
    __db_maintinit(&dbaud->reginfo, addr, AUD_MAINT_SIZE);
    region->maint_off = R_OFFSET(&dbaud->reginfo, addr);
#endif
    if ((ret = __db_shmutex_init(dbenv, &region->flush, 0,
				 0, &dbaud->reginfo,
				 (REGMAINT *)R_ADDR(&dbaud->reginfo,
						    region->maint_off))) != 0)
	return (ret);
#endif

    /* Initialize TOL entries onto a free list.  */
    SH_TAILQ_INIT(&region->tol);
    SH_TAILQ_INIT(&region->free_tol_entries);
/*
printf("region->tol_size=%d\n", region->tol_size);
*/
    for (i = 0; i < region->tol_size; ++i) {
	if ((ret = __db_shalloc(dbaud->reginfo.addr,
				sizeof(AUD_TOL_ENTRY), 0, &entryp)) != 0) {
	  mem_err:
	    __db_err(dbenv, "Unable to allocate memory for the lock table");
	    return (ret);
	}
	SH_TAILQ_INSERT_HEAD(&region->free_tol_entries, entryp,
			     links, __aud_tol_entry);
    }

    return (0);
}

int __aud_open(dbenv)
    DB_ENV *dbenv;
{

    DB_AUD *dbaud;
    AUD *aud;
    int	ret;

#ifndef __AUD
    return 0;
#endif

#ifdef AUD_DEBUG
    printf("\tEnter __aud_open()\n");
#endif

    /* Create/initialize the DB_AUD structure. */
    if ((ret = __os_calloc(dbenv, 1, sizeof(DB_AUD), &dbaud)) != 0)
	return (ret);
    dbaud->dbenv = dbenv;

    /* Join/create the log region. */
    dbaud->reginfo.type = REGION_TYPE_AUD;
    dbaud->reginfo.id = INVALID_REGION_ID;
    dbaud->reginfo.mode = dbenv->db_mode;
    dbaud->reginfo.flags = REGION_JOIN_OK;
    if (F_ISSET(dbenv, DB_ENV_CREATE))
	F_SET(&dbaud->reginfo, REGION_CREATE_OK);
    if ((ret = __db_r_attach(
	dbenv, &dbaud->reginfo, AUD_REG_SIZE)) != 0)
	goto err;

    /* If we created the region, initialize it. */
    if (F_ISSET(&dbaud->reginfo, REGION_CREATE))
	if ((ret = __aud_init(dbenv, dbaud)) != 0)
	    goto err;

    /* Set the local addresses. */
    aud = dbaud->reginfo.primary =
	R_ADDR(&dbaud->reginfo, dbaud->reginfo.rp->primary);
	
    /* in case we are joining another env, we don't enter __aud_init,
     *  we should hook the tables to the dbaud's pointer here.
     * - 01/30/03
    dbaud->tol = R_ADDR(&dbaud->reginfo, aud->tt_htab_off);
    dbaud->tol_size = aud->tol_size;
     */

    /*
     * If the region is threaded, then we have to lock both the handles
     * and the region, and we need to allocate a mutex for that purpose.
     */
    if (F_ISSET(dbenv, DB_ENV_THREAD)) {
#ifdef BDB_4
	if ((ret = __db_mutex_setup(dbenv, &dbaud->reginfo, &dbaud->mutexp,
				    MUTEX_ALLOC | MUTEX_NO_RLOCK | MUTEX_THREAD)) != 0)
	    goto err;
#endif
#ifdef BDB_3
	if ((ret = __db_mutex_alloc(dbenv, &dbaud->reginfo,
				    &dbaud->mutexp)) != 0)
	    goto err;
	if ((ret = __db_mutex_init(dbenv, dbaud->mutexp,
				   0, MUTEX_THREAD)) != 0)
	    goto err;
#endif
    }

    /* Initialize the rest of the structure. */

    /*
     * Set the handle -- we may be about to run recovery, which allocates
     * log cursors.  Log cursors require logging be already configured,
     * and the handle being set is what demonstrates that.
     */
    ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->aud_handler = dbaud;

    R_UNLOCK(dbenv, &(dbaud->reginfo));

    return (0);

  err:
    if (dbaud->reginfo.addr != NULL) {
	if (F_ISSET(&dbaud->reginfo, REGION_CREATE))
	    ret = __db_panic(dbenv, ret);
	R_UNLOCK(dbenv, &dbaud->reginfo);
	(void)__db_r_detach(dbenv, &dbaud->reginfo, 0);
    }

    if (dbaud->mutexp != NULL)
	__db_mutex_free(dbenv, &dbaud->reginfo, dbaud->mutexp);

#ifdef BDB_4
	__os_free(dbenv, dbaud);
#endif
#ifdef BDB_3
	__os_free(dbaud, sizeof(*dbaud));
#endif
    return (ret);
}

int __aud_txn_begin(dbenv, txnid)
    DB_ENV *dbenv;
    u_int32_t txnid;
{
    DB_AUD *dbaud;
    AUD *aud;
    AUD_TOL_ENTRY *tolep, *new_tolep, *prev_tolep;

#ifndef __AUD
    return 0;
#endif

    dbaud = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->aud_handler;
    if (dbaud == NULL)
	return 0;
    aud = dbaud->reginfo.primary;

/*
printf("enter aud_txn_begin: txnid=%d\n", txnid);
*/
    R_LOCK(dbenv, &dbaud->reginfo);

    /* Search on the current TOL for the beginning txn */
    for (tolep = SH_TAILQ_FIRST(&aud->tol, __aud_tol_entry);
	 tolep != NULL;
	 tolep = SH_TAILQ_NEXT(tolep, links, __aud_tol_entry)) {

	if (tolep->txnid == txnid)
	    break;

    }

    /* If the txn is already on TOL, ERROR */
    if (tolep != NULL) {
	printf("10\n");
	goto err;
    }

    /* Put the txn on TOL in its correct order */
    new_tolep = SH_TAILQ_FIRST(&aud->free_tol_entries, __aud_tol_entry);
    if (new_tolep != NULL)
	SH_TAILQ_REMOVE(&aud->free_tol_entries, new_tolep,
			links, __aud_tol_entry);
    else {
	printf("20\n");
	goto err;
    }

    new_tolep->txnid = txnid;
    new_tolep->ready = 0;
    new_tolep->time.tv_sec = time(NULL); /* Better to ask CLK for this
					 upper bound !! */
    new_tolep->time.tv_usec = 0;
    memset(new_tolep->hash_value, 0, AUD_HASH_LEN);
    new_tolep->seq = 1;
    __aud_hash_newtxn(new_tolep);
    
    tolep = SH_TAILQ_FIRST(&aud->tol, __aud_tol_entry);
    if (tolep == NULL) {
	SH_TAILQ_INSERT_HEAD(&aud->tol, new_tolep, links, __aud_tol_entry);
    } else {
	if (__tol_cmp(new_tolep, tolep) < 0) {
	    SH_TAILQ_INSERT_HEAD(&aud->tol, new_tolep, links, __aud_tol_entry);
	} else {
	    prev_tolep = tolep;
	    for (tolep = SH_TAILQ_NEXT(tolep, links, __aud_tol_entry);
		 (tolep != NULL) && (__tol_cmp(new_tolep, tolep) < 0);
		 prev_tolep = tolep,
		     tolep = SH_TAILQ_NEXT(tolep, links, __aud_tol_entry));
	    SH_TAILQ_INSERT_AFTER(&aud->tol, prev_tolep, new_tolep,
				 links, __aud_tol_entry);
	}
    }

    R_UNLOCK(dbenv, &dbaud->reginfo);
    return 0;
  err:
    fprintf(stderr, "Error in __aud_txn_begin\n");
    exit(1);
}

/* in __stp_txn_end() we scaned and hashed all the pages modified
 * by the txn. Then this function is called to finalize the TOL entry
 * representing this txn. If possible, the TOL entry is link-hashed
 * and removed from the TOL.
 * To make the TOL stable, a log entry is written to record that
 * the TOL entry for this txn is finalized in TOL.
 */
int __aud_txn_end(dbenv, txnid_arg)
    DB_ENV *dbenv;
    STP_ID txnid_arg;
{
    DB_AUD *dbaud;
    AUD *aud;
    AUD_TOL_ENTRY *tolep, *curr_tolep, *prev_tolep;
    AUD_TXN_HASH_DATA txn_hash_data;
    char digest[AUD_HASH_LEN];
    DB_STP *dbstp;
    STP_TT_ENTRY *tt;
    u_int32_t txnid;
    int i;
    DB_LSN ret_lsn;

#ifndef __AUD
    return 0;
#endif

    txnid = txnid_arg.txnid;
/*
printf("===> ENTER __aud_txn_end: txnid=%d\n", txnid);
*/

    dbaud = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->aud_handler;
    if (dbaud == NULL)
	return 0;
    aud = dbaud->reginfo.primary;

    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;
    if (dbstp == NULL)
	return 0;

    R_LOCK(dbenv, &dbaud->reginfo);

    STP_TT_FIND(dbstp, dbstp->tt_htab, txnid_arg, tt);

    /* Search on the current TOL for the ending txn */
    for (curr_tolep = SH_TAILQ_FIRST(&aud->tol, __aud_tol_entry);
	 curr_tolep != NULL;
	 curr_tolep = SH_TAILQ_NEXT(curr_tolep, links, __aud_tol_entry)) {

	if (curr_tolep->txnid == txnid)
	    break;

    }

    /* If the txn is not already on TOL, ERROR */
    if (curr_tolep == NULL)
	goto err;

    /* The hash of this transaction is done. So we copy the
     *   intermediate hash value to the final hash value and
     *   release the hash state.
     */
    if (__aud_hash_finish(curr_tolep) != 0) {
	fprintf(stderr, "__aud_hash_finish(): error\n");
	goto err;
    }

    /* If the txn is empty, we don't hash it on the chain. 
     * And we also remove it from the tol.
     */
    for (i=0; i<AUD_HASH_LEN; i++)
	if (curr_tolep->hash_value[i] != 0)
	    break;
    if (i >= AUD_HASH_LEN) {
	SH_TAILQ_REMOVE(&aud->tol, curr_tolep, links, __aud_tol_entry);
/* Added by Huilong to fix a bug */
        SH_TAILQ_INSERT_HEAD(&aud->free_tol_entries, curr_tolep,
                              links, __aud_tol_entry);        
/* End Huilong */
	goto done;
    }

    /* Log the adding of this TOL entry */
    if (__aud_tol_insert_log(dbenv, &(curr_tolep->time),
			     curr_tolep->hash_value, &ret_lsn, 0) != 0) {
	printf("Error return from __aud_tol_insert_log\n");
	goto err;
    }
    
    /* Set the txn to be ready for link-hashing */
    curr_tolep->ready = 1;

    /* Set the txn time */
    curr_tolep->time = tt->time;

    /* Adjust the TOL entry for this txn by removing it from
     * the TOL and reinsert it onto TOL.
     */
    SH_TAILQ_REMOVE(&aud->tol, curr_tolep, links, __aud_tol_entry);

    tolep = SH_TAILQ_FIRST(&aud->tol, __aud_tol_entry);
    if (tolep == NULL) {
	SH_TAILQ_INSERT_HEAD(&aud->tol, curr_tolep, links, __aud_tol_entry);
    } else {
	if (__tol_cmp(curr_tolep, tolep) < 0) {
	    SH_TAILQ_INSERT_HEAD(&aud->tol, curr_tolep, links, __aud_tol_entry);
	} else {
	    prev_tolep = tolep;
	    for (prev_tolep=tolep,
		     tolep=SH_TAILQ_NEXT(prev_tolep, links, __aud_tol_entry);
		 ((tolep != NULL) && (__tol_cmp(curr_tolep, tolep) > 0));
		 tolep = SH_TAILQ_NEXT(tolep, links, __aud_tol_entry)) {
		prev_tolep = tolep;
	    }
	    SH_TAILQ_INSERT_AFTER(&aud->tol, prev_tolep, curr_tolep,
				  links, __aud_tol_entry);
	}
    }

print_tol(dbenv);
    /* See if we can hash-link any txn on the TOL? */
    for (tolep = SH_TAILQ_FIRST(&aud->tol, __aud_tol_entry);
	 (tolep != NULL) && (tolep->ready);
	 tolep = SH_TAILQ_FIRST(&aud->tol, __aud_tol_entry)) {

	/* Do the link-hashing !!! */
	/* Update the current hash-chain tail value and tail time */
	if (!aud->ready_to_hash) {
	    break;
	}
	memcpy(txn_hash_data.prev_hash_value, aud->hash_value, AUD_HASH_LEN);
	txn_hash_data.time = tolep->time;
	memcpy(txn_hash_data.hash_value, tolep->hash_value, AUD_HASH_LEN);
	if (__aud_hash(&txn_hash_data, sizeof(AUD_TXN_HASH_DATA),
		       aud->hash_value) != 0)
	    goto err;
	aud->hash_tail_time = tolep->time;

/*
printf("++++++++++++++++++\n");
printf("Hash time(%d, %d), txnid=(---, %d)\n",
       tolep->time.tv_sec, tolep->time.tv_usec,
       tolep->txnid);
printf("    txn_hash =");
print_hash(tolep->hash_value);
printf("    tail_hash=");
print_hash(aud->hash_value);
*/

	/* Log the update to the current hash-chain tail (including txnid) */
	

	/* Delete the hashed TOL entry */
	SH_TAILQ_REMOVE(&aud->tol, tolep, links, __aud_tol_entry);
	SH_TAILQ_INSERT_HEAD(&aud->free_tol_entries, tolep,
			     links, __aud_tol_entry);
	/* Log the deletion of the hashed TOL entry (includeing txnid) */

    }

  done:
    R_UNLOCK(dbenv, &dbaud->reginfo);
    return 0;
  err:
    fprintf(stderr, "Error in __aud_txn_end\n");
    exit(1);
}

static AUD_TOL_ENTRY *
__aud_tol_find(aud, txnid)
    AUD *aud;
    u_int32_t txnid;
{
    AUD_TOL_ENTRY *tolep;
    for (tolep = SH_TAILQ_FIRST(&aud->tol, __aud_tol_entry);
	 (tolep != NULL) && (tolep->txnid != txnid);
	 tolep = SH_TAILQ_NEXT(tolep, links, __aud_tol_entry));
    return tolep;
}

/* During mpfget(), page is scaned, so we hash tuples on it at
 * the same time.
 * If the tuple is changed, return 1. Otherwise return 0.
 */
int __aud_tuple_hash(dbenv, dbt_data)
    DB_ENV *dbenv;
    DBT dbt_data;
{
    int changed = 0;
    TTDATA ttdata;
    char *data;
    int data_size;
    char digest[AUD_HASH_LEN*2];
    AUD_TOL_ENTRY *tol_entry;
    DB_AUD *dbaud;
    AUD *aud;

#ifndef __AUD
    return 0;
#endif
//    printf("WRONG!");exit(1);

    memcpy (&ttdata, (char *)dbt_data.data, sizeof (TTDATA));
    data = ((char *)(dbt_data.data)) + sizeof(TTDATA);
    data_size = dbt_data.size - sizeof(TTDATA);

/*
printf("enter __aud_tuple_hash: TTDB_DBT_AUD(ttdata.state)=%d\n",
       TTDB_DBT_AUD(ttdata.state));
printf("**********************%s\n", data);
*/

    if (!TTDB_DBT_AUD(ttdata.state))
	return 0;
    
    dbaud = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->aud_handler;
    if (dbaud == NULL)
        return 0;
    aud = dbaud->reginfo.primary;

    R_LOCK(dbenv, &dbaud->reginfo);

    if (TTDB_DBT_HASH1(ttdata.state)) {
	/* If the start_time is not ID now, it's wrong */
	if (TTDB_DBT_STATE(ttdata.state) != ID_UC) {
	    fprintf(stderr, "1-");
	    goto done1;
	    goto err;
	}

	/**************************************/
	/* Hash the tuple on to the start txn */
	if ((tol_entry = __aud_tol_find(aud, 
					ttdata.start.stamperId.txnid)) == NULL) {
	    /* See the comment below for "7-" */
	    fprintf(stderr, "3-");
	    goto done1;
	    goto err;
	}

	/* This is changed into commutative hashing
	if (__aud_hash(data, data_size, digest) != 0) {
	    fprintf(stderr, "2-");
	    goto err;
	}

	if (__aud_xor(tol_entry->hash_value, digest) != 0) {
	    fprintf(stderr, "4-");
	    goto err;
	}
	*/
	/* Commutative hashing
	__aud_comm_hash(data, data_size, tol_entry->hash_value, AUD_HASH_LEN, digest);
	memcpy(tol_entry->hash_value, digest, AUD_HASH_LEN);
	*/
/*	
printf("++>tuple_hash (start time): %s :txn(---,%d),seq=%d\n",
       (char *)data, tol_entry->txnid, tol_entry->seq);
*/

	/* Sequence number */
	ttdata.start_seq = tol_entry->seq;
	tol_entry->seq ++;

	/* Incremental hashing */
	__aud_hash_addtuple(tol_entry, data, data_size);

	/* Update the state */
	TTDB_DBT_CLRHASH1(ttdata.state);
	memcpy(dbt_data.data, &ttdata, sizeof(TTDATA));
	changed = 1;
    }

  done1:
    if (TTDB_DBT_HASH2(ttdata.state)) {
	/* If the start_time is not ID now, it's wrong */
	if (TTDB_DBT_STATE(ttdata.state) != TIME_ID) {
	    fprintf(stderr, "5-%d", TTDB_DBT_STATE(ttdata.state));
	    goto done;
	    goto err;
	}

	/*************************************/
	/* Hash the tuple on to the stop txn */
	if ((tol_entry = __aud_tol_find(aud, ttdata.stop.stamperId.txnid)) == NULL) {
	    /* The tuple was hashed, the txn committed, then before writing the
	     * page to the disk, it crashed. Restart, redo the page, which restore
	     * the tuple to the unhashed state :(
	     * ---> Now simply ignore this. Later think about changing the state
	     * without causing extra page write!!!!*/
	    fprintf(stderr, "7-");
	    goto done;
	    goto err;
	}

	/* This is changed into commutative hashing
	if (__aud_hash(data, data_size, digest) != 0) {
	    fprintf(stderr, "6-");
	    goto err;
	}

	if (__aud_xor(tol_entry->hash_value, digest) != 0) {
	    fprintf(stderr, "8-");
	    goto err;
	}
	*/
	/* Commutative hashing
	__aud_comm_hash(data, data_size, tol_entry->hash_value, AUD_HASH_LEN, digest);
	memcpy(tol_entry->hash_value, digest, AUD_HASH_LEN);
	*/
/*
printf("++>tuple_hash (stop time): %s :txn(---,%d),seq=%d\n",
       (char *)data, tol_entry->txnid, tol_entry->seq);
*/

	/* Sequence number */
	ttdata.stop_seq = tol_entry->seq;
	tol_entry->seq ++;

	/* Incremental hashing */
	__aud_hash_addtuple(tol_entry, data, data_size);

	/* Update the state */
	TTDB_DBT_CLRHASH2(ttdata.state);
	memcpy(dbt_data.data, &ttdata, sizeof(TTDATA));
	changed = 1;
    }

  done:
    R_UNLOCK(dbenv, &dbaud->reginfo);
    return changed;
  err:
    fprintf(stderr, "Error in __aud_tuple_hash\n");
    fflush(stdout);
    exit(1);
}

int
__aud_get_nrequest(dbenv, hash_value, mtime)
    DB_ENV *dbenv;
    char *hash_value;
    TV *mtime;
{
    DB_AUD *dbaud;
    AUD *aud;

#ifndef __AUD
    return 0;
#endif

    dbaud = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->aud_handler;
    if (dbaud == NULL)
        return 1;
    aud = dbaud->reginfo.primary;

    R_LOCK(dbenv, &dbaud->reginfo);

    memcpy(hash_value, aud->hash_value, AUD_HASH_LEN);
    *mtime = aud->hash_tail_time;
//    aud->ready_to_hash = 0;

    R_UNLOCK(dbenv, &dbaud->reginfo);

    return 0;
}

int __aud_rec_hash(dbenv)
    DB_ENV *dbenv;
{
    DB_AUD *dbaud;
    AUD *aud;
    AUD_TOL_ENTRY *tolep, *curr_tolep, *prev_tolep;
    AUD_TXN_HASH_DATA txn_hash_data;

#ifndef __AUD
    return 0;
#endif

    dbaud = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->aud_handler;
    if (dbaud == NULL)
	return 0;
    aud = dbaud->reginfo.primary;

/*
    printf("ENTER __aud_rec_hash\n");
print_tol(dbenv);
*/

    /* Since the recovery has restore the TOL, we hash them all */
    for (tolep = SH_TAILQ_FIRST(&aud->tol, __aud_tol_entry);
	 (tolep != NULL) && (tolep->ready);
	 tolep = SH_TAILQ_FIRST(&aud->tol, __aud_tol_entry)) {

	/* Do the link-hashing !!! */
	/* Update the current hash-chain tail value and tail time
	 * We assume every TOL entry is ready to be hashed
	 */
	memcpy(txn_hash_data.prev_hash_value, aud->hash_value, AUD_HASH_LEN);
	txn_hash_data.time = tolep->time;
	memcpy(txn_hash_data.hash_value, tolep->hash_value, AUD_HASH_LEN);
	if (__aud_hash(&txn_hash_data, sizeof(AUD_TXN_HASH_DATA),
		       aud->hash_value) != 0)
	    goto err;
	aud->hash_tail_time = tolep->time;

/*
printf("Hash time(%d, %d)\n", tolep->time.tv_sec, tolep->time.tv_usec);
printf("  tail_hash=");
print_hash(aud->hash_value);
*/

	/* Log the update to the current hash-chain tail (including txnid) */
        /* We don't have to log here, because right after this recovery there
	 *   will be a checkpoint that records everything.
	 */
 

	/* Delete the hashed TOL entry */
	SH_TAILQ_REMOVE(&aud->tol, tolep, links, __aud_tol_entry);
	SH_TAILQ_INSERT_HEAD(&aud->free_tol_entries, tolep,
			     links, __aud_tol_entry);
    }
  done:
print_tol(dbenv);
printf("EXIT the __aud_rec_hash\n");
    return 0;
  err:
    fprintf(stderr, "Error in __aud_txn_end\n");
    exit(1);
}

int
__aud_io(int flag)
{
    static long cnt = 0;
    if (flag == 0) {
	cnt ++;
    } else if (flag == 1) {
	return cnt;
    }
    return cnt;
}

/* 3/19/2007 Added by Huilong Huang
 * API for notarizor to get the time and hash
 * value for specified database
 */
int __aud_get_time_hash(char *path, char *dbname, char *hash_value, struct timeval *ttime, struct timeval *cur_time)
{
    DB_ENV *dbenv;
    DB *db;
    DB_STP      *dbstp;
    int ret;
    int    flags_env = DB_JOINENV | DB_THREAD;
    int    flags_db  = DB_TEMPORAL|DB_AUDIT;

    /* Create and open the environment */
    if ((ret = db_env_create(&dbenv, 0)) != 0) {
        printf("Error in __aud_get_time_hash: fail in db_env_create()\n");
        return -1;
    }
    if ((ret = dbenv->open(dbenv, path, flags_env,
                           S_IRUSR | S_IWUSR)) != 0){
        printf("Error in __aud_get_time_hash: fail in db environment open\n");
        return -1;
    }

    /* Open the database */
    if ((ret = db_create(&db, dbenv, 0)) != 0) {
        printf("Error in  __aud_get_time_hash: fail in db_create()\n");
        return -1;
    }
    if ((ret = db->open(db, dbname, NULL, DB_BTREE, flags_db, 
                           S_IRUSR | S_IWUSR)) != 0) {
        printf("Error in  __aud_get_time_hash: fail in db open\n");
        return -1;
    }

    /* Prevent the transaction from committing */
    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;
    R_LOCK(dbenv,&dbstp->reginfo);
    
    /* Get the hash value */
    if (__aud_get_nrequest(dbenv, hash_value, ttime) != 0)
    {
        printf("Error in __aud_get_nrequest()!\n");
        return -1;
    }

    /* Get current time */
    gettimeofday(cur_time,NULL);

    /* Unlock the database */ 
    R_UNLOCK(dbenv, &dbstp->reginfo);
     
    /* close database and the environment */
    db->close(db, 0);
    dbenv->close(dbenv, 0);
    
    return 0;
}

