#include "db_config.h"

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/* Original BDB system */
#include "db_int.h"
#include "db_shash.h"
#include "lock.h"
#include "db_page.h"
#include "mp.h"
#include "qam.h"
#include "log.h"

/* TT support */
#include "db.h"
#include "TTDB.h"
#include "clk_txn.h"
#include "stp_int.h"
#include "aud_int.h"
#include "temporal.h"

#include "sha1.h"
#include "memchunk.h"

int
__aud_hash(data, data_size, digest)
    char *data;
    int data_size;
    char *digest;
{
    sha1Param param;
    if (sha1Reset(&param))                                                    
	return -1;                                                        
    if (sha1Update(&param, data, data_size))              
	return -1;                                                        
    if (sha1Digest(&param, digest))                                           
	return -1;
    return 0;
}

int
__aud_xor(target_value, new_value)
    char *target_value;
    char *new_value;
{
    int i;
    for (i=0; i<AUD_HASH_LEN; i++)
	target_value[i] ^= new_value[i];
    return 0;
}

/****************************************************************
 * Commutative hashing (not used any more!!!)
 ****************************************************************/
void min_max(char *d1, int s1, char *d2, int s2, char *data)
{
    int i, j;
    int min_d1 = 1;
    /* compare d1 and d2 */
    if (s1 > s2)
	min_d1 = 0;
    else if (s1 < s2)
	min_d1 = 1;
    else if (memcmp(d1, d2, s1) <= 0)
	min_d1 = 1;
    else
	min_d1 = 0;

    if (min_d1) {
	memcpy(data, d1, s1);
	memcpy(data+s1, d2, s2);
    } else {
	memcpy(data, d2, s2);
	memcpy(data+s2, d1, s1);
    }
}

int __aud_comm_hash(char *d1, int s1, char *d2, int s2, char *digest)
{
    char *data = (char *)malloc(s1 + s2);
    if (data == NULL) {
	fprintf(stderr, "malloc ERROR in __aud_comm_hash\n");
	exit(1);
    }
    min_max(d1, s1, d2, s2, data);
    __aud_hash(data, s1 + s2, digest);
    return 0;
}

/************************************************************
 * Incremental hashing
 ************************************************************/
int
__aud_hash_newtxn(AUD_TOL_ENTRY *tol_entry)
{
    sha1Param *param;
    param = (sha1Param *)malloc(sizeof(sha1Param));
    if (sha1Reset(param))
	return -1;
    tol_entry->hash_param = param;
    return 0;
}

int
__aud_hash_addtuple(AUD_TOL_ENTRY *tol_entry, char *data, int data_size)
{
    sha1Param *param = tol_entry->hash_param;
    if (param == NULL)
	return -1;
    if (sha1Update(param, data, data_size))
	return -1;
    /* hash_value is all 0 means there is no tuple hashed to
     *   this transaction. We set hash_value[0] to 1 to indicate
     *   there is tuple hashed to this txn and this txn is not empty.
     */
    tol_entry->hash_value[0] = 1;

    return 0;
}

static const int hinit[5] = {
    0x67452301U,
    0xefcdab89U,
    0x98badcfeU,
    0x10325476U,
    0xc3d2e1f0U
};

int
__aud_hash_finish(AUD_TOL_ENTRY *tol_entry)
{
    sha1Param *param = tol_entry->hash_param;
    if (param == NULL)
	return -1;

    /* See the comment in __aud_hash_addtuple() */
    if (tol_entry->hash_value[0] == 0) {
	/* printf("**************empty txn!!\n"); */
	return 0;
    }

    if (sha1Digest(param, tol_entry->hash_value))                                           
	return -1;

    free(param);
    tol_entry->hash_param = NULL;
    return 0;
}
