#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/times.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>

/* Original BDB system */
#include "db_config.h"
#include "db_int.h"
#include "db_shash.h"
#include "lock.h"
#include "db_page.h"
#include "mp.h"
#include "qam.h"
#include "log.h"

/* TT support */
#include "db.h"
/*typedef struct timeval TV;*/
#include "TTDB.h"
#include "clk_txn.h"
#include "stp_int.h"
#include "aud_int.h"
#include "temporal.h"

#define MAXLINE 100

typedef struct sockaddr_in SAI;
typedef struct sockaddr SA;

char *env_dir;

void  env_open();

DB_ENV *dbenv;
DB *db;

/* For debug */
static int
print_hash(char *h)
{
    int i;
    for (i=0; i<AUD_HASH_LEN; i++)
	printf("%d-", h[i]);
    printf("\n");
    return 0;
}

void usage ()
{
    printf ("cmd env_home N_serv_inet_addr N_serv_port\n");
    exit (1);
}

int key_comp(DB *dbp, const DBT *a, const DBT *b)
{
    TV at, bt;

    memcpy(&at, a->data, sizeof(at));
    memcpy(&bt, b->data, sizeof(bt));

    if (at.tv_sec < bt.tv_sec)
	return -1;
    else if (at.tv_sec > bt.tv_sec)
	return 1;
    else if (at.tv_usec < bt.tv_usec)
	return -1;
    else if (at.tv_usec > bt.tv_usec)
	return 1;
    else
	return 0;
}

void
db_open(char *name, int dups, int join)
{
    int ret;
    u_int32_t flags, flags1, flags0;

    flags0 = DB_CREATE|DB_TEMPORAL;
    flags1 = DB_TEMPORAL;

    if (join == 1)
	flags = flags1;
    else
	flags = flags0;

    if ((ret = db_create(&db, dbenv, 0)) != 0) {
	dbenv->err(dbenv, ret, "db_create");
	exit (1);
    }

    if ((ret = db->set_bt_compare(db, key_comp)) != 0) {
	dbenv->err(dbenv, ret, "DB->set_bt_compare");
	exit (1);
    }

    if ((ret = db->open(db, name, NULL,
			DB_BTREE, flags, S_IRUSR | S_IWUSR)) != 0) {
	dbenv->err(dbenv, ret, "db->open: %s", name);
	exit (1);
    }	
}

void
env_open(int join)
{
    int ret;
    u_int32_t flags, flags1, flags0;
    int num_retry = 0;

    flags0 = DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG |
	DB_INIT_MPOOL | DB_INIT_TXN | DB_RECOVER | DB_THREAD;
    flags1 = DB_JOINENV | DB_THREAD;

    if (join == 1)
	flags = flags1;
    else
	flags = flags0;

    /* Create the environment handle. */
    if ((ret = db_env_create(&dbenv, 0)) != 0) {
	fprintf(stderr,
		"txnapp: db_env_create: %s\n", db_strerror(ret));
	exit (1);
    }

    /* Set up error handling. */
    dbenv->set_errpfx(dbenv, "txnapp");

    /*
     * Open a transactional environment:
     *	create if it doesn't exist
     *	free-threaded handle
     *	run recovery
     *	read/write owner only
     */
  retry:
    if ((ret = dbenv->open(dbenv, env_dir, flags,
			   S_IRUSR | S_IWUSR)) != 0) {
	sleep(1);
	if (num_retry++ < 10)
	    goto retry;
	dbenv->err(dbenv, ret, "dbenv->open: %s", env_dir);
	exit (1);
    }
}

int
add_fruit(char *fruit, char *name,int time,int time2,int myid)
{
    DBT key, data;
	
    DB_TXN *tid;
	
    int ret, t_ret;
    struct timeval mytime;

    /* Initialization. */
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    key.data = fruit;
    key.size = strlen(fruit);
    data.data = name;
    data.size = strlen(name);
    data.flags = DB_DBT_MALLOC;

    sleep(time);  /* await for the time to start */
    printf("Enter add fruit\n");

    /* Begin the transaction. */
    if ((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0) {
	dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
	exit (1);
    }
    env_current(&mytime,7);
    printf("P - Current time is (((%ld,%ld))): ---------->",
	   mytime.tv_sec, mytime.tv_usec);
    printf("P - Start transaction %d(Transaction ID:) \n",myid);
	

    sleep(time2);

    env_current(&mytime,7);
    printf("P - Current time is (((%ld,%ld))): --------->",
	   mytime.tv_sec,mytime.tv_usec);
    printf("P - Put Operation in transaction %d(Transaction ID:) \n",myid);
    db->put(db, tid, &key, &data, 0); 

    switch (db->put(db, tid, &key, &data, 0)) {

    case 0:
	env_current(&mytime,7);
	        
	printf("P - Current time is (((%ld,%ld))): ------->",
	       mytime.tv_sec, mytime.tv_usec);
	printf("P - Commit transaction %d(Transaction ID: ) \n",myid);

	if ((ret = txn_commit(tid, 0)) != 0) {
	    dbenv->err(dbenv, ret, "DB_TXN->commit");
	    exit (1);
	}

	memp_sync(dbenv, NULL);

	break;
    case DB_LOCK_DEADLOCK:
    default:
	env_current(&mytime,7);
	printf("P - Current time is (((%ld,%ld))): ----->",
	       mytime.tv_sec, mytime.tv_usec);
	printf("P - ABORT this transaction %d() !!!!!!\n",myid);		
		
	if ((t_ret = txn_abort(tid)) != 0) {
	    dbenv->err(dbenv, t_ret, "DB_TXN->abort");
	    exit (1);
	}
	break;
    }
    return;
}

int
main(int argc, char *argv[])
{
    char hash_value[AUD_HASH_LEN];
    TV ttime, old_ttime;
    int ch, i;
    int sockfd;
    SAI servaddr;
    int n, ret;
    in_port_t port;
    NID nid;
    DBT key, data;
    DB_TXN *tid;
    DB_AUD *dbaud;
    AUD *aud;
    int cnt;
    
    if (argc != 4) {
	usage();
	exit(1);
    }
    env_dir = argv[1];
    port = atoi(argv[3]);
	
    printf("Audit Notarization Request Utility Starts ...\n");
    printf("env_home=%s, N-serv_IP=%s, port=%d\n", env_dir, argv[2], port);

    /* Establish UDP */
    memset(&servaddr, 0, sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_port = htons(port);
    inet_pton(AF_INET, argv[2], &servaddr.sin_addr);
    sockfd = socket(AF_INET, SOCK_DGRAM, 0);

    env_open(1);
    printf("Env joined\n");

    db_open("db_aud_n_history", 0, 0);
    printf("N-history db created and/or opened\n");
    
    dbaud = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->aud_handler;
    if (dbaud == NULL)
        return 1;
    aud = dbaud->reginfo.primary;

    cnt = 0;
    memset(&old_ttime, 0, sizeof(old_ttime));
    for (i=0; ;) {
	if (__aud_get_nrequest(dbenv, hash_value, &ttime) != 0)
	    goto err;
	printf("oldtime=(%d,%d), newtime=(%d,%d)\n",
	       old_ttime.tv_sec, old_ttime.tv_usec,
	       ttime.tv_sec, ttime.tv_usec);
	if (memcmp(&old_ttime, &ttime, sizeof(ttime)) == 0) {
	    if (cnt < 2) {
		cnt ++;
		R_LOCK(dbenv, &dbaud->reginfo);
		memcpy(aud->hash_value, nid.nid, AUD_HASH_LEN);
		aud->ready_to_hash = 1;
		R_UNLOCK(dbenv, &dbaud->reginfo);
		goto retry;
	    }
	    printf("No more transactions! Exit\n");
	    break;
	}
	cnt = 0;

	memcpy(&old_ttime, &ttime, sizeof(ttime));

	sendto(sockfd, hash_value, AUD_HASH_LEN, 0, 
	       (SA *)&servaddr, sizeof(servaddr));
	printf("Request sent: ttime=(%u, %u), hash=",
	       ttime.tv_sec, ttime.tv_usec);
	print_hash(hash_value);

	n = recvfrom(sockfd, &nid, sizeof(nid), 0, NULL, NULL);
	printf("  N-id received: ntime=(%u, %u), nid_hash=", 
	       nid.ntime.tv_sec, nid.ntime.tv_usec);
	print_hash(nid.nid);

	R_LOCK(dbenv, &dbaud->reginfo);
	memcpy(aud->hash_value, nid.nid, AUD_HASH_LEN);
	aud->ready_to_hash = 1;
	R_UNLOCK(dbenv, &dbaud->reginfo);

	/* Initialization. */
	memset(&key, 0, sizeof(key));
	memset(&data, 0, sizeof(data));
	key.data = &ttime;
	key.size = sizeof(ttime);
	data.data = &nid;
	data.size = sizeof(nid);

	/* Begin the transaction. */
	if ((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0) {
	    dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
	    exit (1);
	}

	if (db->put(db, tid, &key, &data, 0) != 0) {
	    dbenv->err(dbenv, ret, "DB->put");
	    exit (1);
	}

	if ((ret = txn_commit(tid, 0)) != 0) {
	    dbenv->err(dbenv, ret, "DB_TXN->commit");
	    exit (1);
	}

	printf("  N-id stored in the history table.\n");
      retry:
	sleep(5);
    }
    
    db->close(db, 0);
    dbenv->close(dbenv, 0);

    {
        struct tms TimeReading;
        times(&TimeReading);
        printf("CPU Time=%f sec\n", ((TimeReading.tms_stime + TimeReading.tms_utime) / 100.0));
    }
    printf("IO=%d\n", __aud_io(1));

    printf("Stop.\n");
    return (0);
  err:
    fprintf(stderr, "Error in main()!\n");
    exit(1);
}
