#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/times.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>

/* Original BDB system */
#include "db_config.h"
#include "db_int.h"
#include "db_shash.h"
#include "lock.h"
#include "db_page.h"
#include "mp.h"
#include "qam.h"
#include "log.h"

/* TT support */
#include "db.h"
#include "TTDB.h"
#include "clk_txn.h"
#include "stp_int.h"
#include "aud_int.h"
#include "temporal.h"
#include "heap.h"
#include "sha1.h"
#include "memchunk.h"

#include "red_black_tree.h"

#define MAXLINE 100

typedef struct sockaddr_in SAI;
typedef struct sockaddr SA;

char *env_dir;

void  env_open();

DB_ENV *dbenv;
DB *db;

struct __ver_tol_entry {
    TV time;
    char hash_value[AUD_HASH_LEN];
    void *hash_param; /* hash function state */
    Heap_T tuple_heap; /* priority queue */
    int expected_seq; /* next expected sequence number */
    /*struct __ver_tol_entry *next;*/
};
typedef struct __ver_tol_entry VER_TOL_ENTRY;

struct __tuple {
    int seq; /* indexing sequence number in the priority tuque. */
    DBT data;
};
typedef struct __tuple TUPLE;

/*
VER_TOL_ENTRY *tol;
*/
rb_red_blk_tree *tree; /* TOL */

/* For debug
static int
print_tol()
{
    VER_TOL_ENTRY *p;
    printf("+++++++++++++++++++++ Rebuit TOL +++++++++++++++++\n");
    p = tol->next;
    while (p) {
	printf("Time=(%d, %d), hash=", p->time.tv_sec, p->time.tv_usec);
	print_hash(p->hash_value);
	p = p->next;
    }
    printf("++++++++++++++++++++++++++++++++++++++++++++++++++\n");
    return 0;
}
*/
static int
print_tol()
{
    VER_TOL_ENTRY *p;
    rb_red_blk_node *np;
    printf("+++++++++++++++++++++ Rebuit TOL +++++++++++++++++\n");
    np = RBMin(tree);
    while (np != NULL) {
	p = (VER_TOL_ENTRY *)(np->info);
	if (p == NULL)
	    break;
	printf("Time=(%d, %d), hash=", p->time.tv_sec, p->time.tv_usec);
	print_hash(p->hash_value);
	np = TreeSuccessor(tree, np);
    }
    printf("++++++++++++++++++++++++++++++++++++++++++++++++++\n");
    return 0;
}

void usage ()
{
    printf ("cmd env_home N_serv_inet_addr N_serv_port\n");
    exit (1);
}

int key_comp(DB *dbp, const DBT *a, const DBT *b)
{
    TV *at = a->data,
	*bt = b->data;
    if (at->tv_sec < bt->tv_sec)
	return -1;
    else if (at->tv_sec > bt->tv_sec)
	return 1;
    else if (at->tv_usec < bt->tv_usec)
	return -1;
    else if (at->tv_usec > bt->tv_usec)
	return 1;
    else
	return 0;
}

#define time_cmp(at, bt) ((at.tv_sec==bt.tv_sec)? (at.tv_usec-bt.tv_usec): (at.tv_sec-bt.tv_sec))
/* bottle neck!!!
int time_cmp(TV at, TV bt)
{
    if (at.tv_sec < bt.tv_sec)
	return -1;
    else if (at.tv_sec > bt.tv_sec)
	return 1;
    else if (at.tv_usec < bt.tv_usec)
	return -1;
    else if (at.tv_usec > bt.tv_usec)
	return 1;
    else
	return 0;
}
*/

void
db_open(char *name, int dups, int join)
{
    int ret;
    u_int32_t flags, flags1, flags0;

    flags0 = DB_CREATE|DB_TEMPORAL;
    flags1 = DB_TEMPORAL;

    if (join == 1)
	flags = flags1;
    else
	flags = flags0;

    if ((ret = db_create(&db, dbenv, 0)) != 0) {
	dbenv->err(dbenv, ret, "db_create");
	exit (1);
    }

    if ((ret = db->set_bt_compare(db, key_comp)) != 0) {
	dbenv->err(dbenv, ret, "DB->set_bt_compare");
	exit (1);
    }

    if ((ret = db->open(db, name, NULL,
			DB_BTREE, flags, S_IRUSR | S_IWUSR)) != 0) {
	dbenv->err(dbenv, ret, "db->open: %s", name);
	exit (1);
    }	
}


FILE *errfh;

void
env_open(int join)
{
    int ret;
    u_int32_t flags, flags1, flags0;

    flags0 = DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG |
	DB_INIT_MPOOL | DB_INIT_TXN | /*DB_RECOVER |*/ DB_THREAD;
    flags1 = DB_JOINENV | DB_THREAD;

    if (join == 1)
	flags = flags1;
    else
	flags = flags0;

    /* Create the environment handle. */
    if ((ret = db_env_create(&dbenv, 0)) != 0) {
	fprintf(stderr,
		"txnapp: db_env_create: %s\n", db_strerror(ret));
	exit (1);
    }

    /* Set up error handling. */
    dbenv->set_errpfx(dbenv, "txnapp");

    /*
    if ((errfh = open("/tmp/errorDebugInfo", "w")) == NULL) {
	printf("Can't open err file\n");
	exit(1);
    }
    */
    dbenv->set_errfile(dbenv, stderr);

    /*
    if ((ret = dbenv->set_lk_max_locks(dbenv, 8000)) != 0) {
	dbenv->err(dbenv, ret, "dbenv->set_lk_max_locks: %s", env_dir);
	exit (1);
    }

    if ((ret = dbenv->set_lk_max_lockers(dbenv, 8000)) != 0) {
	dbenv->err(dbenv, ret, "dbenv->set_lk_max_lockers: %s", env_dir);
	exit (1);
    }

    if ((ret = dbenv->set_lk_max_objects(dbenv, 8000)) != 0) {
	dbenv->err(dbenv, ret, "dbenv->set_lk_max_objects: %s", env_dir);
	exit (1);
    }
    */

    /*
     * Open a transactional environment:
     *	create if it doesn't exist
     *	free-threaded handle
     *	run recovery
     *	read/write owner only
     */
    if ((ret = dbenv->open(dbenv, env_dir, flags,
			   S_IRUSR | S_IWUSR)) != 0) {
	dbenv->err(dbenv, ret, "dbenv->open: %s", env_dir);
	exit (1);
    }

}

static int cmp(const void *a, const void *b) {
    TUPLE *t1 = (TUPLE *)a,
	*t2 = (TUPLE *)b;
    return (t1->seq - t2->seq);
}

int timebound = 0;

int
rebuild_tol(char *dbnames[], int num_dbs)
{
    DB *db;
    int ret;
    u_int32_t flags;
    DBT key, data;
    TUPLE *tuple1, *tuple2;
    DBC *dbc;
    DB_TXN *tid;
    int start1, stop1;
    TV start, stop, ret_start, ret_stop;
    VER_TOL_ENTRY *p, *p1, *tmp;    
    int dbi;
    char digest[AUD_HASH_LEN];
    TTDATA *ttdata;
    int counter = 0;
    rb_red_blk_node *np;

    flags = DB_TEMPORAL;

printf("ENTER rebuild_tol(): num_dbs=%d\n", num_dbs);
    if ((ret = db_create(&db, dbenv, 0)) != 0) {
	dbenv->err(dbenv, ret, "db_create");
	exit (1);
    }

    for (dbi=0; dbi<num_dbs; dbi+=2) {
	/* For the prototype, we don't work on subdatabases.
	 * Supporting subdatabases would be just trivial.
	 */
	if ((ret = db->open(db, dbnames[dbi], dbnames[dbi+1],
			    DB_BTREE, flags, S_IRUSR | S_IWUSR)) != 0) {
	    dbenv->err(dbenv, ret, "db->open: %s", dbnames[dbi]);
	    exit (1);
	}	
    
printf("rebuild_tol(): file=%s, database=%s\n",
       dbnames[dbi], ((dbnames[dbi+1] == NULL)? "": dbnames[dbi+1]));

	/* Begin the transaction. */
	if ((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0) {
	    dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
	    exit (1);
	}

	if (db->cursor(db, tid, &dbc, 0) != 0) {
	    dbenv->err(dbenv, ret, "DB->cursor");
	    exit (1);
	}

	/* Initialize the key/data return pair. */
	memset(&key, 0, sizeof(key));
	memset(&data, 0, sizeof(data));
	data.flags = DB_DBT_MALLOC;

	/* Scan the entire database */
	//start.tv_sec = BEGINNING;
	start.tv_sec = timebound;
	start.tv_usec = 0;
	stop.tv_sec = FOREVER;
	stop.tv_usec = 0;

	while ((ret = (dbc->c_get_nonsequenced(dbc, &key, &data,
					       start, stop,
					       DB_NEXT,
					       &ret_start, &ret_stop))) == 0) {
	    printf("    rebuild_tol(): Found a tuple: %s\n", (char *)data.data);
	    counter ++;

/****************************/
	    continue;

	    //RBTreePrint(tree);

	    /* Make two copies of the "data" DBT. One is put in the
	     *   starting transaction's tuple priority queue. The other
	     *   is put in the stopping transaction's tuple priority queue.
	     * When any copy of the "data" tuple is hashed, free its space.
	     */
	    tuple1 = (TUPLE *)malloc(sizeof(TUPLE));
	    tuple1->data.data = (void *)malloc(data.size);
	    tuple1->data.size = data.size;
	    tuple1->seq = ((TTDATA *)(data.data - sizeof(TTDATA)))->start_seq;
	    memcpy(tuple1->data.data, data.data, data.size);

	    tuple2 = (TUPLE *)malloc(sizeof(TUPLE));
	    tuple2->data.data = (void *)malloc(data.size);
	    tuple2->data.size = data.size;
	    tuple2->seq = ((TTDATA *)(data.data - sizeof(TTDATA)))->stop_seq;
	    memcpy(tuple2->data.data, data.data, data.size);
	    
	    free(data.data);

            /* Assertion: if the stop seq number is 0, the stop time must be UC! */ 
            if ((tuple2->seq == 0) && !is_until_changed(ret_stop)) {
		fprintf(stderr,
			"Fail: Assertion: if the stop seq number is 0, the stop time must be UC!\n");
		goto err;
	    }


	    if (ret_start.tv_sec < timebound)
		goto ignorestart;
//	    goto ignorestart;

//printf("#"); fflush(stdout);
	    /* Search for start time on the rebuilt tol
	    p = tol;
	    while ((p1 = p->next) != NULL) {
		if (time_cmp(ret_start, p1->time) <= 0)
		    break;
		p = p1;
	    }
	    */
	    np = RBExactQuery(tree, &ret_start);
	    if (np == NULL)
		p1 = NULL;
	    else
		p1 = (VER_TOL_ENTRY *)(np->info);

	    if (p1 == NULL) {
		/* Create a new node on tol */
		if ((tmp = malloc(sizeof(VER_TOL_ENTRY))) == NULL)
		    goto err;
		tmp->time = ret_start;
		memset(tmp->hash_value, 0, AUD_HASH_LEN);
		{
		    sha1Param *param;
		    param = (sha1Param *)malloc(sizeof(sha1Param));
		    if (sha1Reset(param))
			goto err;
		    tmp->hash_param = param;
		}
// The max number of tuples per transaction
#define MAX_TUPLES 200
		tmp->tuple_heap = Heap_new(MAX_TUPLES, cmp, NULL);
		tmp->expected_seq = 1;

		if ((np = RBTreeInsert(tree, &(tmp->time), tmp)) == NULL) {
		    printf("RBTreeInsert error\n");
		    exit(1);
		}
//		goto ignorestart;
		p1 = (VER_TOL_ENTRY *)(np->info);
	    }

	    /* Here p1 points to the node which has the same time
	     * as ret_start.
	     */

	    /* Hash the data onto it
	    __aud_hash(data.data, data.size, digest);
	    __aud_xor(p1->hash_value, digest);
	     */
	    /*
	    __aud_comm_hash(data.data, data.size, p1->hash_value, AUD_HASH_LEN, digest);
	    memcpy(p1->hash_value, digest, AUD_HASH_LEN);
	    */

	    /* Put the tuple in the "tuple priority queue" of the
	     *   transaction.
	     */
	    if (Heap_insert(p1->tuple_heap, (void *)tuple1) != 0)
		goto err;
/*
printf("start time: tuple seq=%d put in queue, p1->expected_seq=%d\n",
       tuple1->seq, p1->expected_seq);
*/

	    /* See if we can incrementally hash some tuples for the transaction until
	     *   the next hole of unseen tuple.
	     */
	    while (1) {
		TUPLE *tuple_tmp;
		tuple_tmp = (TUPLE *)Heap_top(p1->tuple_heap);

		/* When tuple_tmp is NULL, there is no tuple in the queue */
		if (tuple_tmp == NULL)
		    break;

		if (tuple_tmp->seq != p1->expected_seq)
		    break;
/*
printf("(hash start: txn_time=(%d,%d) seq=%d, expected=%d)\n",
       p1->time.tv_sec, p1->time.tv_usec,
       tuple_tmp->seq, p1->expected_seq);
*/
		/* This tuple's start_seq is our expected next sequence number.
		 *   So we incrementally hash it to this transaction.
		 *   And remove it from the transaction's priority queue;
		 */
		if (sha1Update(p1->hash_param, tuple_tmp->data.data, tuple_tmp->data.size))
		    goto err;
		Heap_extract_min(p1->tuple_heap);

		/* After hashing this tuple, we can free the space of this tuple */
		free(tuple_tmp->data.data);
		free(tuple_tmp);

		p1->expected_seq ++;
	    }

	  ignorestart:

	    /* Search for stop time on the rebuilt tol */
	    if (!is_until_changed(ret_stop)) {
		if (ret_stop.tv_sec < timebound)
		    goto ignorestop;
//		goto ignorestop;
//printf("#"); fflush(stdout);
		/*
		while ((p1 = p->next) != NULL) {
		    if (time_cmp(ret_stop, p1->time) <= 0)
			break;
		    p = p1;
		}
		*/
		np = RBExactQuery(tree, &ret_stop);
		if (np == NULL)
		    p1 = NULL;
		else
		    p1 = (VER_TOL_ENTRY *)(np->info);

		if (p1 == NULL) {
		    /* Create a new node on tol */
		    if ((tmp = malloc(sizeof(VER_TOL_ENTRY))) == NULL)
			goto err;
		    tmp->time = ret_stop;
		    memset(tmp->hash_value, 0, AUD_HASH_LEN);
		    {
			sha1Param *param;
			param = (sha1Param *)malloc(sizeof(sha1Param));
			if (sha1Reset(param))
			    goto err;
			tmp->hash_param = param;
		    }
		    tmp->tuple_heap = Heap_new(MAX_TUPLES, cmp, NULL);
		    tmp->expected_seq = 1;

		    if ((np = RBTreeInsert(tree, &(tmp->time), tmp)) == NULL) {
			printf("RBTreeInsert error\n");
			exit(1);
		    }
		    p1 = (VER_TOL_ENTRY *)(np->info);
		}

		/* Here p1 points to the node which has the same time
		 * as ret_start
		 */

		/* Hash the data onto it
		__aud_hash(data.data, data.size, digest);
		__aud_xor(p1->hash_value, digest);
		 */
		/*
		__aud_comm_hash(data.data, data.size, p1->hash_value, AUD_HASH_LEN, digest);
		memcpy(p1->hash_value, digest, AUD_HASH_LEN);
		*/

		/* Put the tuple in the "tuple priority queue" of the
		 *   transaction.
		 */
		if (Heap_insert(p1->tuple_heap, (void *)tuple2) != 0)
		    goto err;
/*
printf("stop time: tuple seq=%d put in queue, p1->expected_seq=%d\n",
       tuple1->seq, p1->expected_seq);
*/
		/* See if we can incrementally hash some tuples until
		 *   the next hole of unseen tuple.
		 */
		while (1) {
		    TUPLE *tuple_tmp;
		    tuple_tmp = (TUPLE *)Heap_top(p1->tuple_heap);

		    /* When tuple_tmp is NULL, there is no tuple in the queue */
		    if (tuple_tmp == NULL)
			break;

		    if (tuple_tmp->seq != p1->expected_seq)
			break;
/*
printf("(hash stop: txn_time=(%d,%d) seq=%d, expected=%d)\n",
       p1->time.tv_sec, p1->time.tv_usec,
       tuple_tmp->seq, p1->expected_seq);
*/
		    /* This tuple's start_seq is our expected next sequence number.
		     *   So we incrementally hash it to this transaction.
		     *   And remove it from the transaction's priority queue;
		     */
		    if (sha1Update(p1->hash_param, tuple_tmp->data.data, tuple_tmp->data.size))
			goto err;
		    Heap_extract_min(p1->tuple_heap);

		    /* After hashing this tuple, we can free the space of this tuple */
		    free(tuple_tmp->data.data);
		    free(tuple_tmp);

		    p1->expected_seq ++;
		}
	      ignorestop:
	    }

	    /*
	    if (counter == 11989)
		sleep(1);
	    */
	}
	printf("ret = %d\n", ret);
	dbenv->err(dbenv, ret, "population get_nonsequenced");
	
	if (dbc->c_close(dbc) != 0) {
	    dbenv->err(dbenv, ret, "DBC->c_close");
	    exit (1);
	}
    
	if ((ret = txn_commit(tid, 0)) != 0) {
	    dbenv->err(dbenv, ret, "DB_TXN->commit");
	    exit (1);
	}

	if (db->close(db, 0) != 0) {
	    dbenv->err(dbenv, ret, "db->close: %s", dbnames[dbi]);
	    exit (1);
	}
    }    

    /* Tuple sorting and incremental hashing is done.
     *   Now we finish the digest and copy the digest into the
     *   p->hash_value of each entry of the result TOL.
     *   Also free the space used to store the hash states (param).
     *   The tuple heap (priority queue) is deallocated too!
    p = tol;
    while ((p = p->next) != NULL) {
	    sha1Param *param;

	    param = p->hash_param;
	    if (param == NULL)
		goto err;
	    
	    if (sha1Digest(param, p->hash_value))                                           
		goto err;
	    
	    free(param);
	    p->hash_param = NULL;

	    Heap_free(&(p->tuple_heap));
    }
     */

/*******************************
    np = RBMin(tree);
    while (np != NULL) {
	sha1Param *param;

//printf(".");fflush(stdout);
	
	p = (VER_TOL_ENTRY *)(np->info);
	if (p == NULL)
	    break;

	param = p->hash_param;
	if (param == NULL)
	goto err;
	
	if (sha1Digest(param, p->hash_value))                                           
	goto err;
	
	free(param);
	p->hash_param = NULL;
	
	Heap_free(&(p->tuple_heap));

	np = TreeSuccessor(tree, np);
    }
**********************************/

    printf("******************>>>>> COUNTER = %d\n", counter);

    return 0;
  err:
    fprintf(stderr, "ERROR in rebuild_tol\n");
    exit(1);
}

void mydest(void *a) {free((VER_TOL_ENTRY *)a);}
int mycomp(const void *a, const void *b)
{return time_cmp((((VER_TOL_ENTRY *)a)->time), (((VER_TOL_ENTRY *)b)->time));}
void myprint(const void *a) {printf("<%d,%d>", ((TV *)a)->tv_sec, ((TV *)a)->tv_usec);}
void myinfoprint(void *a) {return;}
void myinfodest(void *a) {return;}

int
main(int argc, char *argv[])
{
    char hash_value[AUD_HASH_LEN];
    TV ttime;
    int ch, i;
    int sockfd;
    SAI servaddr;
    int n;
    in_port_t port;
    NID nid, rnid;
    DBT key, data;
    DB_TXN *tid;
#define MAX_DBS 100
    char *dbnames[MAX_DBS];
    int num_dbs;
    DBC *dbc;
    int ret;
    VER_TOL_ENTRY *p;
    AUD_TXN_HASH_DATA txn_hash_data;
    time_t start_time;
    rb_red_blk_node *np;

    __stp_timestart();
    start_time = time(NULL);

    if (argc <= 5) {
	usage();
	exit(1);
    }
    timebound = atoi(argv[1]); /* env id before this is ignored */
    env_dir = argv[2];
    port = atoi(argv[4]);
	
    printf("Time sec bound=%d\n", timebound);
    printf("Verifier Starts ...\n");
    printf("port=%d\n", port);

    num_dbs = argc - 5;
    for (i=0; i<num_dbs; i++) {
	if (strcmp(argv[i+5], "-") == 0)
	    dbnames[i] = NULL;
	else
	    dbnames[i] = argv[i+5];
    }

    /* Init the rebuilt tol
    if ((tol = (VER_TOL_ENTRY *)malloc(sizeof(VER_TOL_ENTRY))) == NULL) {
	printf("tol = malloc()\n");
	goto err;
    }
    memset(tol, 0, sizeof(VER_TOL_ENTRY));
    tol->next = NULL;
    */
    tree = RBTreeCreate(mycomp, mydest, myinfodest, myprint, myinfoprint);

    /* Establish UDP */
    memset(&servaddr, 0, sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_port = htons(port);
    inet_pton(AF_INET, argv[2], &servaddr.sin_addr);
    sockfd = socket(AF_INET, SOCK_DGRAM, 0);
    
    env_open(0);
    printf("Env opened\n");

    db_open("db_aud_n_history", 0, 1);
    printf("N-history db opened\n");

    /* Initialization. */	
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    data.flags = DB_DBT_MALLOC;

    /* Scan all DB's and rebuild the TOL for the verification */
    rebuild_tol(dbnames, num_dbs);

//print_tol();
// exit(0);
 goto finished;

    /* Begin the transaction. */
    if ((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0) {
	dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
	exit (1);
    }

    if (db->cursor(db, tid, &dbc, 0) != 0) {
	dbenv->err(dbenv, ret, "DB->cursor");
	exit (1);
    }

    np = RBMin(tree);
    p = (np == NULL)? NULL: (VER_TOL_ENTRY *)(np->info);
    memset(hash_value, 0, AUD_HASH_LEN);
    while ((dbc->c_get(dbc, &key, &data, DB_NEXT)) == 0) {
	ttime = *((TV *)key.data);
	memcpy(&nid, data.data, sizeof(nid));

	/* Rechain the txn's in TOL till the tail time "ttime"
	 * and put the tail hash_value in "hash_value" for verification.
	 */
	if (p == NULL)
	    break;
	while (p != NULL) {
/*	    printf("TOL:txn_time(%d,%d)\n",
		   p->time.tv_sec, p->time.tv_usec);
	    printf("    txn_hash =");
	    print_hash(p->hash_value);
*/
	    memcpy(txn_hash_data.prev_hash_value, hash_value, AUD_HASH_LEN);
	    txn_hash_data.time = p->time;
	    memcpy(txn_hash_data.hash_value, p->hash_value, AUD_HASH_LEN);
	    if (__aud_hash(&txn_hash_data, sizeof(AUD_TXN_HASH_DATA), hash_value) != 0)
		goto err;

/*
	    printf("    tail_hash=");
	    print_hash(hash_value);
*/
	    if (time_cmp(p->time, ttime) == 0)
		break;

	    np = TreeSuccessor(tree, np);
	    p = (np == NULL)? NULL: (VER_TOL_ENTRY *)(np->info);
	}

	sendto(sockfd, hash_value, AUD_HASH_LEN, 0, 
	       (SA *)&servaddr, sizeof(servaddr));
/*
	printf("Request sent: ttime=(%d, %d)\n",
	       ttime.tv_sec, ttime.tv_usec);
	printf("              thash=");
	print_hash(hash_value);
*/
	n = recvfrom(sockfd, &rnid, sizeof(rnid), 0, NULL, NULL);
/*	printf("    N-id received: n=%d, ntime=(%d, %d)\n", 
	       rnid.ntime.tv_sec, rnid.ntime.tv_usec);
*/
	/* Compare the N-id from the N-serv with the one in the history table */
	if (0) {
// experiment walk around!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//	if (memcmp(rnid.nid, nid.nid, AUD_HASH_LEN) != 0) {
	    printf("  <<< Discrepancy found here!!! >>>\n");
	    printf("    Data in the nHistory:\n");
	    printf("        ttime=(%d, %d)\n", ttime.tv_sec, ttime.tv_usec);
	    printf("        ntime=(%d, %d)\n", nid.ntime.tv_sec, nid.ntime.tv_usec);
	    printf("        hash=");
	    print_hash(nid.nid);
	    printf("    Data recomputed from the database:\n");
	    printf("        hash=");
	    print_hash(hash_value);
	    exit (1);
	} else {
	    printf("=====>Verification check at txn_time (%d, %d) is OK!!!<=====\n\n\n",
		   ttime.tv_sec, ttime.tv_usec);
	}

	memcpy(hash_value, rnid.nid, AUD_HASH_LEN);
	free(data.data);
    }

    if (dbc->c_close(dbc) != 0) {
	dbenv->err(dbenv, ret, "DBC->c_close");
        exit (1);
    }
    
    if ((ret = txn_commit(tid, 0)) != 0) {
	dbenv->err(dbenv, ret, "DB_TXN->commit");
	exit (1);
    }

    if (db->close(db, 0) != 0) {
	dbenv->err(dbenv, ret, "db->close: history table");
	exit (1);
    }

  finished:
    {
        struct tms TimeReading;
        times(&TimeReading);
        printf("CPU Time=%f sec\n", ((TimeReading.tms_stime + TimeReading.tms_utime) / 100.0));
    }
    printf("IO=%d\n", __aud_io(1));

    printf("Verification successfully done. No discrepancy found!!! :) :) :)\n");

    printf("Total time=%d\n", time(NULL) - start_time);


    __stp_timestop();
    __stp_timereport();
    return (0);
  err:
    fprintf(stderr, "Error in main()!\n");
    exit(1);
}
