 /*-
 * Developed (2002-2003) by:
 *     TAU Project
 *     Computer Science Dept., University of Arizona
 *     Tucson, AZ 85721, USA
 *     URL  : http://www.cs.arizona.edu/tau
 *     Email: tau@cs.arizona.edu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
/*-
 * Transaction Time Support to BDB - Stamper Subsystem
 *
 * By     : Shilong Yao           (yao@CS.Arizona.EDU)
 *          Yuhong Liu            (yuhongl@CS.Ariozna.EDU)
 * Advisor: Dr. Richard Snodgrass (rts@CS.Arizona.EDU)
 * Computer Science Dept., The University of Arizona
 *
 * 04/08/02: Create
 *     Only BTree is implemented. Hooks for other page layout are provided.
 * 06/20/02: Version 2 Begin Coding
 * 12/27/02: Version 2 finished
 * 04/09/05: version 3 Finished
 */

/***************************************************************************
 *                                                                         *
 * STP (Transaction Time Stamper) module is a lower level module embedded  *
 * in the BerkeleyDB system to timestamp the data records of the           *
 * transaction-time supported database. This file implements functions to  *
 * display the TT table, BT table, and to do the research experiments:     *
 *     int printTT(DB_ENV *dbenv)                                          *
 *     int print_bt(DB_STP *dbstp)                                         *
 *     int timeval_subtract (struct timeval *result, struct timeval *x,    *
 *                      struct timeval *y)                                 *
 *     void __stp_timereset()                                              *
 *     void __stp_timestart()                                              *
 *     void __stp_timestop()                                               *
 *     void __stp_timereport()                                             *
 *     int printleafstart()                                                *
 *     int printleafstop()                                                 *
 *     int printpagenum(db_pgno_t pgno)                                    *
 *     int count_ttscan()                                                  *
 *     int report_ttscan()                                                 *
 *                                                                         *
 ***************************************************************************/

#include "db_config.h"

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/* Original BDB system */
#include "db_int.h"
#include "db_shash.h"
#include "lock.h"
#include "db_page.h"
#include "mp.h"
#include "qam.h"
#include "log.h"

/* TT support */
#include "db.h"
#include "TTDB.h"
#include "clk_txn.h"
#include "stp_int.h"
#include "stp_tt.h"
/* Huilong 7/16/06*/
#ifdef TBDB_AUDIT
#include "aud_int.h"
#endif
/* end */
#include "temporal.h"

extern DB_ENV *gdbenv;
extern int stp_pause;


/* print out the TT table -- for debugging purpose */
int
printTT(DB_ENV *dbenv)
{
    DB_STP	*dbstp;
    STP_BTN  	*btn;
    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    if (dbstp == NULL)
	return 0;
    print_tt(dbstp);
    return (0);
}

int
print_tt(dbstp)
    DB_STP	*dbstp;
{
    u_int32_t tt_index;
    STP_TT_ENTRY *tt;

    printf("\t=====TT=====\n");
    for (tt_index=0; tt_index < dbstp->tt_htab_size; tt_index++) {
	for (tt = SH_TAILQ_FIRST(&(dbstp->tt_htab[tt_index]), __stp_tt_entry);
	     tt != NULL;
	     tt = SH_TAILQ_NEXT(tt, links, __stp_tt_entry)) {

/* modified by Huilong at 17:05 7/11/2006 */
            printf("\t(bucket #%d) tid=((%d,%d),%d),time=(%d,%d),od_pgcnt=%d\n",
                tt_index, tt->txnid.envid.tv_sec,tt->txnid.envid.tv_msec , tt->txnid.txnid,

//	    printf("\t(bucket #%d) tid=(%d,%d),time=(%d,%d),od_pgcnt=%d\n",                
//		   tt_index, tt->txnid.envid, tt->txnid.txnid,
/* end of modification */
		   tt->time.tv_sec, tt->time.tv_usec, tt->od_pgcnt);
	    fflush(stdout);
	    print_btn_v(dbstp, tt);
	}
    }
    printf("\t============\n");
}

int
print_bt(dbstp)
    DB_STP	*dbstp;
{
    u_int32_t bt_index, i;
    STP_BT_ENTRY *bt;

    printf("\t=====BT=====\n");
    for (bt_index=0; bt_index < dbstp->bt_htab_size; bt_index++) {
	for (bt = SH_TAILQ_FIRST(&(dbstp->bt_htab[bt_index]), __stp_bt_entry);
	     bt != NULL;
	     bt = SH_TAILQ_NEXT(bt, links, __stp_bt_entry)) {
	    printf("\t(bucket#%d)pgno=%d,fileid=", bt_index, bt->pgno);
	    for (i=0; i<DB_FILE_ID_LEN; i++)
		printf("%d,", bt->fileid[i]);
	    printf("\n");
	    fflush(stdout);
	    print_btn_h(dbstp, bt);
	}
    }
    printf("\t============\n");
}

int
print_btn_h(dbstp, bt)
    DB_STP  	*dbstp;
    STP_BT_ENTRY *bt;
{
    int begin;
    STP_BTN *btnp;
    begin = 1;
    for (btnp = SH_TAILQ_FIRST(&(bt->btn), __stp_bt_node);
	 btnp != NULL;
	 btnp = SH_TAILQ_NEXT(btnp, hlinks, __stp_bt_node)) {
	if (begin == 1) {
	    printf("\t\t=====BTN=====\n");
	    begin = 0;
	}
/* modified by Huilong at 17:05 4/17/2007 */
//	printf("\t\ttt->txnid=(%d,%d)\n",
//	       btnp->tt->txnid.envid, btnp->tt->txnid.txnid);
        printf("\t\ttt->txnid=((%d,%d)%d)\n",
               btnp->tt->txnid.envid.tv_sec, btnp->tt->txnid.envid.tv_msec, btnp->tt->txnid.txnid);
/* End Huilong */
    }
    if (begin == 0)
	printf("\t\t============\n");
}


int
print_btn_v(dbstp, tt)
    DB_STP  	*dbstp;
    STP_TT_ENTRY *tt;
{
    int begin;
    STP_BTN *btnp;
    begin = 1;
    for (btnp = SH_TAILQ_FIRST(&(tt->btn), __stp_bt_node);
	 btnp != NULL;
	 btnp = SH_TAILQ_NEXT(btnp, vlinks, __stp_bt_node)) {
	if (begin == 1) {
	    printf("\t\t=====BTN=====\n");
	    begin = 0;
	}
/* modified by Huilong at 17:05 4/17/2007 */
//	printf("\t\ttt->txnid=(%d,%d)\n",
//	       btnp->tt->txnid.envid, btnp->tt->txnid.txnid);
        printf("\t\ttt->txnid=((%d,%d)%d)\n",
               btnp->tt->txnid.envid.tv_sec, btnp->tt->txnid.envid.tv_msec, btnp->tt->txnid.txnid);
/* End Huilong */
    }
    if (begin == 0)
	printf("\t\t============\n");
}

#ifdef EXPERIMENT
/*************************************************************
 *************************************************************
 * Facilities for research experiments.                      *
 *************************************************************
 *************************************************************/

/*******************************************
 * Number of calls to STP
 *******************************************/
static int numCalls[2]; /* number of calls to STP module */

int resetcall() { numCalls[0] = numCalls[1] = 0; }
int addcall(int n) { numCalls[n] ++; }
int reportcall() { printf("\nTotal number of calls to STP is %d, %d.\n", numCalls[0], numCalls[1]); }

/*******************************************
 * CPU time at STP
 *******************************************/
static struct timeval cpu_time; /* total time spent */
static int running; /* if the timer is running */
static struct timeval old_time;

/**
 * this function is for computing the time difference between timeval x and y
 * the result is stored in result
 */
int
timeval_subtract (struct timeval *result, struct timeval *x, struct timeval *y)
{
    /* Perform the carry for the later subtraction by updating y. */
    if (x->tv_usec < y->tv_usec) {
	int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
	y->tv_usec -= 1000000 * nsec;
	y->tv_sec += nsec;
    }
    if (x->tv_usec - y->tv_usec > 1000000) {
	int nsec = (x->tv_usec - y->tv_usec) / 1000000;
	y->tv_usec += 1000000 * nsec;
	y->tv_sec -= nsec;
    }

    /* Compute the time remaining to wait.
     * tv_usec is certainly positive.
     */
    result->tv_sec = x->tv_sec - y->tv_sec;
    result->tv_usec = x->tv_usec - y->tv_usec;

    /* Return 1 if result is negative. */
    return x->tv_sec < y->tv_sec;
}

void
__stp_timereset()
{
    cpu_time.tv_sec = 0;
    cpu_time.tv_usec = 0;
    running = 0;
}

void
__stp_timestart()
{
    if (running) {
	printf("timestart: Should stop timer then start it.\n");
	exit(1);
    } else {
	running = 1;
	if ((gettimeofday(&old_time, NULL)) == -1) {
	    perror("gettimeofday");
	    exit(1);
	}
    }
}

void
__stp_timestop()
{
    struct timeval curr, diff, total;
    if (running) {
	if((gettimeofday(&curr, NULL)) == -1) {
	    perror("gettimeofday");
	    exit(1);
	}
	running = 0;
	timeval_subtract(&diff, &curr, &old_time);
	cpu_time.tv_sec += diff.tv_sec;
	cpu_time.tv_usec += diff.tv_usec;
	if (cpu_time.tv_usec > 1000000) {
	    cpu_time.tv_usec -= 1000000;
	    cpu_time.tv_sec ++;
	}
    } else {
	printf("timestop: Should start timer then stop it.\n");
	exit(1);
    }
}

void
__stp_timereport()
{
    printf("**************************************************************\n");
    printf("TIME_REPORT: Totally time spent = (%d,%d) or %d usec\n",
	   cpu_time.tv_sec, cpu_time.tv_usec,
	   cpu_time.tv_usec + 1000000*cpu_time.tv_sec);
    printf("**************************************************************\n");
}

/*******************************************
 * Pring leaf pages
 *******************************************/
int printingleaf = 0;
db_pgno_t currPgno = (db_pgno_t)(-1);

int
printleafstart()
{
    printingleaf = 1;
    currPgno = -1;
}

int
printleafstop()
{
    printingleaf = 0;
    currPgno = -1;
}

int
printpagenum(db_pgno_t pgno)
{
    if (printingleaf == 1) {
        if (pgno != currPgno) {
            currPgno = pgno;
            printf("---------------------- pgno = %d ----------------------\n", currPgno);
        }
    }
    return 0;
}

/*******************************************
 * TT table size (number of entries)
 *******************************************/
/*
int
reportttsize(DB_ENV *dbenv)
{
    int i, cnt;
    DB_STP	*dbstp;
    STP_TT_ENTRY	*tt;
    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;
    tt = dbstp->tt_table;

    i = 0;
    cnt = 0;
    while (tt[i].next != IDX_INVALID) {
	i = tt[i].next;
	cnt ++;
    }
    printf("\nTT table size is %d\n", cnt);
    return (0);
}
*/

/*******************************************
 * number of TT entries scaned (STP_TT_FIND)
 *******************************************/
static num_ttscan=0;
int
count_ttscan()
{
    num_ttscan ++;
    return 0;
}

int
report_ttscan()
{
    printf("Num of TT entries scaned = %d\n", num_ttscan);
    return 0;
}
#endif
