/*-
 * Developed (2002-2003) by:
 *     TAU Project
 *     Computer Science Dept., University of Arizona
 *     Tucson, AZ 85721, USA
 *     URL  : http://www.cs.arizona.edu/tau
 *     Email: tau@cs.arizona.edu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "db_config.h"
#include "db_int.h"
#include "db_page.h"
#include "db_shash.h"
#include "hash.h"
#include "qam.h"
#include "lock.h"
#include "mp.h"
#include "txn.h"
#include "TTDB.h"
#include "ttdb_ext.h"
#include "clk_txn.h"
#include "stp_int.h"
#include "stp.h"
#include "temporal.h"

/*
#define TTDB_DEBUG
*/

/* we want to use the original structures and functions
 * in our code
 */
#undef DB
#undef DBC
#undef db_create

/* These are flags in the database metadata page.
 * The counterpart of the user specified flags are defined
 * in the public accessable db.h file.
 */
#define TEMPORAL_DB  0x02 /* database is temporal */
#define DUP_TT       0x04 /* the temporal database has duplicates */
#define EXISTS_TT_DB 0x08 /* old database. not newly created */

/* PARTITION Mani, 28 Mar 2003, Added flag to represent if a db is
 * partitioned or not
 */
#define PARTITION_DB 0x10 /* the temporal database has partitioned db */
/* Mani End */

DB_ENV *dbenv;

int is_until_changed(struct timeval time_value)
{
    return (time_value.tv_sec == UC);
}

int is_us(struct timeval time_value)
{
    return (time_value.tv_sec == US);
}

int is_other(struct timeval time_value)
{
    return (time_value.tv_sec == OTHER);
}

/*
 * __update_TT_info_create --
 * Updates the temporal information in the database, depending on the flags
 * passed to DB->open and the preexisting flags. Authors: Supratik, Haifeng,
 * Kalyani.
 */
int __update_TT_info_create(dbp, flags, meta)
    DB          *dbp;
    u_int32_t   flags;
    DBMETA      *meta;
{
    u_int8_t *ttdb_flag;
#ifdef TTDB_DEBUG
    FILE *fp;
    fp=fopen("common_debug_file", "a");
    fprintf(fp, "the DB is getting created for the first time.-- _update_tt_info(bt_open.c)\n");
    fflush(fp);
    fclose(fp);
#endif
    /* The database is being created for the first time. Therefore,
     * initialize according to the flags to DB->open.
     * NOTE HERE: WE DO NOT SET THE DB_AM_DUP FOR THE FLAGS IN THE
     * DATABASE BECAUSE THE CHECK FOR THAT MIGHT COME AFTER THIS FUNCTI0N
     * HAS BEEN CALLED.
     */
#ifdef BDB_3
    ttdb_flag = (u_int8_t *)&(meta->unused1[0]);
#endif
#ifdef BDB_4
    ttdb_flag = (u_int8_t *)&(meta->unused1);
#endif

    *ttdb_flag = EXISTS_TT_DB;

    dbp->flags &= (~DB_TEMPORAL); /* Shilong Yao, 04/26/03 */
    dbp->flags |= TT_DB_EXISTS;
    if(flags & DB_TEMPORAL){
	*ttdb_flag |= TEMPORAL_DB;
	dbp->flags |= DB_TEMPORAL;
	if(dbp->flags & DB_AM_DUP){
	    *ttdb_flag |= DUP_TT;
	    dbp->flags |= TT_DUP;
	}

		/* PARTITION Mani 28th Mar 2003, check if user wants to have
		 * the db partitioned if yes set the dbp flag and the
		 * metadata information
		 */
		if(flags & DB_PARTITION){
			*ttdb_flag |= PARTITION_DB;
			dbp->flags |= DB_PARTITION;
		}
		/* Mani End */

	/*dbp->flags |= DB_AM_DUP;*/
    }
    return(0);
}

/*
 * __update_TT_info_open --
 * Updates the temporal information in the database, depending on the flags
 * passed to DB->open and the preexisting flags. Authors: Supratik, Haifeng,
 * Kalyani.
 */
int __update_TT_info_open(dbp, flags, meta)
    DB          *dbp;
    u_int32_t   flags;
    DBMETA      *meta;
{
    u_int8_t *ttdb_flag;

#ifdef BDB_3
    ttdb_flag = (u_int8_t *)&(meta->unused1[0]);
#endif
#ifdef BDB_4
    ttdb_flag = (u_int8_t *)&(meta->unused1);
#endif

    /*
     * error cases + setting values of DB flags
     */
    if(*ttdb_flag & EXISTS_TT_DB){
#ifdef TTDB_DEBUG
	FILE *fp;
	fp=fopen("common_debug_file", "a");
	fprintf(fp, "the DB exists. now checking for errors -- update_tt_info(bt_open.c)\n");
	fflush(fp);
	fclose(fp);
#endif
        /* error if opening temporal DB not with temporal flag and vice versa
	   if((*ttdb_flag & TEMPORAL_DB) && (!(flags & DB_TEMPORAL))){
	   #ifdef TTDB_DEBUG
	   fp=fopen("common_debug_file", "a");
	   fprintf(fp, "error if opening temporal DB not with temporal flag and vice versa --1--update_tt_info(bt_open.c)\n");
	   fflush(fp);
	   fclose(fp);
	   #endif
	   return(TEMPORAL_TUC_ERR);
	   } else
	*/

        if((!(*ttdb_flag & TEMPORAL_DB)) && (flags & DB_TEMPORAL)){
#ifdef TTDB_DEBUG
	    fp=fopen("common_debug_file", "a");
	    fprintf(fp, "error if opening temporal DB not with temporal flag and vice versa --2-- update_tt_info(bt_open.c)\n");
	    fflush(fp);
	    fclose(fp);
#endif
	    /* if ( !(flags & DB_CREATE) ) */
	    return(TEMPORAL_TUC_ERR);
        } else if(*ttdb_flag & TEMPORAL_DB){
#ifdef TTDB_DEBUG
	    fp=fopen("common_debug_file", "a");
	    fprintf(fp, "--3--update_tt_info(bt_open.c)\n");
	    fflush(fp);
	    fclose(fp);
#endif
            if (((TEMPORAL_HANDLE *)(dbp->dbenv->temporal_handle))->stp_handle == NULL)
		return (EINVAL);
            /* we have to put the DB_AM_DUP flag to the DB, because we take it
             * off just before writing to the file (in DB->close). But that's
             * done in the __ttdb_open
             */
            if(*ttdb_flag & DUP_TT){
#ifdef TTDB_DEBUG
		fp=fopen("common_debug_file", "a");
		fprintf(fp, "--4--update_tt_info(bt_open.c)\n");
		fflush(fp);
		fclose(fp);
#endif
                dbp->flags |= TT_DUP;
            }

		/* PARTITION Mani 28th Mar 2003, Checking if the existing db has
		 * partitioned db, if yes then the user has no say, it gets opened
		 * in partitioned mode just like, if the db is temporal, even if the
		 * user doesnt pass DB_TEMPORAL flag the db gets opened in temporal
		 * mode TODO : If the db_create based check remains, then check for
		 * partition needs to be done there also
		 */

		if(meta->unused1[0] & PARTITION_DB){

			#ifdef TTDB_DEBUG
			fp=fopen("common_debug_file", "a");
			fprintf(fp, "db partition flag being set for the db at update_tt_info(TTDB.c)\n");
			fflush(fp);
			fclose(fp);
			#endif
			dbp->flags |= DB_PARTITION;
		}

		/* Mani End */

            dbp->flags |= DB_TEMPORAL; /*Shilong*/
            /*dbp->flags |= DB_AM_DUP;*/
        }

        if ((flags & DB_TEMPORAL) && (flags & DB_CREATE)){
#ifdef TTDB_DEBUG
	    fp=fopen("common_debug_file", "a");
	    fprintf(fp, "--5--update_tt_info(bt_open.c)\n");
	    fclose(fp);
#endif
            dbp->flags |= DB_TEMPORAL;
        }
#ifdef TTDB_DEBUG
	fp=fopen("common_debug_file", "a");
	fprintf(fp, "--6--update_tt_info(bt_open.c)\n");
	fclose(fp);
#endif
        dbp->flags |= TT_DB_EXISTS;
    } else {
	printf("TTDB: Internal error at update_TT_info_open: opened file without TT_DB_EXISTS\n");
	exit(1);
    }
    return(0);
}

/*
 * is_tt  --
 * takes in the database and returns if the database(table) has temporal
 * support
 */
int is_tt(struct __ttdb *ttdb){
    if((ttdb->db)->flags & DB_TEMPORAL){
#ifdef TTDB_DEBUG
        FILE *fp;
        fp=fopen("common_debug_file", "a");
        fprintf(fp, "i returned: TT support -- is_tt(TTDB.c)\n");
        fflush(fp);
        fclose(fp);
#endif
        return(1);
    } else{
#ifdef TTDB_DEBUG
        FILE *fp;
        fp=fopen("common_debug_file", "a");
        fprintf(fp, "i returned: no TT support -- is_tt(TTDB.c)\n");
        fflush(fp);
        fclose(fp);
#endif
        return(0);
    }
}

/*
 * __ttdb_has_dup --
 * tells whether the temporal database has duplicates or not
 */
int __ttdb_has_dup(struct __ttdb *ttdb){
    if((ttdb->db)->flags & TT_DUP){
        return(1);
    } else {
        return(0);
    }
}

/*
 * is_partitioned -
 * PARTITION Mani 28th Mar 2003
 * check if the db has partition db support, which means it has a embedded
 * archive db in it which it uses to move the old records and keeps on the
 * current records in the current db
 */

int is_partitioned (struct __ttdb *ttdb) {

    if((ttdb->db)->flags & DB_PARTITION){
        return(1);
    }
    else{
        return(0);
    }
}

/* Mani End */

/*
 * __ttdb_open --
 * this is analogous to the regular DB->open. This, however, checks for
 * another flag called DB_TEMPORAL, which defines the opened database to have
 * temporal support or not.
 */
#ifdef BDB_4
int __ttdb_open(struct __ttdb *ttdb, DB_TXN *tid, const char *file, const char *database,
		DBTYPE type, u_int32_t flags, int mode){
#endif
#ifdef BDB_3
int __ttdb_open(struct __ttdb *ttdb, const char *file, const char *database,
		DBTYPE type, u_int32_t flags, int mode){
#endif
    struct __db     *db;
    int             ret;
    u_int32_t       temp_flags;
	/* PARTITION Mani, 16th Mar 2003, variable for storing the archive db's
	 * filename
	 */
	char * archiveFile;
	/* Mani End */


#ifdef TTDB_DEBUG
    FILE *fp;
    fp=fopen("common_debug_file", "a");
    fprintf(fp, "TT Support ? %s -- __ttdb_open(TTDB.c)\n",
	    (flags&DB_TEMPORAL)?"YES":"NO");
    fflush(fp);
    fclose(fp);
#endif
    /* we open the database with the regular flags and then check for the
     * flags to determins if the database is temporal or not, has duplicates
     * or not..etc.
     */
#ifdef BDB_4
    if((ret=(ttdb->db)->open((ttdb->db), tid, file, database, type, flags, mode)) !=0) {
        return(ret);
    }
#endif
#ifdef BDB_3
    if((ret=(ttdb->db)->open((ttdb->db), file, database, type, flags, mode)) !=0) {
        return(ret);
    }
#endif


	/* PARTITION Mani, 16th Mar 2003, opening the archival store, when the
	 * current store has been opened. to generate the filename for the
	 * archival db, the flags must be same, but the archival store is not
	 * marked temporal, so remove the temporal flag from the flags, since
	 * otherwise we will be recursively creating archival stores
	 * The best place to decide if archive needs to be opened or not, is
	 * to look for the current db's flag information
	 * 16th Jul 2003, Modified to make the archival store have the temporal
	 * flag, although it wouldnt behave temporally, since its maintained
	 * internally. The flag is required for the Stamper and Clock Module to
	 * consider its pages for stamping and clocking.
	 */

	if (is_tt(ttdb) && is_partitioned(ttdb)) {
		archiveFile = (char *) malloc ((strlen(file)+3)*sizeof(char));
		strcpy(archiveFile, file);
		strcat(archiveFile, ".a");

		#ifdef TTDB_DEBUG
		fp=fopen("common_debug_file", "a");
		fprintf(fp, "\n The archive db file name is : %s \n", archiveFile);
		fflush(fp);
		fclose(fp);
		#endif

		/* temp_flags = flags ^ DB_TEMPORAL; */ /* XOR operation, we know the flag is already set, so it will get reset */
		/* temp_flags = temp_flags ^ DB_PARTITION; */

		#ifdef BDB_3
		if((ret=(ttdb->adb)->open((ttdb->adb), archiveFile, database, type, flags, mode)) !=0){
			return(ret);
		}
		#endif

		#ifdef BDB_4
		if((ret=(ttdb->adb)->open((ttdb->adb), tid, archiveFile, database, type, flags, mode)) !=0){
			return(ret);
		}
		#endif

		(ttdb->adb)->flags |= DB_AM_DUP; /* Since the archive has to be able to hold duplicates */

		#ifdef TTDB_DEBUG
		fp=fopen("common_debug_file", "a");
		fprintf(fp, "\n Archive opened successfully. \n");
		fflush(fp);
		fclose(fp);
		#endif

		/* TODO : free(archiveFile);*/
	}
	/* Mani End */

	/* PARTITION Mani, 27 Mar 2003, In a partitioned db the current store
	 * need not be compelled to have dups unless the data itself has dups,
	 * but if the db is temporal and not partitioned then we need to compel
	 * the current store to allow dups, so the below code is modified accordingly
	 */

	/* all temporal databases, non partitioned current stores have
	 * duplicate support
	 */

	if(is_tt(ttdb) && !(is_partitioned(ttdb))){
        	(ttdb->db)->flags |= DB_AM_DUP;
   	}

	/* Mani, T-BDB Code modified and placed above.
    if((ttdb->db)->flags & DB_TEMPORAL){
        (ttdb->db)->flags |= DB_AM_DUP;
    }
    */

	/* Mani End */

    /* there are certain elements that are accessed from outside. So, they
     * have to be mirrored in the struct __ttdb */
    ttdb->app_private= (ttdb->db)->app_private;

    /* call to the stamper guys to get ready for some action.*/
    /* 5/28/02
      if(is_tt(ttdb)){
      __stp_settemporal(1);
      }
      else{
      __stp_settemporal(0);
      }
    */

    {
        struct timeval current;
	gettimeofday(&current, NULL);
	/*
	  current.tv_sec= time(NULL);
	  current.tv_usec=0;
	*/
        ttdb->set_current(ttdb, current);
    }
    return(ret);
}

/*
 * __getTTvoid -
 * Internal Function:
 * this function returns a malloc'ed void* after combining the two inputs.
 *        _________           ___________           _______________________
 * TTDATA |\\\\\\\|   + void* |   data  |  = void*  |\\\\\\\\|    data    |
 *        ^^^^^^^^^           ^^^^^^^^^^^           ^^^^^^^^^^^^^^^^^^^^^^^ */
/* IMP: we are assuming that the 'char' is of size 1. so when we do a char* =
 * char* + size, then it takes us to the correct place. */
void *__getTTvoid(void *data, u_int32_t size, TTDATA *tt){
    char  *ret;

    ret=(char*)malloc(sizeof(TTDATA) + size);
    memcpy((void*)ret, (void*)tt, sizeof(TTDATA));
    ret=ret + sizeof(TTDATA);
    memcpy((void*)ret, data, size);
    ret=ret - sizeof(TTDATA);
    return((void*)ret);
}

/*
 *   __ttdb_del --
 *   Delete records from current database
 */
int __ttdb_del(struct __ttdb *ttdb, DB_TXN *txnid, DBT *key, u_int32_t flags)
{
    struct  __dbc *Cur;
    TTDATA  *tt;
    int     ret=0, temporal, key_size, foundflag=0;
    void    *key_data;
    DBT     data;
    DBT     key_temp;

    temporal=is_tt(ttdb);
    if(!temporal){
        {
            struct timeval current;
	    gettimeofday(&current, NULL);
	    /*
	      current.tv_sec= time(NULL);
	      current.tv_usec=0;
	    */
            ttdb->set_current(ttdb, current);
        }
        return((ttdb->db)->del((ttdb->db), txnid, key, flags));
    }

    /* temporal database but outside transactions. disallowed. */
    if(txnid==NULL){
        return(TEMPORAL_TUC_ERR);
    }

    if(flags==0) {
	if(ret=(ttdb->db)->cursor((ttdb->db), txnid, &Cur, 0)) {
	    {
		struct timeval current;
		gettimeofday(&current, NULL);
		/*
		  current.tv_sec= time(NULL);
		  current.tv_usec=0;
		*/
		ttdb->set_current(ttdb, current);
	    }
	    return(ret);
	}
	memset(&key_temp, 0, sizeof(key_temp));
	memset(&data, 0, sizeof(data));

	/* Shilong Yao, 03/23/03
	   while((ret=Cur->c_get(Cur, &key_temp, &data, DB_NEXT))==0){
	   if((memcmp(key->data, key_temp.data, key->size)==0) &&
	   (key_temp.size==key->size)){
	*/
	if ((ret=Cur->c_get(Cur, key, &data, DB_SET)) != 0) {
	    goto done;
	}

	/*printf("key=%d, data=%s\n", *(int *)(key->data), (char*)(data.data));*/
	while(1) {
	    tt = (TTDATA*)data.data;

	    if (tt->state == TIME_UC) {
		foundflag=1;

		tt->stop.stamperId.txnid=txnid->txnid;
		tt->stop.stamperId.envid =
		    ((TEMPORAL_HANDLE *)(txnid->mgrp->dbenv->temporal_handle))->envid;
		tt->state = TIME_ID;

		/*stamper is here!
		  tt->stop.curr.tv_sec=time(NULL);
		  tt->stop.curr.tv_usec=0;

		  tt->state = TIME_TIME;
		*/

		/* PARTITION Mani, 13 April 2003, Added code to check for
		 * isPartitioned and if so, then push the
		 * record to the archive and deleted it from the current store,
		 * else "placed the code
		 * that existed for temporal bdb, i.e. logically delete that record,
		 * by updating end time
		 */

			if (is_partitioned(ttdb)) {
				/* Put the key,data information in the archive store
				 * and delete the current record */

				if((ret=(ttdb->adb)->put((ttdb->adb), txnid, key, &data, 0))!=0) {
					Cur->c_close(Cur);
					return(ret);
				}

				if((ret = Cur->c_del (Cur, 0)) != 0) {
					Cur->c_close(Cur);
					return(ret);
				}
			}
			else {

				/* execute the non partitioned related code */
				if((ret = Cur->c_put (Cur, key, &data,
							  DB_CURRENT))!=0) {
					Cur->c_close(Cur); /* Shilong Yao, 04/01/03 */
					return(ret);
				}
			}
			/* Mani End */
	    } else if (tt->state == ID_UC) {
		foundflag=1;
		/* in the same xact, do real delete */
		/* Mani, 2nd April 2003, Modified the If condition to also check envid */
		if(tt->start.stamperId.txnid==txnid->txnid &&
		   tt->start.stamperId.envid==
		   ((TEMPORAL_HANDLE *)(txnid->mgrp->dbenv->temporal_handle))->envid) {
		    if ((ret = Cur->c_del (Cur, 0)) != 0) {
			Cur->c_close(Cur); /* Shilong Yao, 04/01/03 */
			return (ret);
		    }
		} else {
		    /* Mani, 2nd April 2003, Modified the else part to just update the
			 * record into the ID_ID state, if the start ID is not us, which is
			 * what is being done in the TIME_UC case
			 */
		    tt->stop.stamperId.txnid=txnid->txnid;
		    tt->stop.stamperId.envid=
			((TEMPORAL_HANDLE *)(txnid->mgrp->dbenv->temporal_handle))->envid;
		    tt->state = ID_ID;

			/* PARTITION Mani, 13 April 2003, Added code to check for
			 * isPartitioned and if so, then push therecord to the archive
			 * and deleted it from the current store, else "placed the code
			 * that existed for temporal bdb", i.e. logically delete that
			 * record, by updating end time
			 */

			if (is_partitioned(ttdb)) {
				/* Put the key,data information in the archive store and
				 * delete the current record
				 */

				if((ret=(ttdb->adb)->put((ttdb->adb), txnid, key, &data, 0))!=0) {
					Cur->c_close(Cur);
					return(ret);
				}

				if((ret = Cur->c_del (Cur, 0)) != 0) {
					Cur->c_close(Cur);
					return(ret);
				}
			}
			else {
				/* execute the non partitioned related code */
				if((ret = Cur->c_put (Cur, key, &data, DB_CURRENT))!=0) {
					Cur->c_close(Cur);
					return(ret);
				}
			}
			/* Mani End */
		}
	    } else {
		/* in all other cases, we do not make any changes to the
		 * DBT
		 */
	    }

	    /* Get the next tuple to see if it is also the same key
	     * If no, we stop searching.
	     */
	    if ((ret=Cur->c_get(Cur, &key_temp, &data, DB_NEXT_DUP))!=0) {
		break;
	    }

	    if((memcmp(key->data, key_temp.data, key->size)!=0) ||
	       (key_temp.size!=key->size))
		break;
	}

      done:
	/*error return*/
	Cur->c_close(Cur);
	{
	    struct timeval current;
	    gettimeofday(&current, NULL);
	    /*
	      current.tv_sec= time(NULL);
	      current.tv_usec=0;
	    */
	    ttdb->set_current(ttdb, current);
	}
	/* Mani , 2nd April 2003, Commented the use of foundflag, since the success
	 * of del is not decided by whether it was able to delete any record or not,
	 * it deletes records if it finds them, otherwise just returns successfully.
	 * Moreover variable ret will have the value fed by our cursor operation
	 * and may not be a valid return value for del operation, so just return 0
	 *
	 if (foundflag==0)
	 return(ret);
	 else
	 return(0);
	*/

		/* Mani , 13th April 2003, Corrected the return value to DB_NOTFOUND,
		 * when foundflag is 0
		 */
		if (foundflag==0)
			return(DB_NOTFOUND);
		else
			return(0);

    } else {
        /*we haven't taken care of any flags... flags currently unused in
         *db_del. must be set to 0.
         */
        /*error return*/
        {
            struct timeval current;
	    gettimeofday(&current, NULL);
	    /*
	      current.tv_sec= time(NULL);
	      current.tv_usec=0;
	    */
            ttdb->set_current(ttdb, current);
        }
        return(TEMPORAL_TUC_ERR);
    }
}

/*
 * __ttdb_put --
 * similar to DB->put, only takes in a struct __ttdb instead of the __db
 * (normal DB).
 * This function actually needs to put the start and the stop time with the
 * data, so that temporal queries can be performed. We have to make sure that
 * the DBT->size is correct (because, the size changes when we add the start
 * and stop times. */
int __ttdb_put(struct __ttdb *ttdb, DB_TXN *txnid, DBT *key, DBT *data,
	       u_int32_t flags)
{
    int                 ret;
    void                *tt_data;
    TTDATA              tt;
    DBT                 new_dat;

    /* PARTITION Mani, 22 Mar 2003, curr_dat field to be used to hold the
     * record in current store, which is going to be updated
     */
    DBT			curr_dat;
    TTDATA *		ttDataPtr;
    /* Mani End */

#ifdef TTDB_DEBUG
    FILE *fp;
#endif

    if(is_tt(ttdb)){
#ifdef TTDB_DEBUG
	fp=fopen("common_debug_file", "a");
	fprintf(fp, "TT support -- db_put(TTDB.c)\n");
#endif

	if(txnid==NULL){
#ifdef TTDB_DEBUG
	    fprintf(fp, "txnID is null!!!!\n");
#endif
	} else {
#ifdef TTDB_DEBUG
	    fprintf(fp, "txnID is NOT null. Trying to print it out:\n");
	    fflush(fp);
	    fprintf(fp, "the transaction ID is %d \n", txnid->txnid);
	    fprintf(fp, "the logical key & data we are gonna put is %s, %s \n",
		    (char*)(key->data), (char*)(data->data));
	    fflush(fp);
	    fclose(fp);
#endif
	}

        if(txnid==NULL){
            return(TEMPORAL_TUC_ERR);
        }

        tt.state=ID_UC;
        tt.start.stamperId.txnid=txnid->txnid;
	tt.start.stamperId.envid=
	    ((TEMPORAL_HANDLE *)(txnid->mgrp->dbenv->temporal_handle))->envid;
        /* tt.stop.stamperId.txnid=UNTILCHANGED;
	   Commented, since this is redundant, as state info already tells us abt this.
	*/

        /* stamper is here!!
	   tt.state = TIME_UC;
	   tt.start.curr.tv_sec=111;
	   tt.start.curr.tv_usec=222;

	   tt.stop.stamperId=UNTILCHANGED;
	*/

        memset(&new_dat, 0, sizeof(new_dat));
        new_dat.data=malloc(data->size+sizeof(TTDATA));
        memcpy(new_dat.data, __getTTvoid(data->data, data->size, &tt),
	       data->size + sizeof(TTDATA));
        new_dat.size=data->size + sizeof(TTDATA);   /* needs extra size*/
        new_dat.ulen=data->ulen;
        new_dat.dlen=data->dlen;
        new_dat.doff=data->doff;
        new_dat.flags=data->flags;

        /* special things to do when unique keys is set */
        if(!__ttdb_has_dup(ttdb)){

#ifdef TTDB_DEBUG
	    fp=fopen("common_debug_file", "a");
	    fprintf(fp, "unique keys is set!!! --__ttdb_put(TTDB.c)\n");
	    fprintf(fp, "actually we have put key, data:%s, %s --__ttdb_put(TTDB.c)\n",
		    (char*)(key->data), (char*)(new_dat.data));
#endif
	    memcpy(&tt, new_dat.data, sizeof(TTDATA));

#ifdef TTDB_DEBUG
	    fprintf(fp, "the tt.state which we have put is: %d!!!--__ttdb_put(TTDB.c)\n",
		    tt.state);
	    fflush(fp);
	    fclose(fp);
#endif

		/* PARTITION Mani, 22 Mar 2003,
		 * do a get on the current store to get the current version if
		 * it exists
		 * do a put on the current store to update the record if it existed
		 * or add it, which is default behaviour of put
		 * put the record, if one was obtained in the get, after setting the
		 * stopTime for it, into the archive store
		 * if the final record to be put into the archive is of type ID_ID
		 * and if both ID's are the same txn then
		 * the record can be eliminated.
		 */

		if (is_partitioned(ttdb)) {

			memset(&curr_dat, 0, sizeof(curr_dat));
			/* Mani, 27 April 2003,
			 * Corrected the code here to specify the allocation information to bdb layer
			 * removed the allocation on this end. TODO : where should this mem be freed
			 */
			curr_dat.flags = DB_DBT_MALLOC;
			ret = (ttdb->db)->get(ttdb->db, txnid, key, &curr_dat, 0);
			if (ret == 0) {
				/* check the start time and decide if record needs to go to
				 * the archive or not
				 */
				ttDataPtr = (TTDATA *) curr_dat.data;
				if (!(ttDataPtr->state == ID_UC &&
					ttDataPtr->start.stamperId.txnid == txnid->txnid &&
					ttDataPtr->start.stamperId.envid ==
					((TEMPORAL_HANDLE *)(txnid->mgrp->dbenv->temporal_handle))->envid)) {

					/* So the current record need to be pushed to the archive
				 	* Ideally, only records of type ID_UC or TIME_UC can exist in the current store */
					if (ttDataPtr->state == ID_UC) ttDataPtr->state = ID_ID;
					if (ttDataPtr->state == TIME_UC) ttDataPtr->state = TIME_ID;

        			ttDataPtr->stop.stamperId.txnid = txnid->txnid;
					ttDataPtr->stop.stamperId.envid =
						((TEMPORAL_HANDLE *)(txnid->mgrp->dbenv->temporal_handle))->envid;

					ret=(ttdb->adb)->put((ttdb->adb), txnid, key, &curr_dat, flags);
				}
			}
		}
		else {
			/* Mani 28th Mar 2003, Moved the code from the default path to
			 * this path, so that its executed when the db is non partitioned
			 */

			/* delete everything with the same key (unique keys set) */
 	           	ret=__ttdb_del(ttdb, txnid, key, 0);
		}

       	/* Mani 28th Mar 2003, Added Comment
		 * The put will put a new record in the current db or will update
		 * the record based on key depending on the status if the dup allowed
		 * flag in the db, which is set accordingly while opening the db
		 * based on if the db is partitioned or not
		 */

		/* Mani End */

	    ret=(ttdb->db)->put((ttdb->db), txnid, key, &new_dat, flags);
	    {
		struct timeval current;
		gettimeofday(&current, NULL);
		/*
		  current.tv_sec= time(NULL);
		  current.tv_usec=0;
		*/
		ttdb->set_current(ttdb, current);
	    }
	    return(ret);
        }

        /* non unique keys: just store another record */
        ret=(ttdb->db)->put((ttdb->db), txnid, key, &new_dat, flags);
#ifdef TTDB_DEBUG
	fp=fopen("common_debug_file", "a");
	fprintf(fp, "we have put key, data:%s, %s --__ttdb_put(TTDB.c)\n",
		(char*)(key->data), (char*)(new_dat.data));
	fflush(fp);
	fclose(fp);
#endif
        {
            /* set current time in the database to this present time */
            struct timeval current;
	    gettimeofday(&current, NULL);
	    /*
	      current.tv_sec= time(NULL);
	      current.tv_usec=0;
	    */
            ttdb->set_current(ttdb, current);
        }
        return(ret);
    } else { /* non temporal */
#ifdef TTDB_DEBUG
	FILE *fp;
	fp=fopen("common_debug_file", "a");
	fprintf(fp, "no TT support -- db_put(TTDB.c)\n");
	fflush(fp);
	fclose(fp);
#endif
	{
	    struct timeval current;
	    gettimeofday(&current, NULL);
	    /*
	      current.tv_sec= time(NULL);
	      current.tv_usec=0;
	    */
	    ttdb->set_current(ttdb, current);
	}
	ret=(ttdb->db)->put((ttdb->db), txnid, key, data, flags);
	return(ret);
    }
}

/*
 * __validRec --
 * Internal Function:
 * checks whether the data is valid in the particular database
 */
int __validRec(void *data, struct __ttdb *ttdb, DB_TXN *txnid){
    TTDATA  *tt;
    STP_ID stamperId;
    struct timeval recStart, recStop;

    tt=(TTDATA*)malloc(sizeof(TTDATA));
    memcpy((void*)tt, data, sizeof(TTDATA));
    stamperId.txnid = txnid->txnid;
    stamperId.envid =
    	((TEMPORAL_HANDLE *)(txnid->mgrp->dbenv->temporal_handle))->envid;

    switch(tt->state){
    case ID_TIME: /* This case cannot occur, since the stamper would have stamped
		     the record fully and hence the value will be TIME_TIME */
	return(0);
    case ID_ID: /* Find the equivalent times for the ID and see if the current time
		   falls within the time, as done in TIME_TIME code.
		   But, if we are able to see the record that means the event occured
		   before our current time and hence current time will not fall
		   in this interval */
	/* If start and stop are same txn, record doesnt exist */
	if (tt->start.stamperId.txnid == tt->stop.stamperId.txnid &&
	    tt->start.stamperId.envid == tt->stop.stamperId.envid)
	    return(0);

	/* If stop is US then record has been killed by us, so it need not be visible to us
	 * If start is US then it cannot be stopped by someone else, till we commit, so that
	 * case is anyways invalid */
	if ( (tt->start.stamperId.txnid == stamperId.txnid &&
	      tt->start.stamperId.envid == stamperId.envid) ||
	     (tt->stop.stamperId.txnid == stamperId.txnid &&
	      tt->stop.stamperId.envid == stamperId.envid)
	    )
	    return(0);
	/* At this point of the code start and stop are not the same txn and are
	 * not US
	 */
	recStart = __stp_get_ctime(txnid->mgrp->dbenv, tt->start.stamperId);
	recStop = __stp_get_ctime(txnid->mgrp->dbenv, tt->stop.stamperId);
	if ((recStart.tv_sec != 0) && (recStop.tv_sec != 0)) {
	    if (((recStart.tv_sec < ttdb->current_time.tv_sec) ||
		 (recStart.tv_sec == ttdb->current_time.tv_sec
		  &&
		  recStart.tv_usec <= ttdb->current_time.tv_usec))
		&&
		((recStop.tv_sec > ttdb->current_time.tv_sec) ||
		 (recStop.tv_sec == ttdb->current_time.tv_sec
		  &&
		  recStop.tv_usec >= ttdb->current_time.tv_usec)))

		return (1);
	}
	return (0);

    case ID_UC:
	/* The current transaction put this data, or some other txn put this data
	 * the ID can be converted to TIME and we can confirm that current time is
	 * greater than TIME. I think this will be always true since
	 * if we are seeing that record it would have been created by a txn which is
	 * earlier than us.Just to make sure, we can behave like TIME_UC after
	 * we get the TIME for the ID.
	 * The time returned by the stamper should ideally be non zero, since if i
	 * am seeing an ID then that ID must have committed and must have a time,
	 * as we are not doing a dirty read.
	 */
	/* If ID is US then i should see the record */
	if  (tt->start.stamperId.txnid == stamperId.txnid &&
	     tt->start.stamperId.envid == stamperId.envid)
	    return(1);

	recStart = __stp_get_ctime(txnid->mgrp->dbenv, tt->start.stamperId);

	if (recStart.tv_sec != 0 && ((recStart.tv_sec < ttdb->current_time.tv_sec) ||
				     (recStart.tv_sec == ttdb->current_time.tv_sec &&
				      recStart.tv_usec <= ttdb->current_time.tv_usec)))
	    return (1);
	else
	    return (0);

    case TIME_ID:
	/* The current transation deleted this. Or some other txn deleted this.
	 * In this case we have check if the time of the other txn is greater than
	 * current time, then this record can be considered valid. Here again the
	 * argument is the txn that deleted this would have been earlier to us
	 * so our current time will be surely greater than the time of ID
	 * Just to make sure we can behave like TIME_TIME after we get the
	 * TIME for the ID
	 */
	return(0);
    case TIME_UC:
	if((tt->start.curr.tv_sec < ttdb->current_time.tv_sec)||
	   ((tt->start.curr.tv_sec == ttdb->current_time.tv_sec)
	    &&
	    (tt->start.curr.tv_usec <=ttdb->current_time.tv_usec))
	    )
	    return(1);
	else
	    return(0);
	break;
    case TIME_TIME:
	if(((tt->start.curr.tv_sec < ttdb->current_time.tv_sec)||
	    ((tt->start.curr.tv_sec == ttdb->current_time.tv_sec)
	     &&
	     (tt->start.curr.tv_usec <=ttdb->current_time.tv_usec))
	    )&&((tt->stop.curr.tv_sec > ttdb->current_time.tv_sec)||
		((tt->stop.curr.tv_sec == ttdb->current_time.tv_sec)
		 &&
		 (tt->stop.curr.tv_usec >=ttdb->current_time.tv_usec)))
	    )
	    return(1);
	else
	    return(0);
	break;
    default:
	return(0);  /* invalid record...*/
    }
}

/*
 * convertNonTem --
 * Internal Function:
 * converts temporally-enhanced DBT --to--> regular DBT
 */
DBT *convertNonTem(DBT *data){
    char    *temp;

    data->size= (data->size)- sizeof(TTDATA);
    temp= (char*)(data->data);
    temp= temp+ sizeof(TTDATA);
    data->data= (void*)temp;
    return(data);
}

/*
 * __ttdb_get --
 * similar to DB->get. only that the get function would actually check if the
 * 'current_time' parameter in the DB allows only particular records to be
 * accessed. We have to take care of the flags that are sent. Due to that, the
 * code for tt_get becomes substantially complicated. Design decision: we will
 * use the underlying database utilities to code in our functions (for
 * example, we would not use the temporal DBCursor functions from inside the
 * DB->... functions*/
int __ttdb_get(struct __ttdb *ttdb, DB_TXN *txnid, DBT *key, DBT *data,
	       u_int32_t flags)
{

    struct __dbc        *Cur;
    TTDATA  *tt;
    int     ret, temporal, key_size;
    void    *val_data, *key_data;
    DBT     key_temp;
    struct timeval curr_tmp;

    gettimeofday(&curr_tmp, NULL);
    temporal=is_tt(ttdb);
    if(!temporal){
        ret=(ttdb->db)->get((ttdb->db), txnid, key, data, flags);
        {
            struct timeval current;
	    gettimeofday(&current, NULL);
	    /*
	      current.tv_sec= time(NULL);
	      current.tv_usec=0;
	    */
            ttdb->set_current(ttdb, current);
        }
        return(ret);
    }

    /* the txnid should not be null for temporal DB */
    if(temporal && (txnid==NULL)){
        {
            struct timeval current;
	    gettimeofday(&current, NULL);
	    /*
	      current.tv_sec= time(NULL);
	      current.tv_usec=0;
	    */
            ttdb->set_current(ttdb, current);
        }
        return(TEMPORAL_TUC_ERR);
    }

	/* PARTITION Mani, 10th April 2003, Adding code to call the underlying
	 * get and convert the data returned to non-temporal form
	 * and return it to the user. This might handle some flags, but all flag
	 * related issues need to be handled in a separate iteration. The code
	 * will be executed only if the db is partitioned, if the db is not
	 * partitioned then the default code of t-bdb will be executed.
	 */

    if (is_partitioned(ttdb)) {

		if (flags == 0) {

			ret=(ttdb->db)->get((ttdb->db), txnid, key, data, flags);
			convertNonTem(data);
			{
				struct timeval current;
				gettimeofday(&current, NULL);
				ttdb->set_current(ttdb, current);
			}
			return(ret);
		}
		else {
			/* we haven't taken care of any flags <- TO DO */
			/* error return*/
			{
				struct timeval current;
				gettimeofday(&current, NULL);
				ttdb->set_current(ttdb, current);
			}
			return(TEMPORAL_TUC_ERR);
		}
	}

	/* Mani, 10th April 2003, End of Added Code. Since the above code has a
	 * return in the if and else part
	 * if the db is partitioned, the code will not flow to the code below.
	 */

	/* Mani End */

    /* the queries can not search in the future */
    if (ttdb->current_time.tv_sec > curr_tmp.tv_sec)
	return (TEMPORAL_TUC_ERR);
    else if ((ttdb->current_time.tv_sec == curr_tmp.tv_sec) &&
	     (ttdb->current_time.tv_usec > curr_tmp.tv_usec))
	return (TEMPORAL_TUC_ERR);
    /*
      if(ttdb->current_time.tv_sec > time(NULL)){
      return(TEMPORAL_TUC_ERR);
      }
    */

    if(flags==0){               /*no special flags*/
        if((ret=(ttdb->db)->cursor((ttdb->db), txnid, &Cur, 0))!=0){
            {
                struct timeval current;
		gettimeofday(&current, NULL);
		/*
		  current.tv_sec= time(NULL);
		  current.tv_usec=0;
		*/
                ttdb->set_current(ttdb, current);
            }
            return(ret);
        }
        memset(&key_temp, 0, sizeof(key_temp));
	/* Shilong Yao, 05/06/03
	   while((ret=Cur->c_get(Cur, &key_temp, data, DB_NEXT))==0){
	*/
	if ((ret=Cur->c_get(Cur, key, data, DB_SET)) != 0) {
            goto done;
        }
	while (1) {
	    /*
	      printf("DEBUG: HERE, I did get the key, %s\n",key_temp.data);
	      printf("DEBUG: HERE, size %d, key size %d\n", key_temp.size, key->size);
	      printf("DEBUG: HERE, the original key is %s\n",key->data);
	      printf("DEBUG: memcp=%d\n",memcmp(key->data, key_temp.data, key->size));
	    */
	    /* Shilong Yao, 05/06/03
	       if((memcmp(key->data, key_temp.data, key->size)==0) &&
	       (key_temp.size==key->size))
	    */
            {
		if (__validRec(data->data, ttdb, txnid)) {
		    convertNonTem(data);
		    /*success return*/
		    Cur->c_close(Cur);
		    {
			struct timeval current;
			gettimeofday(&current, NULL);
			/*
			  current.tv_sec= time(NULL);
			  current.tv_usec=0;
			*/
			ttdb->set_current(ttdb, current);
		    }
		    return(ret);
		}
		/*printf("DEBUG: did not pass the validation\n");*/
            }

	    /* Shilong Yao, 05/06/03 */
	    /* Get the next tuple to see if it is also the same key
	     * If no, we stop searching.
	     */
            if ((ret=Cur->c_get(Cur, &key_temp, data, DB_NEXT_DUP))!=0) {
                break;
            }
        }
      done:
	printf("DEBUG: There is error here, ret is %d\n", ret);
        /* error return*/
        Cur->c_close(Cur);
        {
            struct timeval current;
	    gettimeofday(&current, NULL);
	    /*
	      current.tv_sec= time(NULL);
	      current.tv_usec=0;
	    */
            ttdb->set_current(ttdb, current);
        }
        return(ret);
    } else {
        /* we haven't taken care of any flags <- TO DO */
        /* error return*/
        {
            struct timeval current;
	    gettimeofday(&current, NULL);
	    /*
	      current.tv_sec= time(NULL);
	      current.tv_usec=0;
	    */
            ttdb->set_current(ttdb, current);
        }
        return(TEMPORAL_TUC_ERR);
    }
}

/*
 *  __ttdb_c_del --
 *  Deletes the current record pointed to by the cursor. If the record has
 *  already been deleted, returns DB_KEYEMPTY or if the cursor hasn't been
 *  initialized, then returns EINVAL. 1 or 2 CHANGES TO BE MADE HERE AFTER
 *  INSTALLING YAO'S + MINGDE'S MODULES.
*/
int __ttdb_c_del(struct __ttdbc *ttdbc, u_int32_t flags)
{
    DBT     key, data;
    TTDATA  tt;
    int     ret;
    STP_ID  stamperId;

#ifdef TTDB_DEBUG
    FILE *fp;
    fp=fopen("common_debug_file", "a");
    fprintf(fp, "enter this function... --ttdb_c_del(TTDB.c)\n");
    fflush(fp);
    fclose(fp);
#endif

    if(!is_tt(ttdbc->ttdb)){
#ifdef TTDB_DEBUG
	fp=fopen("common_debug_file", "a");
	fprintf(fp, "no transactiontime support... --ttdb_c_del(TTDB.c)\n");
	fflush(fp);
	fclose(fp);
#endif

	ret=(ttdbc->dbc)->c_del((ttdbc->dbc), flags);
	{
	    struct timeval current;
	    gettimeofday(&current, NULL);
	    /*
	      current.tv_sec= time(NULL);
	      current.tv_usec=0;
	    */
	    ttdbc->ttdb->set_current(ttdbc->ttdb, current);
	}
	return(ret);
    } else if(ttdbc->dbc->txn==NULL){
#ifdef TTDB_DEBUG
	fp=fopen("common_debug_file", "a");
	fprintf(fp, "NULL transaction... --ttdb_c_del(TTDB.c)\n");
	fflush(fp);
	fclose(fp);
#endif
	return(TEMPORAL_TUC_ERR);
    } else {
        memset(&key, 0, sizeof(DBT));
        memset(&data, 0, sizeof(DBT));
#ifdef TTDB_DEBUG
	fp=fopen("common_debug_file", "a");
	fprintf(fp, "just before calling the underlying c_get --ttdb_c_del(TTDB.c)\n",
		db_strerror(ret));
	fprintf(fp, "sizeof(ttdbc->dbc) %d, sizeof(ttdbc) %d, sizeof(struct __ttdbc) %d --ttdb_c_del(TTDB.c)\n",
				sizeof(*(ttdbc->dbc)), sizeof(*ttdbc), sizeof(struct __ttdbc));
	fflush(fp);
	fclose(fp);
#endif
        if(ret=(ttdbc->dbc)->c_get((ttdbc->dbc), &key, &data, DB_CURRENT)){
            return(ret);
        }
        memcpy(&tt,data.data, sizeof(TTDATA));

        /* PARTITION Mani, 26th May 2003, Added code to check for
         * partitioned store.
		 * we can copy the current record, move it to the archive
		 * based on (logical, physical) condition then call c_del, so that
		 * the DB_EMPTY issue is handled by it and the cursor position is
		 * determined by BDB, we dont get into its issues. Since the version
		 * of delete doesnt use flags, it has been passed as 0 in the
		 * underlying calls. If flags are introduced, then this code needs
		 * to be modified accordingly.
         */

        if (is_partitioned(ttdbc->ttdb)) {

			/* The current store should contain only ID-UC or TIME-UC records, so if it is otherwise
				return err temporal,
				TODO : set up an ERR_PARTITIONED for use in places where
				the error is with respect to what the partitioned store expects */

			if (tt.state != ID_UC && tt.state != TIME_UC) return(TEMPORAL_TUC_ERR);


			if ((tt.state == ID_UC &&
				(!(tt.start.stamperId.txnid ==
					stamperId.txnid && tt.start.stamperId.envid == stamperId.envid)))
				|| tt.state == TIME_UC) {

				/* If ID-UC and the deleting txn is different from the
				 * creating txn then push to archive OR the state is TIME-UC
				 * then push to archive */

				if (tt.state == ID_UC) ((TTDATA*)data.data)->state = ID_ID;
				if (tt.state == TIME_UC) ((TTDATA*)data.data)->state = TIME_ID;

                ((TTDATA*)data.data)->stop.stamperId.txnid=
                								ttdbc->dbc->txn->txnid;
				((TTDATA*)data.data)->stop.stamperId.envid =
						((TEMPORAL_HANDLE *)(ttdbc->dbc->txn->mgrp->dbenv->temporal_handle))->envid;

				if((ret=(ttdbc->ttdb->adb)->put((ttdbc->ttdb->adb),
										ttdbc->dbc->txn, &key, &data, 0))!=0)
					return(ret);
			}

			return(ttdbc->dbc->c_del((ttdbc->dbc),0));

		}

		/* Mani End */

#ifdef TTDB_DEBUG
	fp=fopen("common_debug_file", "a");
	fprintf(fp,"ok. we completed the call to c_get...key,data: %s, %s--ttdb_c_del(TTDB.c)\n",
		(char*)(key.data), (char*)(data.data));
	fprintf(fp, "the tt.state is %d-ttdb_c_del(TTDB.c)\n", tt.state);
	fflush(fp);
	fclose(fp);
#endif
	stamperId.txnid = (ttdbc->dbc->txn)->txnid;
	stamperId.envid =
	    ((TEMPORAL_HANDLE *)(ttdbc->dbc->txn->mgrp->dbenv->temporal_handle))->envid;

        switch(tt.state){
	case ID_TIME:
	    /* one ID wasn't converted to time. error-
	     * stamper! */
	    return(TEMPORAL_TUC_ERR);
	case ID_ID:
	    /* check if this transaction created and deleted.
	     * should not be here. error-ttdb_c_del, elseif some other txn created it
	     * and some other txn deleted it then the record is valid and we can return
	     * DB_KEYEMPTY and not TEMPORAL_TUC_ERR */
#ifdef TTDB_DEBUG
	    fp=fopen("common_debug_file", "a");
	    fprintf(fp, "some ID error. --ttdb_c_del(TTDB.c)\n");
	    fflush(fp);
	    fclose(fp);
#endif
	    /* Checking if anyone of the ID is US */

	    /* If stop is US then record has been killed by us, so it should not
	     * be present If start is US then it cannot be stopped by someone else,
	     *  till we commit, so that case is anyways invalid
	     */
	    if ((tt.start.stamperId.txnid == stamperId.txnid &&
		 tt.start.stamperId.envid == stamperId.envid) ||
		(tt.stop.stamperId.txnid == stamperId.txnid &&
		 tt.stop.stamperId.envid == stamperId.envid))

		return(TEMPORAL_TUC_ERR);

	    /* If code reaches here then both start and stop are different from US
	     * so the record was created and deleted by someone else */
	    return (DB_KEYEMPTY);

	    break;
	case TIME_TIME:
	    /* some transaction created. some transaction
	     * deleted. the record is not there!*/
	case TIME_ID:
	    /* means that this record has already been deleted in this
	     * transaction or by some other txn*/
#ifdef TTDB_DEBUG
	    fp=fopen("common_debug_file", "a");
	    fprintf(fp, "someone else deleted ?... --ttdb_c_del(TTDB.c)\n");
	    fflush(fp);
	    fclose(fp);
#endif
	    return(DB_KEYEMPTY);
	    break;
	case ID_UC:
#ifdef TTDB_DEBUG
	    fp=fopen("common_debug_file", "a");
	    fprintf(fp, "delete it completely --ttdb_c_del(TTDB.c)\n");
	    fflush(fp);
	    fclose(fp);
#endif
	    /* means that this record created in this transaction or by some other
	     * txn, if same txn then physical delete, if some other txn then
	     * logical delete as done in the TIME_UC case
	     */

	     /****** The condition check here is not necessary since they have to be
	             the same txn, otherwise, there will be an error.
	             Modified by Yuhong (05/20/05)
	      ******/
	 	/*
	    if (tt.start.stamperId.txnid == stamperId.txnid &&
		tt.start.stamperId.envid == stamperId.envid) {
		if(ret=ttdbc->dbc->c_del((ttdbc->dbc),0))
		    return(ret);
		return(0);
	    }
	    */
		if(ret=ttdbc->dbc->c_del((ttdbc->dbc),0))
		    return(ret);
		return(0);

	    /* code reaches here if the start is not US, in that case a logical
	     * delete of the record needs to be done, which is the same for TIME_UC
	     * so let the code flow, without breaking out of switch */

	case TIME_UC:
#ifdef TTDB_DEBUG
	    fp=fopen("common_debug_file", "a");
	    fprintf(fp, "put a stop time --ttdb_c_del(TTDB.c)\n");
	    fflush(fp);
	    fclose(fp);
#endif
	    /* finally an undeleted transaction. we logically delete it by
	     * changing the limits. */
	    tt.state=TIME_TIME;
	    {
		struct timeval current;
		gettimeofday(&current, NULL);
		/*
		  current.tv_sec= time(NULL);
		  current.tv_usec=0;
		*/
		ttdbc->ttdb->set_current(ttdbc->ttdb, current);
	    }
	    /* this has to be changed to a transaction ID */
	    /* stamper is here!!
	       tt.stop.curr=time(NULL);
	    */
	    tt.stop.stamperId.txnid= (ttdbc->dbc->txn)->txnid;
	    tt.stop.stamperId.envid =
		((TEMPORAL_HANDLE *)(ttdbc->dbc->txn->mgrp->dbenv->temporal_handle))->envid;
	    memcpy(data.data,&tt, sizeof(TTDATA));


	    if(ret=ttdbc->dbc->c_put((ttdbc->dbc), &key, &data,
				     DB_CURRENT))
		return(ret);
	    return(0);
	    break;
	default:
#ifdef TTDB_DEBUG
	    fp=fopen("common_debug_file", "a");
	    fprintf(fp, "comes to default in the SWITCH statement!...--ttdb_c_del(TTDB.c)\n",
		    db_strerror(ret));
	    fflush(fp);
	    fclose(fp);
#endif
	    break;
        }
    }
}

/*
 * __ttdb_c_get --
 * temporal get using cursor
 * PUBLIC : int ttdb_c_get ( ( TTDBC *, DBT *, DBT *, u_int32_t));
 */
int
__ttdb_c_get(TTDBC *dbc_arg, DBT *key, DBT *data, u_int32_t flags)
{

    struct __ttdb *ttdbp;
    struct __db *db;
    struct __tt_wrapper  *tt_data;
    int not_valid_record;
    int ret;
    time_t ttdb_current_time;
    char *p;
#ifdef TTDB_DEBUG
    FILE *fp;
#endif

    ttdbp  =  dbc_arg->ttdb;
    {
	struct timeval current;
	gettimeofday(&current, NULL);
	/*
	  current.tv_sec= time(NULL);
	  current.tv_usec=0;
	*/
	ttdbp->set_current(ttdbp, current);
    }
    ttdb_current_time =  ttdbp->current_time.tv_sec;
#ifdef TTDB_DEBUG
    fp=fopen("common_debug_file", "a");
    fprintf(fp, "comes to this function key, data,flags: %s, %s, %d--ttdb_c_get(TTDB.c)\n",
	    key?((key->data)?key->data:"NULL"):"NULL",
	    data?((data->data)?data->data:"NULL"):"NULL", flags);
    fflush(fp);
    fclose(fp);
#endif

    /* check if the underlying db supports temporal queries */
    if (!is_tt(ttdbp)){
#ifdef TTDB_DEBUG
	fp=fopen("common_debug_file", "a");
	fprintf(fp, "no TT support --ttdb_c_get(TTDB.c)\n");
	fflush(fp);
	fclose(fp);
#endif

	return( (dbc_arg->dbc)->c_get( (dbc_arg->dbc),key,data,flags));
    } else {

	   /* PARTITION Mani, 26 May 2003, if the db is partitioned, then the
	    * implementation of c_get is same as normal db, since the current
	    * store has only valid (current) records, just do a get
	    * and return after stripping time related content
	    */

		if (is_partitioned(ttdbp)) {

				ret = (dbc_arg->dbc)->c_get( (dbc_arg->dbc),key,data,flags);
				convertNonTem(data);
				{
					struct timeval current;
					gettimeofday(&current, NULL);
					ttdbp->set_current(ttdbp, current);
				}
				return(ret);
		}

		/* Mani End */
	/* Mani, 28th Oct 2003, Checking for txnid = NULL, since MySQL tries
	 * call c_get without txnid when "use database" command is executed
	 */
	if (dbc_arg->dbc->txn == NULL) return(TEMPORAL_TUC_ERR);
	/* Mani, End 28th Oct 2003 */
	not_valid_record = 1 ;
	while (not_valid_record){
#ifdef TTDB_DEBUG
	    fp=fopen("common_debug_file", "a");
	    fprintf(fp, "while_not_valid_rec --ttdb_c_get(TTDB.c)\n");
	    fflush(fp);
	    fclose(fp);
#endif
	    if((ret = (dbc_arg->dbc)->c_get(
		(dbc_arg->dbc),key,data,flags))!=0){
#ifdef TTDB_DEBUG
		fp=fopen("common_debug_file", "a");
		fprintf(fp, "c_get returns error! %s--ttdb_c_get(TTDB.c)\n", db_strerror(ret));
		fflush(fp);
		fclose(fp);
#endif
		{
		    struct timeval current;
		    gettimeofday(&current, NULL);
		    /*
		      current.tv_sec= time(NULL);
		      current.tv_usec=0;
		    */
		    ttdbp->set_current(ttdbp, current);
		}
		return(ret);
	    }

	    /* retrieve the time data from the data field */
	    tt_data = (TTDATA *) data->data;
	    if(__validRec(data->data,ttdbp, dbc_arg->dbc->txn)){
#ifdef TTDB_DEBUG
		fp=fopen("common_debug_file", "a");
		fprintf(fp, "this one is a valid record --ttdb_c_get(TTDB.c)\n");
		fflush(fp);
		fclose(fp);
#endif
		convertNonTem (data);
		{
		    struct timeval current;
		    gettimeofday(&current, NULL);
		    /*
		      current.tv_sec= time(NULL);
		      current.tv_usec=0;
		    */
		    ttdbp->set_current(ttdbp, current);
		}
		return ret;
	    }
	}/*temporal check*/
    }
    {
        struct timeval current;
	gettimeofday(&current, NULL);
	/*
	  current.tv_sec= time(NULL);
	  current.tv_usec=0;
	*/
        ttdbp->set_current(ttdbp, current);
    }
    return ret;
}

/*
 * __ttdb_c_put --
 * Temporal put using cursor
 * PUBLIC : int put ( (TTDBC *, DBT *, DBT *,u_int32_t));
 */
int
__ttdb_c_put (TTDBC * dbc_arg, DBT *key, DBT *data, u_int32_t flags)
{

    int                 ret;
    void                *tt_data;
    TTDATA              tt;
    DBT                 new_dat;
    time_t              temp;
    TTDB *ttdb;

    /* PARTITION Mani, 26 May 2003, curr_dat field to be used to hold the
     * record in current store, which is going to be updated
     */
    DBT			curr_dat, curr_key;
    TTDATA *		ttDataPtr;
    /* Mani End */

    ttdb = dbc_arg->ttdb;

    if(!is_tt(ttdb)){
#ifdef TTDB_DEBUG
	FILE *fp;
	fp=fopen("common_debug_file", "a");
	fprintf(fp, "no TT support -- c_put(TTDB.c)\n");
	fflush(fp);
	fclose(fp);
#endif
	return( (dbc_arg->dbc)->c_put((dbc_arg->dbc),key,
				      (DBT*)data,flags));
    } else {
#ifdef TTDB_DEBUG
	FILE *fp;
	fp=fopen("common_debug_file", "a");
	fprintf(fp, "TT support -- c_put(TTDB.c)\n");
	fprintf(fp, "transaction ID: %d", (dbc_arg->dbc->txn)->txnid);
	fflush(fp);
	fclose(fp);
#endif
	if(dbc_arg->dbc->txn==NULL){
	    return(TEMPORAL_TUC_ERR);
	}

	tt.state=ID_UC;
	tt.start.stamperId.txnid=(dbc_arg->dbc->txn)->txnid;
	/* Mani, 26th May 2003, corrected code to set the env id also for the new
	 * record.
	 */
	tt.start.stamperId.envid=
		((TEMPORAL_HANDLE *)(dbc_arg->dbc->txn->mgrp->dbenv->temporal_handle))->envid;

	/* tt.stop.stamperId.txnid=UNTILCHANGED;
	   Commented, since state information tells us abt this */

	/* stamper is here!!
	   tt.state=TIME_UC;
	   tt.start.curr=time(NULL);
	   tt.stop.stamperId=UNTILCHANGED;
	*/

	tt_data=__getTTvoid(data->data, data->size, &tt);
	/* the above returns a malloc'ed void* after preappending the TTDATA*/

	memset(&new_dat, 0, sizeof(new_dat));
	new_dat.data=tt_data;
	new_dat.size=data->size + sizeof(TTDATA);       /* needs extra size*/
	new_dat.ulen=data->ulen;
	new_dat.dlen=data->dlen;
	new_dat.doff=data->doff;
	new_dat.flags=data->flags;

	/* PARTITION Mani, 26th May 2003, If partitioned then
	 * if flag = DB_CURRENT update the current record and archive it based
	 * on logical/physical condition,
	 * if flag = DB_FIRST or DB_LAST then pass the data to the underlying
	 * c_put with the time info attached to the data, so that a new record
	 * is created. Basically, underlying c_put needs to be called with
	 * appropriate data based on flags. Before this call the record may
	 * have to be archived
	 */
	if (is_partitioned(ttdb)) {

		if (flags == DB_CURRENT) {

			memset(&curr_dat, 0, sizeof(curr_dat));
			memset(&curr_key, 0, sizeof(curr_key));
			curr_dat.flags = DB_DBT_MALLOC;
			ret = (dbc_arg->dbc)->c_get((dbc_arg->dbc), &curr_key, &curr_dat, DB_CURRENT);
			if (ret == 0) {
				/* check the start time and decide if record needs to go to the archive or not */
				ttDataPtr = (TTDATA *) curr_dat.data;
				if (!(ttDataPtr->state == ID_UC &&
					  ttDataPtr->start.stamperId.txnid ==
					  dbc_arg->dbc->txn->txnid &&
					  ttDataPtr->start.stamperId.envid ==
					  ((TEMPORAL_HANDLE *)(dbc_arg->dbc->txn->mgrp->dbenv->temporal_handle))->envid)) {

					/* So the current record need to be pushed to the archive
					* Ideally, only records of type ID_UC or TIME_UC can exist in the current store */
					if (ttDataPtr->state == ID_UC) ttDataPtr->state = ID_ID;
					if (ttDataPtr->state == TIME_UC) ttDataPtr->state = TIME_ID;

					ttDataPtr->stop.stamperId.txnid = dbc_arg->dbc->txn->txnid;
					ttDataPtr->stop.stamperId.envid =
						((TEMPORAL_HANDLE *)(dbc_arg->dbc->txn->mgrp->dbenv->temporal_handle))->envid;
					ret=(ttdb->adb)->put((ttdb->adb), dbc_arg->dbc->txn, &curr_key, &curr_dat, flags);
				}
			}
		}
		{
			struct timeval current;
			gettimeofday(&current, NULL);
			ttdb->set_current(ttdb, current);
		}
		return( (dbc_arg->dbc)->c_put((dbc_arg->dbc),key,&new_dat,flags));
	}
	/* Mani End */

	/* special things to do when unique keys is set */
	if(__ttdb_has_dup(ttdb)){
	    temp=ttdb->current_time.tv_sec;
	    {
		struct timeval current;
		gettimeofday(&current, NULL);
		/*
		  current.tv_sec= time(NULL);
		  current.tv_usec=0;
		*/
		ttdb->set_current(ttdb, current);
	    }
	    ret=__ttdb_del(ttdb, (dbc_arg->dbc)->txn, key, 0);
	    ret=(ttdb->db)->put((ttdb->db), (dbc_arg->dbc)->txn,
				key, &new_dat, flags);
	    {
		struct timeval current;
		gettimeofday(&current, NULL);
		/*
		  current.tv_sec= time(NULL);
		  current.tv_usec=0;
		*/
		ttdb->set_current(ttdb, current);
	    }
	    return(ret);
	}
	{
	    struct timeval current;
	    gettimeofday(&current, NULL);
	    /*
	      current.tv_sec= time(NULL);
	      current.tv_usec=0;
	    */
	    ttdb->set_current(ttdb, current);
	}
	return( (dbc_arg->dbc)->c_put((dbc_arg->dbc),key,&new_dat,flags));
    }
}

/*
 * __ttdb_c_close --
 * function to close the cursor
 */

int __ttdb_c_close(TTDBC *cursor){

	/* PARTITION Mani */
	int ret;

	/* Mani, 13 April 2003, Added code to close the archive db cursor,
	 * if partitioned db
	 */
	if (is_partitioned(cursor->ttdb))
		if ((ret = (cursor->adbc)->c_close(cursor->adbc)) != 0) return ret;

	/* Mani End */

    return((cursor->dbc)->c_close(cursor->dbc));
}

/* sets the current time of retrieval from the database */
int set_current(struct __ttdb *ttdb, struct timeval curr){
    ttdb->current_time.tv_sec=curr.tv_sec;
    ttdb->current_time.tv_usec=curr.tv_usec;
    return(0);
}

/* returns the time in the database */
struct timeval get_current(struct __ttdb *ttdb){
    return(ttdb->current_time);
}

/*
 * __ttdb_close --
 * to preserve some sense, we have to take off the DB_AM_DUP flag when we
 * write back a temporal DB which had been opened with unique keys. (recall
 * that during open, every temporal database had the DUP (duplicate) flags set
 * because temporal support requires that. But so that another open does not
 * fail, we will restore the flags
 */
int __ttdb_close(struct __ttdb *ttdb, u_int32_t flags){

	/* PARTITION Mani, 18th Mar 2003 */
	int ret;
	/* Mani End */

    if(is_tt(ttdb) && (!__ttdb_has_dup(ttdb))){
        (ttdb->db)->flags &= (~DB_AM_DUP);
    }

	/* PARTITION Mani, 18th Mar 2003, Modified code to close the archive db
	 * also Not changing the flags of the adb to remove AM_DUP, since the adb
	 * needs to have capability to store duplicates,
	 * further changed on 28th Mar 2003 to check for partitioned flag
	 */
   	if (is_partitioned(ttdb)) {
    		if ((ret = ((ttdb->db)->close)((ttdb->db), flags)) != 0)
				return ret;
			else
				return(((ttdb->adb)->close)((ttdb->adb), flags));
	}
	else
    	return(((ttdb->db)->close)((ttdb->db), flags));

    /* Mani End */

}

/*
 * __ttdb_set_flags --
 * directly passes control to the underlying database
 */
int __ttdb_set_flags(struct __ttdb *ttdb, u_int32_t flags){
    return((ttdb->db)->set_flags((ttdb->db), flags));
}

int __tt_cursor(struct __ttdb *ttdb, DB_TXN *txnid, struct __ttdbc **cursorp, u_int32_t flags){
    struct __dbc     *dbc;
    /* PARTITION Mani, 13 April 2003, declared pointer to archive db cursor */
    struct __dbc	*adbc;
	/* Mani End */
    int ret;
#ifdef TTDB_DEBUG
    FILE *fp;
    fp=fopen("common_debug_file", "a");
    fprintf(fp, "a TTDBC gets created --__tt_cursor(TTDB.c)\n");
    fflush(fp);
    fclose(fp);
#endif
    (*cursorp)=(struct __ttdbc*)malloc(sizeof(struct __ttdbc));
    if((ret=(ttdb->db)->cursor((ttdb->db), txnid, &dbc, flags))==0){
        (*cursorp)->dbc=dbc;

	/* PARTITION Mani, 13 April 2003, added code to create a cursor to run
	 * thru the archive db if the db is partitioned
	 */

	if (is_partitioned(ttdb)) {

		if((ret=(ttdb->adb)->cursor((ttdb->adb), txnid, &adbc, flags))==0){
			(*cursorp)->adbc=adbc;
			/* Mani, 25 May 2003, added code to initialize the member variables */
			(*cursorp)->moveCurrentFlag = 0;
			(*cursorp)->moveArchiveFlag = 0;
			(*cursorp)->firstArchiveRec = 1;
			(*cursorp)->firstCurrentRec = 1;
		}
	}
	/* Mani End */

#ifdef TTDB_DEBUG
	fp=fopen("common_debug_file", "a");
	fprintf(fp, "the size of (cursorp->dbc) is %d, sizeof(cursorp) %d, sizeof(struct __dbc) %d __tt_cursor(TTDB.c)\n",
		sizeof(*((*cursorp)->dbc)), sizeof(**cursorp), sizeof(struct __dbc));
	fflush(fp);
	fclose(fp);
#endif
        (*cursorp)->ttdb=ttdb;
        (*cursorp)->c_del=__ttdb_c_del;
        (*cursorp)->c_get=__ttdb_c_get;
        (*cursorp)->c_get_nonsequenced =__ttdb_c_get_nonsequenced;
        (*cursorp)->c_put=__ttdb_c_put;
        (*cursorp)->c_close=__ttdb_c_close;
    }
    return(ret);
}

/*
 * __ttdb_c_get_nonsequenced --
 * this function chooses records which are valid
 * between particular times, and also puts the begin and end times in the
 * addresses passed to it. returns 0 on success, an error code on failure
 */

/* PARTITION Mani, the new get non sequenced makes calls to the internal non
 * sequenced function with archive db cursor or current db cursor
 */

int __ttdb_c_get_nonsequenced(struct __ttdbc *ttdbc, DBT *key, DBT *data,
        struct timeval start, struct timeval stop, u_int32_t flags, struct
        timeval *outstart, struct timeval *outstop)
{
	struct __dbc *dbCursor;
	struct __ttdb   *ttdb;
	DBT 	archiveKey, archiveData;
	DBT		dummyKey, dummyData;
	struct timeval archiveOutstart, archiveOutstop;
	int 	retFromArchive, retFromCurrent;

    ttdb  =  ttdbc->ttdb;
    if (!is_tt(ttdb)){
        return(TEMPORAL_TUC_ERR);
    }
    else{

		/* Mani, 07 May 2003, checking for partitoned db */
		if (!(is_partitioned(ttdb))) {

			#ifdef TTDB_DEBUG
			printf(" -- Non Seq Select, the curr (envid,txnid) : (%d,%d) \n",
			       ((TEMPORAL_HANDLE *)((ttdbc->dbc)->txn->mgrp->dbenv->temporal_handle))->envid,
			       (ttdbc->dbc)->txn->txnid);
			#endif

			return (__ttdb_c_get_nonsequenced_internal(ttdbc->dbc, key, data,
						start, stop, flags, outstart, outstop));
		}
		else {

			/* Idea : Call the internal get non sequenced, check which data needs to be returned to the user and
			return appropriately, also ensure that the cursor is positioned correctly
			TODO : DB_SET as a special case, since in that case the key has to be preserved
			TODO : Currently the code is oriented towards the flags input being DB_NEXT,
			some scenarios with respect to moveArchiveFlag, moveCurrentFlag, firstCurrentRec, firstArchiveRec
			may have to handled differently, when considering all possible values for the flags input variable. */

			#ifdef TTDB_DEBUG
			printf(" -- Non Seq Select, the curr (envid,txnid) : (%d,%d), archive (envid,txnid) : (%d,%d) \n", 	((TEMPORAL_HANDLE *)((ttdbc->dbc)->txn->mgrp->dbenv->temporal_handle))->envid, (ttdbc->dbc)->txn->txnid, ((TEMPORAL_HANDLE *)((ttdbc->adbc)->txn->mgrp->dbenv->temporal_handle))->envid, (ttdbc->adbc)->txn->txnid);
			#endif

			if (ttdbc->moveCurrentFlag == 1) {

				memset(&dummyKey, 0, sizeof(dummyKey)); /* TODO : DB_SET flag */
				memset(&dummyData, 0, sizeof(dummyData));
				retFromArchive = (ttdbc->dbc)->c_get((ttdbc->dbc), &dummyKey,
							&dummyData, DB_PREV); /* TODO : DB_SET flag */
				ttdbc->moveCurrentFlag = 0;

			}

			if (ttdbc->moveArchiveFlag == 1) {

				memset(&dummyKey, 0, sizeof(dummyKey)); /* TODO : DB_SET flag */
				memset(&dummyData, 0, sizeof(dummyData));
				retFromCurrent = (ttdbc->adbc)->c_get((ttdbc->adbc), &dummyKey,
							&dummyData, DB_PREV); /* TODO : DB_SET flag */
				ttdbc->moveArchiveFlag = 0;
			}

			if (ttdbc->firstCurrentRec == 1) {

				retFromCurrent = __ttdb_c_get_nonsequenced_internal(ttdbc->dbc,
						key, data, start, stop, DB_FIRST, outstart, outstop);
				/* If error is DB NOT FOUND, it could be because the first record did not satisfy the criteria, so try next record */
				if (retFromCurrent == DB_NOTFOUND)
					retFromCurrent = __ttdb_c_get_nonsequenced_internal(
						ttdbc->dbc, key, data, start, stop, flags, outstart, outstop);
			}
			else
				retFromCurrent = __ttdb_c_get_nonsequenced_internal(ttdbc->dbc,
					key, data, start, stop, flags, outstart, outstop);

			if (retFromCurrent == TEMPORAL_TUC_ERR) return (TEMPORAL_TUC_ERR);
			if (retFromCurrent == DB_NOTFOUND) {

				#ifdef TTDB_DEBUG
				printf(" Returned Record From ARCHIVE STORE \n");
				#endif

				if (ttdbc->firstArchiveRec == 1) {

					ttdbc->firstArchiveRec = 0;
					retFromArchive = __ttdb_c_get_nonsequenced_internal(
						ttdbc->adbc, key, data, start, stop, DB_FIRST, outstart, outstop);
					/* If error is DB NOT FOUND, it could be because the first record did not satisfy the criteria, so try next record */
					if (retFromArchive == DB_NOTFOUND)
						return (__ttdb_c_get_nonsequenced_internal(
							ttdbc->adbc, key, data, start, stop, flags, outstart, outstop));
					else
						return (retFromArchive);
				}
				else
					return (__ttdb_c_get_nonsequenced_internal(ttdbc->adbc,
							key, data, start, stop, flags, outstart, outstop));
			}

			/* At this point, the current cursor has returned data, so go ahead and check the archive cursor */
			memset(&archiveKey, 0, sizeof(archiveKey)); /* TODO : DB_SET flag */
			memset(&archiveData, 0, sizeof(archiveData));

			if (ttdbc->firstArchiveRec == 1) {

				ttdbc->firstArchiveRec = 0;
				retFromArchive = __ttdb_c_get_nonsequenced_internal(
					ttdbc->adbc, &archiveKey, &archiveData, start, stop,
					DB_FIRST, &archiveOutstart, &archiveOutstop);
				/* If error is DB NOT FOUND, it could be because the first record did not satisfy the criteria, so try next record */
				if (retFromArchive == DB_NOTFOUND)
					retFromArchive = __ttdb_c_get_nonsequenced_internal(
						ttdbc->adbc, &archiveKey, &archiveData, start, stop,
						flags, &archiveOutstart, &archiveOutstop);
			}
			else
				retFromArchive = __ttdb_c_get_nonsequenced_internal(
					ttdbc->adbc, &archiveKey, &archiveData, start, stop,
					flags, &archiveOutstart, &archiveOutstop);

			if (retFromArchive == TEMPORAL_TUC_ERR) return (TEMPORAL_TUC_ERR);

			if (retFromArchive == DB_NOTFOUND) {

				ttdbc->firstCurrentRec = 0;
				#ifdef TTDB_DEBUG
				printf(" Returned Record From CURRENT STORE \n");
				#endif
				return (0);
			}

			/* At this point both archive and the current have data, so decide which one to return */

			/* What if, the two keys had different sizes. Well, we can assume that the systems using
			BDB will have equally sized keys and data */

			/*
			printf("archive key =%d, data = %s \n", *(int *)(archiveKey.data), (char*)(archiveData.data));
			printf("current key =%d, data = %s \n", *(int *)(key->data), (char*)(data->data));
			printf("memcmp archive, current = %d \n", memcmp(archiveKey.data, key->data, key->size));
			*/


			/* If the archive key is less than or equal to the current store key, then it gets to go first */
			if (memcmp(archiveKey.data, key->data, key->size) <= 0) {

				/* we need to return the archive data and key, by copying them into the user field */

				outstart->tv_sec = archiveOutstart.tv_sec;
				outstart->tv_usec = archiveOutstart.tv_usec;

				outstop->tv_sec = archiveOutstop.tv_sec;
				outstop->tv_usec = archiveOutstop.tv_usec;

				memcpy(key->data, archiveKey.data, key->size);
				memcpy(data->data, archiveData.data, data->size);

				/* the other fields in the DBT structure must have been set by the get call */

				/* move the current store cursor to prev record, so that it will be considered in the
					next call */
				ttdbc->moveCurrentFlag = 1;
				ttdbc->firstArchiveRec = 0;


				/* TODO : where to free the mem allocated to archiveData and archiveKey -> data field */

				#ifdef TTDB_DEBUG
				printf(" Returned Record From ARCHIVE STORE \n");
				#endif

				return (0);

			}
			else {

				/* we need to return the current store record, so no copying required,
					move the archive store to prev record */

				ttdbc->moveArchiveFlag = 1;
				ttdbc->firstCurrentRec = 0;
				#ifdef TTDB_DEBUG
				printf(" Returned Record From CURRENT STORE \n");
				#endif

				return (0);
			}
		}
	}
}

/* Mani End */

/*
 * __ttdb_c_get_nonsequenced_internal --
 *
 * PARTITION Mani, 7 May 2003
 * This function does the get non sequenced operation as designed in T-BDB.
 * The only difference is that it works on the cursor given as input
 * So, modify the function signature to have the cursor as input,
 * and change all occurences of ttdbc->dbc by dbCursor, and rename
 * function as internal
 */

int __ttdb_c_get_nonsequenced_internal(struct __dbc *dbCursor, DBT *key, DBT *data,
        struct timeval start, struct timeval stop, u_int32_t flags, struct
        timeval *outstart, struct timeval *outstop)
{
    struct __ttdb   *ttdb;
    int             ret;
    TTDATA          *ttdata;
    STP_ID		stamperId;
    struct timeval recStart, recStop;

    if (start.tv_sec == BEGINNING) start.tv_sec = -1;
    while(!(ret=(dbCursor)->c_get((dbCursor), key, data, flags))){
	ttdata=(TTDATA*)data->data;
	stamperId.txnid = (dbCursor->txn)->txnid;
	stamperId.envid =
	    ((TEMPORAL_HANDLE *)(dbCursor->txn->mgrp->dbenv->temporal_handle))->envid;

	switch(ttdata->state){
	case ID_ID:
	    /* This case can occur, the ID values might be of different txns */
	    /* both the ID values cannot be US, similarly the start being US and
	     * end being OTHER is also not possible
	     */

	    if ((ttdata->start.stamperId.txnid == ttdata->stop.stamperId.txnid) &&
		(ttdata->start.stamperId.envid == ttdata->stop.stamperId.envid)) {
		return(TEMPORAL_TUC_ERR);
	    }

	    /* If start is US then it cannot be stopped by someone else,
	     * till we commit, so that case is invalid
	     */
	    if ((ttdata->start.stamperId.txnid == stamperId.txnid) &&
		(ttdata->start.stamperId.envid == stamperId.envid)){
		return(TEMPORAL_TUC_ERR);
	    }

	    /* So, after rejecting the wrong cases, pass the OTHER ID to stp
	     * module to get the timeval structure representing that ID. Finally
	     * we will have two timeval structures which represent the start
	     * and stop of the record and we will have two boundary structures
	     * which will be start and stop passed by the user. We have to check
	     * if the record is valid.
	     */

	    if ((ttdata->stop.stamperId.txnid == stamperId.txnid)
		&& (ttdata->stop.stamperId.envid == stamperId.envid)) {
		outstop->tv_sec=US;
		outstop->tv_usec=0;
		recStart = __stp_get_ctime(dbCursor->txn->mgrp->dbenv,
					   ttdata->start.stamperId);
		if (recStart.tv_sec !=0)
		    *outstart = recStart;
		else {
		    outstart->tv_sec = OTHER;
		    outstart->tv_usec = 0;
		}
		convertNonTem(data);
		return(0);
	    } else {
		recStart = __stp_get_ctime(dbCursor->txn->mgrp->dbenv,
					   ttdata->start.stamperId);
		recStop = __stp_get_ctime(dbCursor->txn->mgrp->dbenv,
					  ttdata->stop.stamperId);
		if (recStart.tv_sec != 0 && recStop.tv_sec !=0) {
		    /* Treat the record in the same way as TIME_TIME */
		    if(recStart.tv_sec <= start.tv_sec){
			if(recStop.tv_sec <= start.tv_sec)
			    continue;
			*outstart=start;
			if(recStop.tv_sec >= stop.tv_sec)
			    *outstop=stop;
			else
			    *outstop=ttdata->stop.curr;
			convertNonTem(data);
			return(0);
		    } else if(recStart.tv_sec <= stop.tv_sec){
			*outstart=ttdata->start.curr;
			if(recStop.tv_sec >= stop.tv_sec)
			    *outstop=stop;
			else
			    *outstop=ttdata->stop.curr;
			convertNonTem(data);
			return(0);
		    }
		} else {
		    /* one of them is OTHER, so treat like TIME_ID case */
		    /* TODO : Might have to confirm if the record falls in the
		     * interval start to stop
		     */
		    if (recStart.tv_sec == 0) {
			outstart->tv_sec = OTHER;
			outstart->tv_usec = 0;
		    } else
			*outstart = recStart;

		    if (recStop.tv_sec == 0) {
			outstop->tv_sec = OTHER;
			outstop->tv_usec = 0;
		    } else
			*outstop = recStop;
		    convertNonTem(data);
		    return(0);
		}
	    }
	    continue;
	case ID_UC:
	    /* a record that is current and has been written by this txn or some other txn
	       so the stop time is user passed stop time
	       start time is US if the txn is me and the other txn's time if it is not me
	       the other txn cannot be uncommitted, if so, i shouldnt have seen this record
	       since this is Don't Allow Dirty Read area */

	    if (ttdata->start.stamperId.txnid == stamperId.txnid &&
		ttdata->start.stamperId.envid == stamperId.envid){
		outstart->tv_sec=US;
		outstart->tv_usec=0;

		outstop->tv_sec=UC; /* Stanley Yao (08/01/03) */
		outstop->tv_usec=0;

		convertNonTem(data);
		return(0);
	    } else {
		recStart = __stp_get_ctime(dbCursor->txn->mgrp->dbenv,
					   ttdata->start.stamperId);
		if (recStart.tv_sec == 0) {
		    outstart->tv_sec = OTHER;
		    outstart->tv_usec = 0;

		    outstop->tv_sec=UC; /* Stanley Yao (08/01/03) */
		    outstop->tv_usec=0;

		    convertNonTem(data);
		    return(0);
		} else {
		    /* Treat like TIME_UC */
		    if(recStart.tv_sec >= stop.tv_sec)
			continue;
		    if(recStart.tv_sec <= start.tv_sec)
			*outstart=start;
		    else
			*outstart=recStart;

		    outstop->tv_sec=UC; /* Stanley Yao (08/01/03) */
		    outstop->tv_usec=0;

		    convertNonTem(data);
		    return(0);
		}
	    }
	    /* unreached statement. --Shilong
	       continue;
	    */
	case ID_TIME:
	    /* This case is not possible, since stamper would have stamped
	       both ID's in the record, it would have become TIME_TIME */
	    /*
	      if(ttdata->stop.curr.tv_sec >= stop.tv_sec){
	      outstart->tv_sec=US;
	      outstart->tv_usec=0;
	      outstop->tv_sec=OTHER;
	      outstop->tv_usec=0;
	      convertNonTem(data);
	      return(0);
	      }
	      else
	      continue;
	    */
	    return(TEMPORAL_TUC_ERR);
	case TIME_TIME: /* regular begin and end to data */
	    if(ttdata->start.curr.tv_sec <= start.tv_sec){
		if(ttdata->stop.curr.tv_sec <= start.tv_sec)
		    continue;
		*outstart=start;
		if(ttdata->stop.curr.tv_sec >= stop.tv_sec)
		    *outstop=stop;
		else
		    *outstop=ttdata->stop.curr;
		convertNonTem(data);
		return(0);
	    } else if(ttdata->start.curr.tv_sec <= stop.tv_sec){
		*outstart=ttdata->start.curr;
		if(ttdata->stop.curr.tv_sec >= stop.tv_sec)
		    *outstop=stop;
		else
		    *outstop=ttdata->stop.curr;
		convertNonTem(data);
		return(0);
	    } else /* the data began after the chosen interval: invalid */
		continue;
	case TIME_ID:
	    /* someone else wrote, we have deleted or someone else has deleted
	       So, here again find if the ID represents US or someone else
	       If US then return US, else check if record falls inside
	       the time range based on the TIME of the other txn */

	    if(ttdata->start.curr.tv_sec >= stop.tv_sec)
		continue;
	    if(ttdata->start.curr.tv_sec <= start.tv_sec)
		*outstart=start;
	    else
		*outstart=ttdata->start.curr;

	    if (ttdata->stop.stamperId.txnid == stamperId.txnid &&
		ttdata->stop.stamperId.envid == stamperId.envid) {
		outstop->tv_sec=US;
		outstop->tv_usec=0;
		convertNonTem(data);
		return(0);
	    } else {
		recStop = __stp_get_ctime(dbCursor->txn->mgrp->dbenv,
					  ttdata->stop.stamperId);
		if (recStop.tv_sec == 0) {
		    outstop->tv_sec = OTHER;
		    outstop->tv_usec = 0;
		    convertNonTem(data);
		    return(0);
		} else {
		    recStart = ttdata->start.curr;
		    /* Treat like TIME_TIME, since we have two time values now */
		    if(recStart.tv_sec <= start.tv_sec){
			if(recStop.tv_sec <= start.tv_sec)
			    continue;
			*outstart=start;
			if(recStop.tv_sec >= stop.tv_sec)
			    *outstop=stop;
			else
			    *outstop=ttdata->stop.curr;
			convertNonTem(data);
			return(0);
		    } else if(recStart.tv_sec <= stop.tv_sec){
			*outstart=ttdata->start.curr;
			if(recStop.tv_sec >= stop.tv_sec)
			    *outstop=stop;
			else
			    *outstop=ttdata->stop.curr;
			convertNonTem(data);
			return(0);
		    }
		}
	    }
	    continue;
	case TIME_UC:
	    if(ttdata->start.curr.tv_sec >= stop.tv_sec)
		continue;
	    if(ttdata->start.curr.tv_sec <= start.tv_sec)
		*outstart=start;
	    else
		*outstart=ttdata->start.curr;

	    outstop->tv_sec=UC; /* Stanley Yao (08/01/03) */
	    outstop->tv_usec=0;

	    convertNonTem(data);
	    return(0);
	default:
	    return(TEMPORAL_TUC_ERR);
	}
	/* don't recurse on DB_FIRST or DB_LAST */
	/* Not reached, --Shilong 05/27/03
	   if((flags & DB_FIRST) || (flags & DB_LAST))
	   return DB_NOTFOUND;
	*/
    }
    return DB_NOTFOUND;
}

/* finds the data that are valid in a certain time period */
/*
 * this is the old function that has now been changed
 */
/*
int __ttdb_c_get_nonsequenced(struct __ttdbc *dbc_arg,DBT *key,DBT *data,
        u_int32_t flags,time_t start,time_t stop)
{
   struct __ttdb *ttdbp;
   struct __db *db;
   struct __tt_wrapper  *tt_data;
   int not_valid_record;
   int ret;
   time_t ttdb_current_time;
   char *p;

   ttdbp  =  dbc_arg->ttdb;

   if (!is_tt(ttdbp))
   {
     return( TEMPORAL_TUC_ERR);
   }
   else{
        not_valid_record = 1 ;
        while (not_valid_record){
            if((ret = (dbc_arg->dbc)->c_get(
                          (dbc_arg->dbc),key,data,flags))!=0){
                ttdbp->set_current(ttdbp, ttdbp, time(NULL));
                return(ret);
            }

            /* retrieve the time data  from the data field
            tt_data = (TTDATA *) data->data;
            if (tt_data->state == TIME_UC || tt_data->state == TIME_TIME)
            {
                if (((tt_data->start.curr <= start)
                        &&((tt_data->stop.stamperId.txnid == UNTILCHANGED)
                                ||(tt_data->stop.curr >= start)))
                   ||((tt_data->start.curr >= start)
                        && ( tt_data->start.curr < stop))) {
                        ttdbp->set_current(ttdbp, ttdbp, time(NULL));
                        return ret;
                     }
                }
        }/*temporal check
    }
    ttdbp->set_current(ttdbp, ttdbp, time(NULL));
    return ret;
}
*/

/*
 * __ttdb_err --
 * takes care of one special case, and passes the rest to the underlying
 * database error function
 */
void __ttdb_err(struct __ttdb* ttdb, int err, const char *fmt, ...){
    switch(err) {
    case TEMPORAL_TUC_ERR:
        printf("%s: Temporal Conflict Error\n", fmt);
	break;
    default:
        (ttdb->db)->err((ttdb->db), err, fmt);
    }
}

/* now writing the other functions that are provided by the regular BerkeleyDB */
/*
int __ttdb_associate(struct __ttdb *primary, struct __ttdb *secondary,
        int (*callback)(struct __ttdb *, const DBT *, const DBT *, DBT *),
        u_int32_t flags){

    return((primary->db)->associate((primary->db), (secondary->db), callback,
                flags));
}

int __ttdb_fd(struct __ttdb *db, int *fdp){
    return((db->db)->fd((db->db), fdp));
}

int __ttdb_get_byteswapped(struct __ttdb *db, int *isswapped){
    return((db->db)->get_byteswapped((db->db), isswapped));
}

int __ttdb_get_type(struct __ttdb *db, DBTYPE *type){
    return((db->db)->get_type((db->db), type));
}

int __ttdb_join(struct __ttdb *primary, struct __ttdbc **curslist, struct
        __ttdbc **dbcp, u_int32_t flags){

    return((primary->db)->join(primary->db, curslist, **dbcp, flags));
}
*/

void __ttdb_set_errfile(struct __ttdb *db, FILE *errfile){
    (db->db)->set_errfile(db->db, errfile);
}

void __ttdb_set_errpfx(struct __ttdb *db, const char *errpfx){
    (db->db)->set_errpfx(db->db, errpfx);
}

int __ttdb_set_cachesize(struct __ttdb *db, u_int32_t gbytes, u_int32_t bytes, int
			 ncache){
    (db->db)->set_cachesize(db->db, gbytes, bytes, ncache);
}

int __ttdb_set_pagesize(struct __ttdb *db, u_int32_t pagesize){
    (db->db)->set_pagesize(db->db, pagesize);
}

/*int (*cmp_db)(struct __db *, const DBT *, const DBT *);*/

int (*cmp_ttdb)(struct __ttdb *, const DBT *, const DBT *);

int func_cmp_db(struct __db *db, const DBT *dbt1, const DBT *dbt2){
    struct __ttdb *ttdb1;

    __ttdb_create(&ttdb1, NULL, 0);
    ttdb1->db=db;
    return(cmp_ttdb(ttdb1, dbt1, dbt2));
}

/*
 * __ttdb_set_bt_compare --
 * this function takes in another function that's used to compare two
 * databases. Since the function is sent from the user, we do not change it to
 * accept struct __ttdb... it accepts the regular DB */
int __ttdb_set_bt_compare(struct __ttdb* db, int (*func)(struct __ttdb*,
							 const DBT *, const DBT*)){
    {
        /* we have to pass another function that takes in a __db instead of
         * the function that takes in a __ttdb
         */
        cmp_ttdb= func;
        return((db->db)->set_bt_compare((db->db), func_cmp_db));
    }
}

int  __ttdb_remove (struct __ttdb* db, const char *str1, const char *str2,
		    u_int32_t flags){
    return((db->db)->remove((db->db), str1, str2, flags));
}

#ifdef BDB_4
int  __ttdb_stat(struct __ttdb *db, void *vd1, u_int32_t flags){
    return((db->db)->stat((db->db), vd1, flags));
}
#endif
#ifdef BDB_3
int  __ttdb_stat(struct __ttdb *db, void *vd1, void *(*vd2)(size_t), u_int32_t flags){
    return((db->db)->stat((db->db), vd1, vd2, flags));
}
#endif

int  __ttdb_key_range(struct __ttdb *db, DB_TXN *txn, DBT *dbt, DB_KEY_RANGE
		      *range, u_int32_t flags){
    return((db->db)->key_range((db->db), txn, dbt, range, flags));
}

/*
 * __ttdb_create --
 * redefines the methods to point to the functions that we wrote.
 */
int __ttdb_create(struct __ttdb **ttdb, DB_ENV *env, u_int32_t flags){
    struct __db    *db;
    int      ret;
	/* Mani, 16th Mar 2003 */
	struct __db    *adb;
	/* Mani End */

    *ttdb=(struct __ttdb*)malloc(sizeof(struct __ttdb));
    if(ret=db_create(&db, env, flags))
        return(ret);
    (*ttdb)->db=db;

	/* Mani, 16th Mar 2003, constructing the archive db structure, no checks
	 * done here, since this just the structure creation, which can be done
	 * by default
	 */
	if(ret=db_create(&adb, env, flags))
        	return(ret);
    	(*ttdb)->adb=adb;

	/* Mani End */

    /*
      this modification was found necessary for use in set_bt_compare, but we
      can do without it...
      db->ttdb=(void*)(*ttdb);
    */
    (*ttdb)->open=__ttdb_open;
    (*ttdb)->is_tt=is_tt;
    (*ttdb)->put=__ttdb_put;
    (*ttdb)->get=__ttdb_get;
    (*ttdb)->del=__ttdb_del;
    (*ttdb)->set_current=set_current;
    (*ttdb)->get_current=get_current;
    (*ttdb)->close=__ttdb_close;
    (*ttdb)->set_flags=__ttdb_set_flags;
    (*ttdb)->cursor=__tt_cursor;
    (*ttdb)->err=__ttdb_err;
    (*ttdb)->set_errfile=__ttdb_set_errfile;
    (*ttdb)->set_errpfx=__ttdb_set_errpfx;
    (*ttdb)->set_pagesize=__ttdb_set_pagesize;
    (*ttdb)->set_cachesize=__ttdb_set_cachesize;
    (*ttdb)->set_bt_compare=__ttdb_set_bt_compare;
    (*ttdb)->remove=__ttdb_remove;
    (*ttdb)->stat=__ttdb_stat;
    (*ttdb)->key_range=__ttdb_key_range;
    return(0);
}

